/**
 * \file	blitz.h
 * \brief	Complete API for using the BlitzMax runtime
 *
 * \mainpage BlitzMaxExtended
 *
 * \section	intro_sec Introduction
 *
 * Any BlitzMax application uses this C runtime library and the assembly code
 * created by the bcc from the actual BlitzMax code which is assemled by
 * the platform's assembler.
 *
 * That means this C code has a great influence on the execution of
 * any application or module. Changing the expected behaviour of the API
 * can therefore lead to error-prone behaviour. You have been warned :)
 *
 * \section	what_is_bme_sec What is BlitzMaxExtended?
 *
 * BlitzMaxExtended is a modification of this runtime library to ease development
 * from C/C++, speed up existing applications and add new features and stability.
 * While some features may have been renamed or replaced by newer ones,
 * in general the API stays the same and legacy code should not be affected.
 *
 * \version	0.8.16
 * \copyright zlib/libpng
 * \author	Mark Sibly
 * \author	Fabian Niemann
 * \date	2012
 */

#ifndef BLITZ_H
#define BLITZ_H

#include "c/_common.h"

#include "c/blitz_types.h"
#include "c/blitz_memory.h"
#include "c/blitz_thread.h"
#include "c/blitz_gc.h"
#include "c/blitz_ex.h"
#include "c/blitz_cclib.h"
#include "c/blitz_debug.h"
#include "c/blitz_module.h"
#include "c/blitz_incbin.h"
#include "c/blitz_class.h"
#include "c/blitz_object.h"
#include "c/blitz_string.h"
#include "c/blitz_array.h"
#include "c/blitz_handle.h"
#include "c/blitz_app.h" 

//	Exceptions

#include "c/exceptions/blitz_exception.h"

#include "c/exceptions/blitz_blitzexception.h"
#include "c/exceptions/blitz_nullobjectexception.h"
#include "c/exceptions/blitz_nullmethodexception.h"
#include "c/exceptions/blitz_nullfunctionexception.h"
#include "c/exceptions/blitz_arrayboundsexception.h"
#include "c/exceptions/blitz_outofdataexception.h"
#include "c/exceptions/blitz_runtimeexception.h"

BB_BEGIN_DECLS

#ifdef BB_ENABLE_HELPER

/** Checks if a pointer is NULL
 *
 * The BlitzMax runtime has three compile time objects to represents
 * NULL instances of their specific type (Object, String, Array).
 *
 * Checking for those is performed via GCC builtin functionality.
 * If the given parameter \a x is neither of the three above mentioned,
 * this returns if \a x is the C \c NULL pointer.
 *
 * \see		BBNULLOBJECT, BBNULLSTRING, BBNULLARRAY
 *
 * \param	ptr A pointer to test
 * \retval	true If the pointer is considered to be NULL
 * \retval	false If the pointer is referencing to a valid value
 */
#define BB_ISNULL(ptr) \
	({																			\
		typeof (ptr) _x = (ptr);												\
		__builtin_choose_expr (													\
			__builtin_types_compatible_p (typeof (_x), BBString*),				\
			_x == BBNULLSTRING,													\
			__builtin_choose_expr (												\
				__builtin_types_compatible_p (typeof (_x), BBArray*)			\
				|| __builtin_types_compatible_p (typeof (_x), BBByteArray*)		\
				|| __builtin_types_compatible_p (typeof (_x), BBShortArray*)	\
				|| __builtin_types_compatible_p (typeof (_x), BBIntArray*)		\
				|| __builtin_types_compatible_p (typeof (_x), BBFloatArray*)	\
				|| __builtin_types_compatible_p (typeof (_x), BBDoubleArray*)	\
				|| __builtin_types_compatible_p (typeof (_x), BBStringArray*),	\
				_x == BBNULLARRAY,												\
				__builtin_choose_expr (											\
					__builtin_types_compatible_p (typeof (_x), BBObject*),		\
					_x == BBNULLOBJECT,											\
					_x == NULL													\
				)																\
			)																	\
		);																		\
	})

/** Checks if two objects are equal
 *
 * \todo	Won't work with primitive data types yet.
 *
 * \param	obj The object which provides the Compare() function
 * \param	rhs The right-hand side object which is compared with the first one
 * \retval	true If both instances are considered to be equal
 * \retval	false If the instances differ in some way
 */
#define BB_EQUALS(obj, rhs) \
	({																			\
		typeof (obj) _x = (obj); typeof (rhs) _y = (rhs);						\
		__builtin_choose_expr (													\
			__builtin_types_compatible_p (typeof (_x), BBString*)				\
			&& __builtin_types_compatible_p (typeof (_y), BBString*),			\
			bbStringEquals(_x, _y),												\
			__builtin_choose_expr (												\
				__builtin_types_compatible_p (typeof (_x), BBArray*)			\
				&& __builtin_types_compatible_p (typeof (_y), BBArray*),		\
				bbArrayEquals(_x, _y),											\
				__builtin_choose_expr (											\
					__builtin_types_compatible_p (typeof (_x), BBObject*)		\
					&& __builtin_types_compatible_p (typeof (_x), BBObject*),	\
					_x->clas->Compare(_x, _y) == 0,								\
					false														\
				)																\
			)																	\
		);																		\
	})

/** Throws a BlitzMax exception
 *
 * \param	cstring_or_obj A string or object describing the problem
 */
#define BB_THROW(cstring_or_obj) \
	({																			\
		typeof (cstring_or_obj) _x = (cstring_or_obj);							\
		__builtin_choose_expr (													\
			__builtin_types_compatible_p (typeof (_x), const char*)				\
			|| __builtin_types_compatible_p (typeof (_x), char*)				\
			|| __builtin_types_compatible_p (typeof (_x), char[]),				\
			bbExThrowCString(_x),												\
			bbExThrow(_x)														\
		);																		\
	})

#endif	// BB_ENABLE_HELPER


BB_END_DECLS

#endif	// BLITZ_H
