#include "_common.h"
#include "../blitz.h"

BBString *bbAppTitle	= BBNULLSTRING;
BBString *bbAppDir		= BBNULLSTRING;
BBString *bbAppFile		= BBNULLSTRING;
BBString *bbLaunchDir	= BBNULLSTRING;
BBArray  *bbAppArgs		= BBNULLARRAY;

void **bbGCStackTop;

void bbEnd()
{
	exit(EXIT_SUCCESS);
}

void bbOnEnd(void (*callback)())
{
	atexit(callback);
}

void bbWriteStdout(const BBString *str)
{
	char *buf = bbStringToCString(str);
	fprintf(stdout, "%s", buf);
	fflush(stdout);
	bbMemFree(buf);
}

void bbWriteStderr(const BBString *str)
{
	char *buf = bbStringToCString(str);
	fprintf(stderr, "%s", buf);
	fflush(stderr);
	bbMemFree(buf);
}

BBString* bbReadStdin()
{
#	define INBUF_SIZE		256
	BBSize size = 0;
	char *c_str = 0;

	char in_buf[INBUF_SIZE];
	size_t in_size;
	
	char *chr_pos;
	
	for (BBSize i = 0; i < BBSTRING_MAX_LENGTH; ++i)
	{
		in_buf[0] = 0;
		fgets(in_buf, INBUF_SIZE, stdin);
		in_buf[INBUF_SIZE - 1] = 0;
		
		if (chr_pos = strchr(in_buf, '\n')) {
			in_size = chr_pos - in_buf;
			if (in_size && isspace(in_buf[in_size - 1])) {
				--in_size;
			}
		} else {
			in_size = strlen(in_buf);
		}
		
		c_str = (char *)bbMemExtend(c_str, size, size + in_size);
		memcpy(c_str + size, in_buf, in_size);
		
		size += in_size;
		if (in_size < (INBUF_SIZE - 1)) {
			break;
		}
	}
	
	BBString *str;
	
	if (size) {
		str = bbStringFromBytes(c_str, size);
	} else {
		str = &bbEmptyString;
	}
	
	bbMemFree(c_str);
	return str;
}

#if defined BB_OS_MACOS

#	include <CoreServices/CoreServices.h>

#	include <limits.h>
#	include <unistd.h>
#	include <pthread.h>

static pthread_t MainThread;

static void startup()
{
	MainThread = pthread_self();
}

void bbDelay(BBInt millisecs)
{
	if (millisecs < 0) {
		return;
	}
	
	usleep(millisecs * 1000);
}

BBInt bbMilliSecs()
{
	UnsignedWide uw;
	Microseconds(&uw);
	
	double time = (uw.hi << (32 - 9)) | (uw.lo >> 9);	// divide by 512...!
	
	return (BBInt) (long long) (time / (1000.0 / 512.0));
}

BBBool bbIsMainThread()
{
	return pthread_self() == MainThread;
}

#elif defined BB_OS_WIN32

#	include <direct.h>

int _bbusew;	// internal 'use unicode' flag

static DWORD MainThread;

static void startup()
{
	MainThread = GetCurrentThreadId();
}

void bbDelay(BBInt millisecs)
{
	if (millisecs < 0) {
		return;
	}
	
	Sleep(millisecs);
}

BBInt bbMilliSecs()
{
	return timeGetTime();
}

BBBool bbIsMainThread()
{
	return GetCurrentThreadId() == MainThread;
}

#elif defined BB_OS_LINUX

#	include <unistd.h>
#	include <pthread.h>
#	include <limits.h>
#	include <signal.h>
#	include <sys/time.h>
#	include <sys/sysinfo.h>

static int BaseTime;
static pthread_t MainThread;

static void startup()
{
	struct sysinfo info;
	
	MainThread = pthread_self();

	sysinfo(&info);
	BaseTime = bbMilliSecs() - info.uptime * 1000;
}

//***** ThreadSafe! *****
void bbDelay(BBInt millisecs)
{
	if (millisecs < 0) {
		return;
	}
		
	int end_time = bbMilliSecs() + millisecs;
	
	for (int i = 0 ;; ++i)
	{
		int delta = end_time - bbMilliSecs();
		
		if (delta <= 0) {
			if (!i) {
				usleep(0);	// sleep at least once
			}			
			break;
		}
		
		usleep(delta * 1000);
	}
}

//***** ThreadSafe! *****
BBInt bbMilliSecs()
{
	struct timeval tv;
	gettimeofday(&tv, 0);
	
	int time = tv.tv_sec * 1000;
	time += tv.tv_usec / 1000;
	return time - BaseTime;
}

BBBool bbIsMainThread()
{
	return pthread_self() == MainThread;
}

#endif	// BB_OS_LINUX

void bbStartup(int argc, char *argv[], void *dummy1, void *dummy2)
{
#if BB_OS_LINUX
	
	char *ebp;
	char buf[PATH_MAX];
	char lnk[PATH_MAX];
	
	asm("movl %%ebp,%0;":"=r"(ebp));
	
	bbGCStackTop = ebp + 28;
	
	bbThreadStartup();
	bbGCStartup();
	
	getcwd(buf, PATH_MAX);
	bbLaunchDir = bbStringFromUTF8String(buf);
	
	pid_t pid = getpid();
	sprintf(lnk, "/proc/%i/exe", pid);
	
	ssize_t buf_len = readlink(lnk, buf, PATH_MAX);
	if (buf_len > 0)
	{
		buf[buf_len] = 0;
		bbAppFile = bbStringFromUTF8String(buf);
		
		char *pos = strrchr(buf, '/');
		
		if (pos) {
			*pos = 0;
			bbAppDir = bbStringFromUTF8String(buf);
		} else {
			bbAppDir = &bbEmptyString;
		}
	} else {
		bbAppFile = &bbEmptyString;
		bbAppDir  = &bbEmptyString;
	}
	
	chdir(bbTmpUTF8String(bbAppDir));	

#elif defined BB_OS_MACOS
	
#	ifdef BB_ARGP
	bbGCStackTop = bbArgp(0);
#	else
	bbGCStackTop = &argc;
#	endif
	
	bbThreadStartup();
	bbGCStartup();
	
	char buf[PATH_MAX];
	getcwd(buf, PATH_MAX);
	bbLaunchDir = bbStringFromUTF8String(buf);
	
	CFURLRef url = CFBundleCopyExecutableURL(CFBundleGetMainBundle());
	CFURLGetFileSystemRepresentation(url, true, (UInt8 *)buf, PATH_MAX);
	CFRelease(url);
	bbAppFile = bbStringFromUTF8String(buf);
	
	char *pos;
	
	if (pos = strstr(buf, ".app/Contents/MacOS/")) {
		*pos = 0;
	}
	
	if (pos = strrchr(buf, '/')) {
		*pos = 0;
		bbAppDir = bbStringFromUTF8String(buf);
		chdir(buf);
	} else {
		bbAppDir = &bbEmptyString;
	}
	
#elif defined BB_OS_WIN32

	char *ebp;
	OSVERSIONINFO os = { sizeof (os) };
	
	asm("movl %%ebp,%0;":"=r"(ebp));
	
	bbGCStackTop = ebp + 28;
	
	bbThreadStartup();
	bbGCStartup();

	if (GetVersionEx(&os)) {
		if (os.dwPlatformId == VER_PLATFORM_WIN32_NT) {
			_bbusew = 1;
		}
	}
	
	int i;
	
	if (_bbusew) {
		wchar_t buf[MAX_PATH];
		
		_wgetcwd(buf, MAX_PATH);
		for (i = 0; buf[i]; ++i)
		{
			if (buf[i] == '\\') {
				buf[i] = '/';
			}
		}
		
		bbLaunchDir = bbStringFromWString(buf);
		
		GetModuleFileNameW(GetModuleHandleW(0), buf, MAX_PATH);
		int last_separator_pos = 0;
		
		for (i = 0; buf[i]; ++i)
		{
			if (buf[i] == '\\') {
				buf[i] = '/';
			}
			if (buf[i] == '/') {
				last_separator_pos = i;
			}
		}
		
		bbAppFile = bbStringFromWString(buf);

		if (last_separator_pos) {
			if (buf[last_separator_pos - 1] == ':') {
				++last_separator_pos;
			}
			bbAppDir = bbStringFromShorts(buf, last_separator_pos);
		} else {
			bbAppDir = &bbEmptyString;
		}

		_wchdir(bbTmpWString(bbAppDir));
		
	} else {
		char buf[MAX_PATH];
		_getcwd(buf, MAX_PATH);
		
		for (i = 0; buf[i]; ++i)
		{
			if (buf[i] == '\\') {
				buf[i] = '/';
			}
		}
		
		bbLaunchDir = bbStringFromCString(buf);
		
		GetModuleFileNameA(GetModuleHandleA(0), buf, MAX_PATH);
		
		size_t last_separator_pos = 0;
		
		for (i = 0; buf[i]; ++i)
		{
			if (buf[i] == '\\') {
				buf[i] = '/';
			}
			if (buf[i] == '/') {
				last_separator_pos = i;
			}
		}
		
		bbAppFile = bbStringFromCString(buf);

		if (last_separator_pos) {
			if (buf[last_separator_pos - 1] == ':') {
				++last_separator_pos;
			}
			bbAppDir = bbStringFromBytes(buf, last_separator_pos);
		} else {
			bbAppDir = &bbEmptyString;
		}

		_chdir(bbTmpCString(bbAppDir));
	}
	
#endif

	BB_INCREFS(bbLaunchDir);
	BB_INCREFS(bbAppDir);
	BB_INCREFS(bbAppFile);	
	
	const BBChar default_apptitle[] = {
		'B','l','i','t','z','M','a','x',' ',
		'A','p','p','l','i','c','a','t','i','o','n'
	};
	
	bbAppTitle = BBSTRING_FROMBUFFER(default_apptitle, 20);
	BB_INCREFS(bbAppTitle);
	
	bbAppArgs = bb_array_alloc1d("$", argc);
	BB_INCREFS(bbAppArgs);

	BBString **arguments = (BBString **)BBARRAY_DATAPTR_CALC(bbAppArgs, 1);
	
	for (unsigned k = 0; k < argc; ++k)
	{
		size_t arg_len = strlen(argv[k]);	// todo: Can an argument be empty?
		
		if (arg_len) {
			BBString *arg = bbStringFromBytes(argv[k], arg_len);
			BB_INCREFS(arg);
			*arguments++ = arg;
		} else {
			*arguments++ = BBNULLSTRING;
		}
	}
	
	//	Register exceptions implemented in C

	bb_class_register_multiple(
		7, /* Number of classes to register */
		&bbExceptionClass,
		&bbBlitzExceptionClass,
		&bbArrayBoundsExceptionClass,
		&bbNullFunctionExceptionClass,
		&bbNullMethodExceptionClass,
		&bbNullObjectExceptionClass,
		&bbRuntimeExceptionClass
	);
	
	startup();
}

void bbLibStartup()
{
	startup();
}
