/**
 * \file	blitz_app.h
 * \brief	Runtime environment setup and basic application flow functions
 *
 * \author	Mark Sibly
 * \author	Fabian Niemann
 */

#ifndef BLITZ_APP_H
#define BLITZ_APP_H

#include "_common.h"

BB_BEGIN_DECLS

/** Contains the absolute path to the executable file's directory
 *
 * The path value is automatically filled in at startup.
 *
 * \see		bbAppFile
 */
extern BBString* bbAppDir;

/** Contains the absolute path to the executable file
 *
 * The path value is automatically filled in at startup.
 *
 * \see		bbAppDir
 */
extern BBString* bbAppFile;

/** Contains the application's title
 *
 * This global string variable has the default value "BlitzMax Application".
 * It can be changed to the name of the application and is used in a number
 * of ways throughout BlitzMax. One of which is the title of the application's
 * main window. But third-party code might depend on it, so try not to make
 * it too long or there could be problems when displaying it.
 */
extern BBString* bbAppTitle;

/** Contains the directory from which the application was launched
 *
 * The path value is automatically filled in at startup.
 */
extern BBString* bbLaunchDir;

/** Contains the application's arguments
 *
 * All command line arguments passed to the application are stored
 * in this global array as strings.
 *
 * \note	If you change the values, consider that other parts of the
 *			code might depend on it. If you are unsure, simply copy them.
 */
extern BBArray*  bbAppArgs;

extern void** bbGCStackTop;

/** Exits the application
 * 
 * Ending the application's execution with this is considered a normal termination,
 * meaning a termination without an error.
 *
 * On execution all callbacks registered with bbOnEnd are executed in reverse
 * order of their registration. Then after a low-level clean-up, control is returned
 * to the operating system.
 */
void bbEnd();

/** Registers a callback function to be executed before exiting the application
 *
 * All functions are called in reverse order of their registration.
 * At least 32 callbacks are supported by the system.
 *
 * \param	callback The function to be called on exiting the application
 */
void bbOnEnd(void (*callback)());

/** Reads a string from the standard input stream
 *
 * This reads as many characters as are needed until a linebreak appears.
 *
 * \note	Usually the stdin stream comes from keyboard input
 * \return	The line of text that the user did input
 */
BBString* bbReadStdin();

/** Writes a string to the standard output stream
 *
 * After writing the message to the standard output stream (stdout),
 * that stream is immediately flushed. That means the results are handed
 * to the system right away, without waiting for the internal buffer to be filled.
 *
 * \note	Usually the stdio stream is the console.
 * \param	str The message to be written to the standard output stream.
 */
void bbWriteStdout(const BBString *str);

/** Writes a string to the standard error stream
 *
 * After writing the message to the standard error stream (stderr),
 * that stream is immediately flushed. That means the results are handed
 * to the system right away, without waiting for the internal buffer to be filled.
 *
 * \note	Usually the stderr stream is the console
 * \param	str The message to be written to the standard error stream
 */
void bbWriteStderr(const BBString *str);

/** Delays the application's execution for a given time
 *
 * The execution is suspended for at least the given amount of milliseconds.
 * A millisecond is one thousandth of a second. Therefore a delay of 1000
 * milliseconds would halt execution for (nearly) 1 second.
 *
 * \note	If you pass a negative value as a parameter, this function does nothing.
 *
 * \param	millisecs The number of milliseconds that the execution should be delayed.
 */
void bbDelay(BBInt millisecs);

/** Gives the number of milliseconds since startup
 *
 * The number of milliseconds since the system was turned on is often used
 * to spice up pseudo-random numbers, for example in games.
 *
 * \return The number of milliseconds since the system's startup 
 */
BBInt bbMilliSecs();

/** Checks if the current thread is the main one.
 *
 * Some functionality only works if executed in the main thread.
 * Using this function you can make sure that everything is alright.
 *
 * \retval	BBTRUE if called from the current main thread
 * \retval	BBFALSE if called from any other thread
 */
BBBool bbIsMainThread();

/* Starts the GC and threading environment. Also initializes
 * the global application variables:
 *
 * - bbAppDir
 * - bbAppFile
 * - bbAppTitle
 * - bbLaunchDir
 * - bbAppArgs
 */
void bbStartup(int argc, char *argv[], void *dummy1, void *dummy2);

BB_END_DECLS

#endif	// BLITZ_APP_H
