#include "blitz_array.h"
#include "blitz_string.h"

typedef uint_fast32_t	strpos_t;
typedef uint_fast32_t	arrpos_t;

static void array_free(BBArray *arr);
static bool array_sub_equals(const BBArray * restrict arr, const BBArray * restrict rhs,
							 BBSize start_pos, BBSize len)
							__attribute__ ((pure, always_inline, nonnull(1, 2), optimize(3)));

static BBDebugScope ClassScope =
{
	.kind			= BBDEBUGSCOPE_USERTYPE,
	.name			= "Array",
	.decls			= {
		BBDEBUGDECL_END
	}
};

BBClass bbArrayClass =
{
	.super			= &bbObjectClass,
	.free			= array_free,
	.debug_scope	= &ClassScope,
	.instance_size	= 0,
	
	.ctor			= 0,
	.dtor			= 0,
	
	.ToString		= bbArrayToString,
	.Compare		= bbArrayCompare,
	.SendMessage	= bbObjectSendMessage,
	.Clone			= bbArrayClone,
	
	._reserved1_	= bbObjectReserved,
	._reserved2_	= bbObjectReserved,
	
	.vfns			= {
		bbArraySort,
		bbArrayDimensions
	}
};

BBArray bbEmptyArray =
{
	.clas		= &bbArrayClass,
	.refs		= BBGC_MANYREFS,
	
	.type		= "",
	.dims		= 0,
	.size		= 0,
	.scales[0]	= 0
};

// Only used by ref counting GC.
static void array_free(BBArray *arr)
{
#ifdef BB_GC_RC
	if (arr == &bbEmptyArray) {
		arr->refs = BBGC_MANYREFS;
		return;
	}
	
	//	complex types must be dereferenced in ref-counting GC
	if (arr->type[0] == ':' || arr->type[0] == '$' || arr->type[0] == '[')
	{
		BBObject **data = (BBObject **)BBARRAY_DATAPTR(arr);
		BBObject *obj;
		
		for (BBSize i = arr->scales[0]; i > 0; --i) {
			obj = *data++;
			BB_DECREFS(obj);
		}
	}
	
	bbGCDeallocObject(arr, BBARRAY_SIZE(arr));
#endif
}

//	contains the GC allocation flags indexed by BBTypeKey
static const int ArrayAllocFlags[] =
{
	[BBTYPEKEY_VOID]	= BBGC_ATOMIC,
	[BBTYPEKEY_BYTE]	= BBGC_ATOMIC,
	[BBTYPEKEY_SHORT]	= BBGC_ATOMIC,
	[BBTYPEKEY_INT]		= BBGC_ATOMIC,
	[BBTYPEKEY_LONG]	= BBGC_ATOMIC,
	[BBTYPEKEY_FLOAT]	= BBGC_ATOMIC,
	[BBTYPEKEY_DOUBLE]	= BBGC_ATOMIC,
	[BBTYPEKEY_POINTER]	= BBGC_ATOMIC,
	
	[BBTYPEKEY_OBJECT]	= 0,
	[BBTYPEKEY_STRING]	= 0,
	[BBTYPEKEY_ARRAY]	= 0,
	[BBTYPEKEY_FUNCTION]= BBGC_ATOMIC
};

BBArray* bb_array_alloc(type, dims, lens)
 const char *type;
 const BBSize dims;
 const BBSize *lens;
{
	BBSize length = 1;
	BBInt i;
	
	for (i = 0; i < dims; ++i)
	{
		if (lens[i] < 1) {
			return &bbEmptyArray;
		}
		length *= lens[i];
	}
	
	const BBSize size = BBTYPETAG_SIZE(type) * length;

	BBArray *arr = BBARRAY_ALLOC(size, dims, ArrayAllocFlags[BBTYPETAG_KEY(type)]);
	
	arr->type = type;
	arr->dims = dims;
	arr->size = size;

	for (i = 0; i < dims; ++i) {
		arr->scales[i] = lens[i];
	}
	
	for (i = dims - 2; i >= 0; --i) {
		arr->scales[i] *= arr->scales[i + 1];
	}

	return arr;
}

BBArray* bb_array_alloc1d(type, len)
 const char *type;
 const BBSize len;
{
	const BBSize size = len * BBTYPETAG_SIZE(type);
	BBArray *arr = BBARRAY_ALLOC(size, 1, ArrayAllocFlags[BBTYPETAG_KEY(type)]);
	
	arr->type = type;
	arr->dims = 1;
	arr->size = size;
	arr->scales[0] = len;
	
	return arr;
}

void bb_array_init(arr)
 BBArray *arr;
{
	if (arr->scales[0] < 1) {
		return;
	}
	
	void **arr_data  = BBARRAY_DATAPTR(arr);
	const void *init = BBTYPETAG_INIT(arr->type);
	
	if (init) {
		for (BBSize i = arr->scales[0]; i > 0; --i) {
			*arr_data++ = init;
		}
	} else {
		memset(arr_data, 0, arr->size);
	}
}

static inline bool array_sub_equals(arr, rhs, start_pos, cmp_len)
 const BBArray * restrict arr;
 const BBArray * restrict rhs;
 BBSize start_pos;
 const BBSize cmp_len;
{
	const BBSize arr_len = arr->scales[0];
	const BBSize rhs_len = rhs->scales[0];
	
	if (arr_len < rhs_len ||
		start_pos + cmp_len > arr_len ||
		cmp_len > rhs_len) {
		return false;
	}
	
	const BBTypeKey arr_tkey = BBTYPETAG_KEY(arr->type);
	const BBTypeKey rhs_tkey = BBTYPETAG_KEY(rhs->type);
	
	arrpos_t pos;
	
	//	Only compare arrays of equal type
	if (arr_tkey == rhs_tkey)
	{
		if (BBTYPEKEY_ISPRIMITIVE(arr_tkey)) {
			return !memcmp(BBARRAY_DATAPTR(arr), BBARRAY_DATAPTR(rhs),
						   rhs_len  * BBTYPEKEY_SIZE(arr_tkey));
		} else if (arr_tkey == BBTYPEKEY_STRING)
		{
			const BBString **arr_data = ((BBStringArray *)arr)->data;
			const BBString **rhs_data = ((BBStringArray *)rhs)->data;
			
			for (pos = 0; pos < cmp_len; ++pos)
			{
				if (!bbStringEquals(arr_data[start_pos + pos], rhs_data[pos])) {
					return false;
				}
			}
		} else if (arr_tkey == BBTYPEKEY_OBJECT)
		{
			const BBObject **arr_data = ((BBObjectArray *)arr)->data;
			const BBObject **rhs_data = ((BBObjectArray *)rhs)->data;
			
			arr_data += start_pos;
			
			BBClassCompareFunc cmp_func;
			
			for (pos = 0; pos < cmp_len; ++pos)
			{
				cmp_func = arr_data[pos]->clas->Compare;
				
				if (cmp_func(arr_data[pos], rhs_data[pos])) {
					return false;
				}
			}
		} else if (arr_tkey == BBTYPEKEY_ARRAY)
		{
			const BBArray **arr_data = ((BBArrayArray *)arr)->data;
			const BBArray **rhs_data = ((BBArrayArray *)rhs)->data;
			
			arr_data += start_pos;
			
			for (pos = 0; pos < cmp_len; ++pos)
			{
				if (!bbArrayEquals(arr_data[pos], rhs_data[pos])) {
					return false;
				}
			}
		}
	}
	
	return true;
}

//	API

BBInt bbArrayCompare(const BBArray *arr, const BBArray *rhs)
{
	if (arr == rhs) {
		return 0;
	}
	
	if (arr->scales[0] != rhs->scales[0]) {
		return arr->scales[0] > rhs->scales[0] ? 1 : -1;
	} else if (arr->dims != rhs->dims) {
		return arr->dims > rhs->dims ? 1 : -1;
	} else if (arr->size != rhs->size) {
		return arr->size > rhs->size ? 1 : -1;
	}
	
	const BBTypeKey arr_typekey = BBTYPETAG_KEY(arr->type);
	const BBTypeKey rhs_typekey = BBTYPETAG_KEY(rhs->type);
	
	if (arr_typekey != rhs_typekey) {
		return arr_typekey > rhs_typekey ? 1 : -1;
	}
	
	int cmp = memcmp(BBARRAY_DATAPTR(arr), BBARRAY_DATAPTR(rhs), arr->size);
	if (cmp) {
		return cmp > 0 ? 1 : -1;
	}
	
	return 0;
}

BBArray *bbArrayNew(const char *type, BBInt dims, ...)
{
#if BB_ARGP
	BBInt *lens = (BBInt *)bbArgp(8);
#else
	BBInt *lens = &dims + 1;
#endif

	BB_ASSERT(lens[0] >= 0, "Array length cannot be negative");
	BB_ASSERT(lens[0] < (BBARRAY_MAX_LENGTH / BBTYPETAG_SIZE(type)), "Array is too long");		
	
	BBArray *arr = bb_array_alloc(type, dims, lens);
	bb_array_init(arr);
	
	return arr;
}
 
BBArray* bbArrayNewEx(const char *type, BBInt dims, BBInt *lens)
{
	BB_ASSERT(lens[0] >= 0, "Array length cannot be negative");
	BB_ASSERT(lens[0] < (BBARRAY_MAX_LENGTH / BBTYPETAG_SIZE(type)), "Array is too long");	
	
	BBArray *arr = bb_array_alloc(type, dims, lens);
	bb_array_init(arr);
	
	return arr;
}

//	manually inlined, since this is used most often
BBArray* bbArrayNew1D(const char *type, BBInt len)
{
	const BBTypeKey tkey = BBTYPETAG_KEY(type);
	const BBSize size = len * BBTYPEKEY_SIZE(tkey);

	BB_ASSERT(len >= 0, "Array length cannot be negative");
	BB_ASSERT(len < (BBARRAY_MAX_LENGTH / BBTYPEKEY_SIZE(tkey)), "Array is too long");
	
	BBArray *arr = BBARRAY_ALLOC(size, 1, ArrayAllocFlags[tkey]);
	
	arr->type = type;
	arr->dims = 1;
	arr->size = size;
	arr->scales[0] = len;
	
	void **arr_data  = BBARRAY_DATAPTR(arr);
	const void *init = BBTYPEKEY_INIT(tkey);
	
	if (init) {
		for (BBInt i = len; i > 0; --i) {
			*arr_data++ = init;
		}
	} else {
		memset(arr_data, 0, size);
	}
	
	return arr;
}

BBArray* bbArraySlice(const char *type, const BBArray *in_arr, BBInt beg, BBInt end)
{
	const int length = end - beg;
	if (length < 1) {
		return &bbEmptyArray;
	}

	BBArray *out_arr = bb_array_alloc1d(type, length);

	const BBSize el_size = out_arr->size / length;
	
	const void * const initializer = BBTYPETAG_INIT(out_arr->type);
	char *data_ptr = (char *)BBARRAY_DATAPTR_CALC(out_arr, 1);

	int k;
	
	int n = -beg;
	if (n > 0) {
		if (beg + n > end) {
			n = end - beg;
		}
		
		if (initializer) {
			void **dst = (void **)data_ptr;
			for (k = 0; k < n; ++k) {
				*dst++ = initializer;
			}
			
			data_ptr = (char*)dst;
		} else {
			memset(data_ptr, 0, n * el_size);
			data_ptr += n * el_size;
		}
		
		beg += n;
		if (beg == end) {
			return out_arr;
		}
	}
	
	n = in_arr->scales[0] - beg;
	if (n > 0) {
		if (beg + n > end) {
			n = end - beg;
		}
#ifdef BB_GC_RC
		if (type[0] == ':' || type[0] == '$' || type[0] == '[')
		{
			BBObject **dst = (BBObject **)data_ptr;
			BBObject **src = (BBObject **)BBARRAY_DATAPTR(in_arr) + beg;
			
			for (k = 0; k < n; ++k) { 
				BBObject *o = *src++;
				BB_INCREFS(o);
				*dst++ = o; 
			}
			
			data_ptr = (char *)dst;
		} else {
			memcpy(data_ptr, (char *)BBARRAY_DATAPTR(in_arr) + beg * el_size, n * el_size);
			data_ptr += n * el_size;
		}
#else
		memcpy(data_ptr, (char *)BBARRAY_DATAPTR(in_arr) + beg * el_size, n * el_size);
		data_ptr += n * el_size;
#endif
		beg += n;
		if (beg == end) {
			return out_arr;
		}
	}
	
	n = end - beg;
	if (n > 0) {
		if (initializer) {
			void **dst = (void **)data_ptr;
			for (k = 0; k < n; ++k) {
				*dst++ = initializer;
			}
		} else {
			memset(data_ptr, 0, n * el_size);
		}
	}
	
	return out_arr;
}

BBArray* bbArrayConcat(const char *type, const BBArray *x, const BBArray *y)
{
	const BBSize out_len = x->scales[0] + y->scales[0];
	
	if (out_len < 1) {
		return &bbEmptyArray;
	}
	
	BBArray *out = bb_array_alloc1d(type, out_len);
	char *data = (char *)BBARRAY_DATAPTR_CALC(out, 1);
	
	memcpy(data, BBARRAY_DATAPTR_CALC(x, 1), x->size);
	memcpy(data + x->size, BBARRAY_DATAPTR_CALC(y, 1), y->size);
	
#ifdef BB_GC_RC
	//	Increasing the reference count of all complex data type elements
	//	is necessary to let the GC know that the new array is using them.
	
	if (type[0] == ':' || type[0] == '$' || type[0] == '[')
	{
		BBObject **objects = (BBObject **)data;
		
		for (arrpos_t i = 0; i < out_len; ++i) {
			BB_INCREFS(objects[i]);
		}
	}
#endif
	
	return out;
}

BBArray* bbArrayFromData(const char *type, BBInt length, const void *data)
{
	if (length < 1) {
		return &bbEmptyArray;
	}
	
	BBArray *arr = bb_array_alloc1d(type, length);

	if (type[0] == 'b') {
		BBByte *arr_data = BBARRAY_DATAPTR_CALC(arr, 1);
		for (arrpos_t k = 0; k < length; ++k) {
			arr_data[k] = ((int *)data)[k];
		}
	} else if (type[0] == 's') {
		BBShort *arr_data = BBARRAY_DATAPTR_CALC(arr, 1);
		for (arrpos_t k = 0; k < length; ++k) {
			arr_data[k] = ((int *)data)[k];
		}
	} else {
		memcpy(BBARRAY_DATAPTR_CALC(arr, 1), data, arr->size);
	}
 
	return arr;
}

BBArray* bbArrayDimensions(const BBArray *arr)
{
	if (arr->scales[0] < 1) {
		return &bbEmptyArray;
	}
		
	BBSize num_dims = arr->dims;
	
	BBArray *out = bb_array_alloc1d("i", num_dims);
	BBInt *out_data = (BBInt *)BBARRAY_DATAPTR_CALC(out, 1);

	--num_dims;
	
	for (arrpos_t i = 0; i < num_dims; ++i) {
		*++out_data = arr->scales[i] / arr->scales[i+1];
	}
	*++out_data = arr->scales[num_dims - 1];
	
	return out;
}

BBArray* bbArrayCastFromObject(const BBObject *obj, const char *typetag)
{
	const BBArray *arr = (BBArray *)obj;
	
	if (arr == &bbEmptyArray) {
		return arr;
	}
	
	if (arr->clas != &bbArrayClass) {
		return (BBArray *)BBNULLOBJECT;
	}
	
	if (arr->type[0] == ':' && typetag[0] == ':') {
		return arr;
	}
	
	if (strcmp(arr->type, typetag)) {
		return (BBArray *)BBNULLOBJECT;
	}
	
	return arr;
}

#define SWAP(x, y)	{ t = *(x); *(x) = *(y); *(y) = t; }
#define QSORTARRAY( TYPE,IDENT )\
static void IDENT( TYPE *lo,TYPE *hi ){\
	TYPE t;\
	TYPE *i;\
	TYPE *x;\
	TYPE *y;\
	if( hi<=lo ) return;\
	if( lo+1==hi ){\
		if( LESSTHAN(hi,lo) ) SWAP(lo,hi);\
		return;\
	}\
	i=(hi-lo)/2+lo;\
	if( LESSTHAN(i,lo) ) SWAP(i,lo);\
	if( LESSTHAN(hi,i) ){\
		SWAP(i,hi);\
		if( LESSTHAN(i,lo) ) SWAP(i,lo);\
	}\
	x=lo+1;\
	y=hi-1;\
	do{\
		while( LESSTHAN(x,i) ) ++x;\
		while( LESSTHAN(i,y) ) --y;\
		if( x>y ) break;\
		if( x<y ){\
			SWAP(x,y);\
			if( i==x ) i=y;\
			else if( i==y ) i=x;\
		}\
		++x;\
		--y;\
	}while( x<=y );\
	IDENT(lo,y);\
	IDENT(x,hi);\
}

#undef LESSTHAN
#define LESSTHAN(X,Y) (*(X) < *(Y))
QSORTARRAY(BBByte, _qsort_b)
QSORTARRAY(BBShort, qsort_s)
QSORTARRAY(BBInt, qsort_i)
QSORTARRAY(BBLong, qsort_l);
QSORTARRAY(BBFloat, qsort_f);
QSORTARRAY(BBDouble, qsort_d);
#undef LESSTHAN
#define LESSTHAN(X,Y) ((*X)->clas->Compare(*(X),*(Y)) < 0)
QSORTARRAY( BBObject*,qsort_obj );
#undef LESSTHAN
#define LESSTHAN(X,Y) (*(X)>*(Y))
QSORTARRAY(BBByte, qsort_b_d)
QSORTARRAY(BBShort, qsort_s_d)
QSORTARRAY(BBInt, qsort_i_d)
QSORTARRAY(BBLong, qsort_l_d);
QSORTARRAY(BBFloat, qsort_f_d);
QSORTARRAY(BBDouble, qsort_d_d);
#undef LESSTHAN
#define LESSTHAN(X,Y) ((*X)->clas->Compare(*(X),*(Y)) > 0)
QSORTARRAY( BBObject*,qsort_obj_d );

void bbArraySort(BBArray *arr, BBInt ascending)
{
	int n = arr->scales[0] - 1;
	if (n <= 0) {
		return;
	}
	
	void *p = BBARRAY_DATAPTR(arr);
	
	if (ascending)
	{
		switch (arr->type[0])
		{
		case 'b':
			_qsort_b((BBByte *)p, (BBByte *)p + n);
			break;
		case 's':
			qsort_s((BBShort *)p, (BBShort *)p + n);
			break;
		case 'i':
			qsort_i((BBInt *)p, (BBInt *)p + n);
			break;
		case 'l':
			qsort_l((BBLong *)p, (BBLong *)p + n);
			break;
		case 'f':
			qsort_f((BBFloat *)p, (BBFloat *)p + n );
			break;
		case 'd':
			qsort_d((BBDouble *)p, (BBDouble *)p + n);
			break;
		case '$': case ':':
			qsort_obj((BBObject **)p, (BBObject **)p + n);
			break;
		}
	} else
	{
		switch (arr->type[0])
		{
		case 'b':
			qsort_b_d((BBByte *)p, (BBByte *)p + n);
			break;
		case 's':
			qsort_s_d((BBShort *)p, (BBShort *)p + n);
			break;
		case 'i':
			qsort_i_d((BBInt *)p, (BBInt *)p + n);
			break;
		case 'l':
			qsort_l_d((BBLong *)p,(BBLong *)p + n);
			break;
		case 'f':
			qsort_f_d((BBFloat *)p,(BBFloat *)p + n);
			break;
		case 'd':
			qsort_d_d((BBDouble *)p, (BBDouble *)p + n);
			break;
		case '$': case ':':
			qsort_obj_d((BBObject **)p, (BBObject **)p + n);
			break;
		}
	}
}

//	Additions by BME

static const BBChar EmptyArrayMessageBuf[] = {
	'(','A','r','r','a','y',')',
	' ','N','u','l','l'
};

#define TOSTRING_NUMBUF_SIZE	32

//	still a bit buggy .. aparently the array properties are not
//	changed when its resized or shit .. I'm not awake :D
BBString* bbArrayToString(const BBArray *arr)
{
	if (arr == &bbEmptyArray) {
		return BBSTRING_FROMBUFFER(EmptyArrayMessageBuf, 12);
	}
	
	const BBTypeKey typekey = BBTYPETAG_KEY(arr->type);
	
	const size_t buf_len = 256 + arr->dims * TOSTRING_NUMBUF_SIZE;
	
	//char buf[buf_len];
	char *buf = malloc(buf_len);
	snprintf(buf, buf_len, "(Array) %s[", bbTypeKeyInfo[typekey].ident);
	
	char num_buf[TOSTRING_NUMBUF_SIZE];
	const BBSize num_dims = arr->dims;
	
	if (num_dims == 1) {
		snprintf(buf, buf_len, "%s%i]", buf, arr->scales[0]);
	} else {
		int d;
		for (d = 0; d < num_dims - 1; ++d) {
			snprintf(num_buf, TOSTRING_NUMBUF_SIZE, "%i,", arr->scales[d] / arr->scales[d+1]);
			strncat(buf, num_buf, buf_len);
		}
		snprintf(num_buf, TOSTRING_NUMBUF_SIZE, "%i]", arr->scales[arr->dims-1]);
		strncat(buf, num_buf, buf_len);
	}	
	
	const size_t str_len = strlen(buf);
	
	if (!str_len)
	{
		free(buf);
		return BBSTRING_FROMBUFFER(EmptyArrayMessageBuf, 12);
	}
	
	BBString *str = bbStringFromBytes(buf, str_len);
	free(buf);
	return str;
}

BBArray* bbArrayClone(const BBArray *arr)
{
	const BBSize len = arr->scales[0];
	
	if (len < 1) {
		return &bbEmptyArray;
	}
	
	BBArray *out = bb_array_alloc(arr->type, arr->dims, arr->scales);
	
	const BBTypeKey tkey = BBTYPETAG_KEY(arr->type);
	
	if (BBTYPEKEY_ISPRIMITIVE(tkey)) {
		memcpy(BBARRAY_DATAPTR(out), BBARRAY_DATAPTR(arr), len * BBTYPEKEY_SIZE(tkey));
	} else
	{		
		BBObject **out_data = ((BBObjectArray *)out)->data;
		const BBObject **arr_data = ((BBObjectArray *)arr)->data;
		
		for (arrpos_t pos = 0; pos < len; ++pos)
		{
			out_data[pos] = arr_data[pos];
			BB_RETAIN(out_data[pos]);
		}
	}
	
	return out;
}

BBBool bbArrayEquals(const BBArray *arr, const BBArray *rhs)
{
	if (arr == rhs) {
		return BBTRUE;
	}

	if (arr->scales[0] != rhs->scales[0] ||
		arr->dims != rhs->dims ||
		arr->size != rhs->size ||
		arr->type[0] != rhs->type[0]) {
		return false;
	}
	
	return array_sub_equals(arr, rhs, 0, rhs->scales[0]);
}



BBInt bbArrayFind(const BBArray *arr, const BBArray *sub, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	const BBSize sub_len = sub->scales[0];
	
	if (sub_len < 1 || sub_len > arr_len) {
		return BBARRAY_NOTFOUND;
	}

	if (start_pos < 1) {
		start_pos = 0;
	} else if (start_pos + sub_len > arr_len) {
		return BBARRAY_NOTFOUND;
	} else if (arr == sub) {
		// When the both array's are the same, the 'sub' one can only
		// be found if the start position for the search is zero
		return !start_pos;
	}	

	for (arrpos_t arr_pos = start_pos; arr_pos < arr_len; ++arr_pos)
	{
		if (array_sub_equals(arr, sub, arr_pos, sub_len)) {
			return arr_pos;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindInt(const BBArray *arr, const BBInt num, BBInt start_pos)
{
	const BBTypeKey tkey = BBTYPETAG_KEY(arr->type);	
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || tkey > BBTYPEKEY_INT) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	if (tkey == BBTYPEKEY_INT)
	{
		const BBInt *data = ((BBIntArray *)arr)->data;	
		
		for (arrpos_t k = start_pos; k < arr_len; ++k)
		{
			if (num == data[k]) {
				return k;
			}
		}
	} else
	if (tkey == BBTYPEKEY_BYTE)
	{
		const BBByte *data = ((BBByteArray *)arr)->data;
		
		for (arrpos_t k = start_pos; k < arr_len; ++k)
		{
			if (num == data[k]) {
				return k;
			}
		}
	} else // Array must be BBShortArray now
	{
		const BBShort *data = ((BBShortArray *)arr)->data;
		
		for (arrpos_t k = start_pos; k < arr_len; ++k)
		{
			if (num == data[k]) {
				return k;
			}
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindLong(const BBLongArray *arr, const BBLong num, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || BBTYPETAG_KEY(arr->type) != BBTYPEKEY_LONG) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	const BBLong *data = arr->data;
		
	for (arrpos_t k = start_pos; k < arr_len; ++k)
	{
		if (num == data[k]) {
			return k;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindFloat(const BBFloatArray *arr, const BBFloat num, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || BBTYPETAG_KEY(arr->type) != BBTYPEKEY_FLOAT) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	const BBFloat *data = arr->data;
	
	for (arrpos_t k = 0; k < arr_len; ++k)
	{
		if (num == data[k]) {
			return k;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindDouble(const BBDoubleArray *arr, BBDouble num, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || BBTYPETAG_KEY(arr->type) != BBTYPEKEY_DOUBLE) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	const BBDouble *data = arr->data;
	
	for (arrpos_t k = start_pos; k < arr_len; ++k)
	{
		if (num == data[k]) {
			return k;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindObject(const BBObjectArray *arr, const BBObject *obj, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || BBTYPETAG_KEY(arr->type) != BBTYPEKEY_OBJECT) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	const BBClassCompareFunc cmp_func = obj->clas->Compare;
	
	const BBObject **data = arr->data;
	
	for (arrpos_t k = start_pos; k < arr_len; ++k)
	{
		if (!cmp_func(obj, data[k])) {
			return k;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBInt bbArrayFindString(const BBStringArray *arr, const BBString * const str, BBInt start_pos)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 1 || BBTYPETAG_KEY(arr->type) != BBTYPEKEY_STRING) {
		return BBARRAY_NOTFOUND;
	}
	
	while (start_pos < 0) {
		start_pos += arr_len;
	}
	if (start_pos >= arr_len) {
		return BBARRAY_NOTFOUND;
	}
	
	const BBString **data = arr->data;
		
	for (arrpos_t k = start_pos; k < arr_len; ++k)
	{
		if (bbStringEquals(str, data[k])) {
			return k;
		}
	}
	
	return BBARRAY_NOTFOUND;
}

BBBool bbArrayStartsWith(const BBArray *arr, const BBArray *rhs)
{
	const BBSize rhs_len = rhs->scales[0];
	
	if (arr->scales[0] < rhs_len || arr->type[0] != rhs->type[0]) {
		return BBFALSE;
	} else if (arr == rhs) {
		return BBTRUE;
	}

	const BBTypeKey arr_tkey = BBTYPETAG_KEY(arr->type);
	
	if (arr_tkey < BBTYPEKEY_OBJECT)
	{
		return !memcmp(BBARRAY_DATAPTR(arr),
					  BBARRAY_DATAPTR(rhs),
					  rhs_len  * BBTYPEKEY_SIZE(arr_tkey)
					 );
	} else
	if (arr_tkey == BBTYPEKEY_STRING)
	{
		const BBString **arr_data = ((BBStringArray *)arr)->data;
		const BBString **rhs_data = ((BBStringArray *)rhs)->data;
		
		for (arrpos_t k = 0; k < rhs_len; ++k)
		{
			if (!bbStringEquals(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
	} else
	if (arr_tkey == BBTYPEKEY_OBJECT)
	{
		const BBObject **arr_data = ((BBObjectArray *)arr)->data;
		const BBObject **rhs_data = ((BBObjectArray *)rhs)->data;
		
		BBClassCompareFunc cmp_func;
		
		for (arrpos_t k = 0; k < rhs_len; ++k)
		{
			cmp_func = arr_data[k]->clas->Compare;
			
			if (cmp_func(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
	} else
	if (arr_tkey == BBTYPEKEY_ARRAY)
	{
		const BBArray **arr_data = ((BBArrayArray *)arr)->data;
		const BBArray **rhs_data = ((BBArrayArray *)rhs)->data;
		
		for (arrpos_t k = 0; k < rhs_len; ++k)
		{
			if (!bbArrayEquals(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
	}
	
	return BBTRUE;
}

BBBool bbArrayEndsWith(const BBArray *arr, const BBArray *rhs)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < rhs->scales[0] || arr->type[0] != rhs->type[0]) {
		return BBFALSE;
	} else if (arr == rhs) {
		return BBTRUE;
	}
	
	const BBTypeKey arr_tkey = BBTYPETAG_KEY(arr->type);
	
	if (arr_tkey < BBTYPEKEY_OBJECT)
	{
		size_t cmp_len = BBTYPEKEY_SIZE(arr_tkey) * (arr_len - rhs->scales[0]);
		
		return !memcmp(
					((char *)BBARRAY_DATAPTR(arr)) + cmp_len,
					BBARRAY_DATAPTR(rhs),
					rhs->scales[0] * BBTYPEKEY_SIZE(arr_tkey)
				);
	} else
	if (arr_tkey == BBTYPEKEY_STRING)
	{
		const BBString **arr_data = ((BBStringArray *)arr)->data;
		const BBString **rhs_data = ((BBStringArray *)rhs)->data;
		
		for (arrpos_t k = rhs->scales[0]; k > arr_len; ++k)
		{
			if (!bbStringEquals(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
	} else
	if (arr_tkey == BBTYPEKEY_OBJECT)
	{
		const BBObject **arr_data = ((BBObjectArray *)arr)->data;
		const BBObject **rhs_data = ((BBObjectArray *)rhs)->data;
		
		const BBClassCompareFunc cmp_func = arr->clas->Compare;
		
		for (arrpos_t k = rhs->scales[0]; k > arr_len; ++k)
		{
			if (cmp_func(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
	} else
	if (arr_tkey == BBTYPEKEY_ARRAY)
	{
		const BBArray **arr_data = ((BBArrayArray *)arr)->data;
		const BBArray **rhs_data = ((BBArrayArray *)rhs)->data;
		
		for (arrpos_t k = rhs->scales[0]; k > arr_len; ++k)
		{
			if (!bbArrayEquals(arr_data[k], rhs_data[k])) {
				return BBFALSE;
			}
		}
	}
	
	return BBTRUE;
}



BBBool bbArrayContains(const BBArray *arr, const BBArray *sub)
{
	return bbArrayFind(arr, sub, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsInt(const BBArray *arr, BBInt num)
{
	return bbArrayFindInt(arr, num, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsLong(const BBLongArray *arr, BBLong num)
{
	return bbArrayFindLong(arr, num, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsFloat(const BBFloatArray *arr, BBFloat num)
{
	return bbArrayFindFloat(arr, num, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsDouble(const BBDoubleArray *arr, BBDouble num)
{
	return bbArrayFindDouble(arr, num, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsObject(const BBObjectArray *arr, const BBObject *obj)
{
	return bbArrayFindObject(arr, obj, 0) != BBARRAY_NOTFOUND;
}

BBBool bbArrayContainsString(const BBStringArray *arr, const BBString *str)
{
	return bbArrayFindString(arr, str, 0) != BBARRAY_NOTFOUND;
}



BBArray* bbArrayReverse(const BBArray *arr)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 2) {
		return arr;
	}
	
	const BBTypeKey typekey = BBTYPETAG_KEY(arr->type);
	size_t element_size = BBTYPEKEY_SIZE(typekey);
	
	if (typekey == BBTYPEKEY_INT)
	{
		BBIntArray *in  = arr;
		BBIntArray *out = bb_array_alloc(arr->type, arr->dims, arr->scales);
		
		for (arrpos_t k = 0; k < arr_len; ++k) {
			out->data[k] = in->data[k];
		}
	}
	
	BBArray *out = bb_array_alloc(arr->type, arr->dims, arr->scales);
	char *out_data = BBARRAY_DATAPTR_CALC(out, 1);
	char *arr_data = BBARRAY_DATAPTR_CALC(arr, 1);
	
	for (arrpos_t k = 0; k < arr_len; ++k)
	{
		out_data[k] = arr_data[arr->scales[0] - k];
	}
	
	return out;
}

BBArray* bbArrayRotate(const BBArray *arr, BBInt count)
{
	const BBSize arr_len = arr->scales[0];
	
	if (arr_len < 2 || count == 0 || count == arr_len) {
		return arr;
	}
	
	count = -(count % arr_len);
	
	if (count < 0) {
		count += arr_len;
	}
	
	const BBTypeKey typekey		= BBTYPETAG_KEY(arr->type);
	const BBSize element_size	= BBTYPEKEY_SIZE(typekey);
	
	char *arr_data	= BBARRAY_DATAPTR(arr);
	
	BBArray *out	= bb_array_alloc(arr->type, arr->dims, arr->scales);
	char *out_data	= BBARRAY_DATAPTR(out);
	
	for (arrpos_t pos = 0; pos < arr_len; ++pos, ++count)
	{
		if (count >= arr_len) {
			count = 0;
		}
		
		memcpy(out_data + (pos * element_size),
			   arr_data + ((count) * element_size),
			   element_size);
	}
	
	return out;
}
