/**
 * \file	blitz_array.h
 * \brief	Functionality to work with array data structures
 *
 * 
 *
 * \author	Mark Sibly
 * \author	Fabian Niemann
 */

#ifndef BLITZ_ARRAY_H
#define BLITZ_ARRAY_H

#include "_common.h"

#include "blitz_class.h"
#include "blitz_object.h"

BB_BEGIN_DECLS

/** Fixed-length array object
 *
 * In BlitzMax an array is a complex data type extending BBObject.
 * That means it is usually handled by the GC and can be stored in
 * \c BBObject pointers for convenience.
 *
 * Individual elements of an array are accessed through integer indexes
 * from \c 0 up to the <tt>scales[0] - 1</tt> (zero-based indexing).
 *
 * Due to its fixed-size nature, arrays should not change size without
 * reallocating them. Although it is theoretically possible to make an
 * array shorter by changing its scales member it is not advised to do so.
 *
 * It can be assumed that every array's class is the bbArrayClass,
 * since it's declared as a "Final Type" from the side of BlitzMax.
 * Therefore every pointer of the type BBArray* can be used directly
 * with the string functions instead of looking up the right one in the
 * virtual function table (bbArrayClass.vfns).
 * 
 * \extends	BBObject
 */
struct BBArray
{
	/** The class of the array
	 * 
	 * \warning	Should always point to bbArrayClass
	 * \see		bbArrayClass
	 */
	BBClass*		clas;
	int				refs;
	
	/** The TypeTag of the data type which the array stores */
	const char*		type;

	/** The number of the array's dimensions */
	BBSize			dims;
	
	/** The size of the array's data section in bytes
	 *
	 * \note	The size of the header (the BBArray struct) is not included.
	 * \see		BBARRAY_SIZE
	 */
	BBSize			size;

	/** The number of elements each dimension can hold
	 *
	 * The first dimension of this array (<tt>scales[0]</tt>) contains the number
	 * of all elements the array holds. It can be used as the array's length.
	 */
	BBSize			scales[1];
};

#define BBARRAY_HEADER_SIZE			(sizeof (BBArray) - sizeof (BBSize))

//	Maximum value of an array's data
#define BBARRAY_MAX_DATASIZE		(BBSIZE_MAX - sizeof (BBArray))

//	Absolute maximum length (value of scales[0]) of an array
//	Depends on the data type the array stores
#define BBARRAY_MAX_LENGTH			(BBARRAY_MAX_DATASIZE)

/** Calculates the size of an array
 *
 * This can be used to calculate the size of an array object from
 * its properties. Comes in handy for all sorts of allocations.
 *
 * The amount of memory the array uses is calculated from the size
 * of "struct BBArray" (the so-called header), the size of the \a scales
 * property (which is a variable C array) and the number of dimensions.
 *
 * \see		BBARRAY_SIZE
 *
 * \param	size The value of the array's \a size property
 * \param	num_dims The number of dimensions the array contains
 * \returns	The size of the array in bytes
 */
#define BBARRAY_SIZE_CALC(size, num_dims) \
	(BBARRAY_HEADER_SIZE + (num_dims) * sizeof (int) + (size))

/** Gives the size of an array in bytes
 *
 * \see		BBARRAY_SIZE_CALC
 *
 * \param	arr A pointer to the array
 * \return	The size of the array in bytes
 */
#define BBARRAY_SIZE(arr)			BBARRAY_SIZE_CALC ((arr)->size, (arr)->dims)

/** Calculates a pointer to an array's data section
 *
 * \see		BBARRAY_DATAPTR
 *
 * \param	arr A pointer to the array
 * \param	num_dims The number of dimensions the array contains
 * \return	A pointer of type void* to the data
 */
#define BBARRAY_DATAPTR_CALC(arr, num_dims) \
	((void *)																	\
		((char *)(arr) + BBARRAY_HEADER_SIZE + (num_dims) * sizeof (int))		\
	)

/** Gives a pointer to the beginning of an array's data section
 *
 * This is a useful shorthand if the array is already initialized with the
 * correct information about its properties.
 *
 * \see		BBARRAY_DATAPTR_CALC
 *
 * \param	arr A pointer to the array
 * \returns	A pointer to the first element in the array
 */
#define BBARRAY_DATAPTR(arr)		BBARRAY_DATAPTR_CALC((arr), (arr)->dims)

/** Allocates a new BBArray object with the GC
 *
 * This is a utility macro which does not perform any
 * additional checking. It only fills in the right parameters
 * for allocating a string object with the GC.
 *
 * Only use this if you can ensure that all parameters are valid
 * and you set the array's information like size or type by hand.
 *
 * Generally it is advised to use bb_array_alloc instead.
 *
 * \note	Depending on the data type an array stores, different GC flags
 *			are required. Use with care.
 * \see		bb_array_alloc, bb_array_alloc1d, bbGCAllocObject
 *
 * \param	size The size of each dimension
 * \param	dims The number of dimensions
 * \param	flags Information for the GC on how to handle destruction
 * \return	A pointer to the newly allocated object
 */
#define BBARRAY_ALLOC(size, dims, flags) \
	({																			\
		(BBArray *)																\
		bbGCAllocObject(														\
			BBARRAY_SIZE_CALC((size), (dims)),									\
			&bbArrayClass,														\
			(flags)																\
		);																		\
	})

/** The array class
 *
 * In BlitzMax arrays are a built-in \c Final class. Therefore this structure
 * rarely needs to be accessed directly except for testing if an \c BBObject* parameter
 * is an array. All other functionality can safely be done via the API functions.
 *
 * \note	Contains only the standard BlitzMax functions to ensure maximum
 *			compatibility with legacy code.
 */
extern BBClass bbArrayClass;

/** A typesafe \c NULL pointer for arrays
 *
 * Used as a typesafe \c NULL pointer for arrays in the runtime.
 * 
 * \see		BBNULLARRAY, bbNullObject, bbEmptyString
 * \warning	Do not free this compile-time object.
 */
extern BBArray bbEmptyArray;

/**
 
 * \Alias for bbEmptyArray
 *
 * \see		BBNULLOBJECT, BBNULLSTRING
 */
#define BBNULLARRAY				(&bbEmptyArray)

/** Helper array object storing BBBytes
 *
 * \extends	BBArray
 */
struct BBByteArray
{
	BBClass			*clas;
	int				refs;
	
	const char		*type;
	BBSize			dims;
	BBSize			size;
	BBSize			scales[1];
	
	/** The array's data as a \c BBByte C-style array */
	BBByte			data[];
};

/** Helper array object storing BBShorts
 *
 * \extends	BBArray
 */
struct BBShortArray
{
	BBClass			*clas;
	int				refs;
	
	const char		*type;
	BBSize			dims;
	BBSize			size;
	BBSize			scales[1];
	
	/** The array's data as a \c BBShort C-style array */
	BBShort			data[];
};

/** Helper array object storing BBInts
 *
 * \extends	BBArray
 */
struct BBIntArray
{
	BBClass			*clas;
	int				refs;
	
	const char		*type;
	BBSize			dims;
	BBSize			size;
	BBSize			scales[1];
	
	/** The array's data as a \c BBInt C-style array */
	BBInt			data[];
};

/** Helper array object storing BBLongs
 *
 * \extends	BBArray
 */
struct BBLongArray
{
	BBClass			*clas;
	int				refs;
	
	const char		*type;
	BBSize			dims;
	BBSize			size;
	BBSize			scales[1];
	
	/** The array's data as a \c BBLong C-style array */
	BBLong			data[];
};

/** Helper array object storing BBFloats
 *
 * \extends	BBArray
 */
struct BBFloatArray
{
	BBClass			*clas;
	int				refs;
	
	const char		*type;
	BBSize			dims;
	BBSize			size;
	BBSize			scales[1];
	
	/** The array's data as a \c BBFloat C-style array */
	BBFloat			data[];
};

/** Helper array object storing BBDoubles
 *
 * \extends	BBArray
 */
struct BBDoubleArray
{
	BBClass			*clas;
	int				refs;
	
	const char		*type;
	BBSize			dims;
	BBSize			size;
	BBSize			scales[1];
	
	/** The array's data as a \c BBDouble C-style array */
	BBDouble		data[];
};

/** Helper array object storing BBObjects
 *
 * \extends	BBArray
 */
struct BBObjectArray
{
	BBClass			*clas;
	int				refs;
	
	const char		*type;
	BBSize			dims;
	BBSize			size;
	BBSize			scales[1];
	
	/** The array's data as a \c BBObject* C-style array */
	BBObject		*data[];
};

/** Helper array object storing BBStrings
 *
 * \extends	BBArray
 */
struct BBStringArray
{
	BBClass			*clas;
	int				refs;
	
	const char		*type;
	BBSize			dims;
	BBSize			size;
	BBSize			scales[1];
	
	/** The array's data as a \c BBString* C-style array */
	BBString		*data[];
};

/** Helper array object storing other BBArrays
 *
 * \extends	BBArray
 */
struct BBArrayArray
{
	BBClass			*clas;
	int				refs;
	
	const char		*type;
	BBSize			dims;
	BBSize			size;
	BBSize			scales[1];
	
	/** The array's data as a \c BBArray* C-style array */
	BBArray			*data[];
};


/** \internal
 *
 * \brief	Allocates a new multi-dimensional array
 *
 * This allocates a new multi-dimensional array with the GC and sets
 * all of its required information.
 *
 * \see		bb_array_alloc1d, bb_array_init
 *
 * \param	type The type tag describing the data to store
 * \param	dims The number of dimensions
 * \param	lens The number of elements for each dimension
 * \returns	A new multi-dimensional array
 */
BBArray* bb_array_alloc(const char *type, BBSize dims, const BBSize *lens);

/** \internal
 *
 * \brief	Allocates a new 1-dimensional array.
 * 
 * This allocates a new 1-dimensional array with the GC and sets all
 * of its required information.
 *
 * If you know that you only need one dimension and have all the necessary
 * data to fill the array with, this is the best way to go in terms of speed.
 *
 * \note	The returned array's data is not initialized, you have to either
 *			completely fill it with fitting values or use bb_array_init.
 * \see		bb_array_alloc, bb_array_init
 *
 * \param	type The type tag describing the data to store
 * \param	len The number of array elements
 * \returns	A new 1-dimensional array.
 */
BBArray* bb_array_alloc1d(const char *type, BBSize len);

/** \internal
 *
 * \brief	Initializes an array
 *
 * After allocation via bb_array_alloc or bb_array_alloc1d,
 * the array's parameters are set, but the data is not initialized.
 *
 * Initialized means that each data type has a default value, like
 * any BBInt variable is set to 0 by default.
 *
 * Of course this can also be used to reset an array that has already
 * been used to its initial state.
 *
 * \note	It does not matter how the array was allocated.
 *			However, arrays allocated by bbArrayNew, bbArrayNew1D or bbArrayNewEx
 *			are already fully initialized.
 * 
 * \param	arr The array of which the data should be initialized
 */
void bb_array_init(BBArray *arr);



/** Creates a new array
 *
 * Allocates a new multi-dimensional array with the GC and initializes its data.
 *
 * \param	type Type tag describing the data to store.
 * \param	dims The number of dimensions
 * \return	A new multi-dimensional array
 */
BBArray* bbArrayNew(const char *type, BBInt dims, ...);

/** Creates a new 1-dimensional array
 *
 * Allocates a new 1-dimensional array with the GC and initializes its data.
 * Since this is the most common method of creating an array, it is heavily
 * optimized, so it is a good idea to always use this.
 *
 * \note	The only reason to use the bb_array_alloc1d instead
 *			is when you have all the data to fill the new array
 *			with, so initializing it would be unnecessary.
 *
 * \param	type Type tag describing the data to store.
 * \param	len The number of array elements
 * \returns	A new 1-dimensional array
 */
BBArray* bbArrayNew1D(const char *type, BBInt length);

//	Creates a new array. Only used by brl.reflection as far as I can tell
BBArray* bbArrayNewEx(const char *type, BBInt dims, BBInt *lens);

/** Creates a new array from a subset of the given array
 *
 * \param	type The type tag indicating the array's data type
 * \param	arr The array to be sliced
 * \param	begin The start position of the resulting array section
 * \param	end The last position of the reuslting array section
 * \returns	A new array containing the specified section of the array
 */
BBArray* bbArraySlice(const char *type, const BBArray *arr, BBInt begin, BBInt end);

/** Creates a new array from a block of data
 *
 * \param	type The type tag indicating the new array's data type
 */
BBArray* bbArrayFromData(const char *type, BBInt length, const void *data);

/** Casts an object to an array depending on the type tag
 *
 * \param	obj The object pointer which points to an array
 * \param	typetag The type of the array that is pointed to by \a obj
 * \returns	A pointer to the casted array, or \c BBNULL as a \c BBArray*.
 */
BBArray* bbArrayCastFromObject(const BBObject *obj, const char *typetag);

/** Sorts an array depending on its data type
 *
 * \note	This changes the indexing of the given array.
 *
 * \param	arr The array to be sorted
 * \param	ascending If the order of sorting should be ascending or descending
 */
void bbArraySort(BBArray *arr, BBInt ascending);

/** Gives an array's dimensions
 *
 * If the array's length is \c 0, \c BBNULLARRAY is returned.
 *
 * \param	arr The array to get the dimensions for
 * \returns A new array containing the size of every dimension of \a arr or \c BBNULLARRAY.
 */
BBArray* bbArrayDimensions(const BBArray *arr);

/** Concatenates two arrays to a new one
 *
 * If the resulting array's length would be \c 0, \c BBNULLARRAY is returned.
 *
 * \param	type The type of both array's as a type tag
 * \param	arr The array that is the beginning of the resulting one
 * \param	rhs The right-hand side array that will be concatenated to the first one
 * \returns A new array consisting of both given arrays or \c BBNULLARRAY.
 */
BBArray* bbArrayConcat(const char *type, const BBArray *arr, const BBArray *rhs);

//	Additions by BME

/** Compares two arrays
 *
 * Comparison is done with respect to a number of properties and of course both
 * array's data. The importance of each tested value is listed below.
 *
 * If one of the following properties in the given \a arr is greater than or
 * equal to the corresponing one in the \a rhs array, 1 is returned. Otherwise -1.
 * 
 * \li \c scales[0]
 * \li \c dims
 * \li \c size
 * \li \c type_tag
 *
 * If all properties are equal, the arrays' data is compared using memcmp.
 *
 * \see		bbArrayEquals
 *
 * \param	arr The array to be compared against
 * \param	rhs The righthand side array to compare the first one with
 * \retval	0 If both arrays are equal
 * \retval	1 If \a arr is greater than or equal to \a rhs
 * \retval	-1 If \a arr is lesser than \a rhs
 */
BBInt bbArrayCompare(const BBArray *arr, const BBArray *rhs) __attribute__ ((pure));

/** Gives a string description the array
 *
 * The string describing the given array has the format:
 * (Array) $TYPE[$DIMS]
 *
 * For a 1-dimensional BBFloat array with three elements, it would be:
 * (Array) Float[3]
 *
 * If the given array \a arr is \c BBNULLARRAY, "(Array) Null" is returned.
 *
 * \param	arr The array to get a description of
 * \returns	A new string containing the array's type and dimensions
 */
BBString* bbArrayToString(const BBArray *arr);

/** Creates an exact clone of an array
 *
 * Primitive data type arrays, namely Byte, Short, Int, Long, Float, Double and
 * their respective pointer counterparts are copied by value.
 *
 * Complex data type arrays on the other hand are copied by reference, meaning
 * every element will have an increase reference count, but won't be cloned itself.
 *
 * If \c arr is an empty array, BBNULLARRAY is returned.
 *
 * \param	arr The array to be cloned
 * \returns	A new array containing the values of the given one.
 *			BBNULLARRAY in case of an error.
 */
BBArray* bbArrayClone(const BBArray *arr);

/** Checks two arrays for equality
 *
 * \see		bbArrayCompare
 *
 * \param	arr The array to be checked for equality
 * \param	rhs The other array, that the first one is compared against
 * \retval	BBTRUE if both arrays contain the exact same values.
 * \retval	BBFALSE if there are differences between both arrays.
 */
BBBool bbArrayEquals(const BBArray *arr, const BBArray *rhs) __attribute__ ((pure));



/** Shows that an element could not be found within a given array
 *
 * \see		bbArrayFindInt, bbArrayFindLong, bbArrayFindFloat, bbArrayDouble,
 *			bbArrayFindObject, bbArrayFindString
 */
#define BBARRAY_NOTFOUND	(-1)

/** Finds the first occurance of a subarray's data in an array
 *
 * \param	arr The array to be searched
 * \param	sub The subarray which provides the values to be looked for
 * \returns	The position of the first occurance of all of the values of \a subarray, or \c BBARRAY_NOTFOUND.
 */
BBInt bbArrayFind(const BBArray *arr, const BBArray *sub, BBInt start_pos) __attribute__ ((pure));

/** Finds the first occurance of a BBInt value in an array
 *
 * \param	arr The array to be searched
 * \param	num The BBInt value to be searched for
 * \param	start_pos The start index of the search
 * \returns	The position of the first occurance of \a num, or \c BBARRAY_NOTFOUND.
 */
BBInt bbArrayFindInt(const BBArray *arr, BBInt num, BBInt start_pos) __attribute__ ((pure));

/** Finds the first occurance of a BBLong value in an array
 *
 * \param	arr The array to be searched
 * \param	search The BBLong value to be searched for
 */
BBInt bbArrayFindLong(const BBLongArray *arr, BBLong num, BBInt start_pos) __attribute__ ((pure));

/** Finds the first occurance of a BBFloat value in an array
 *
 * \param	arr The array to be searched
 * \param	search The BBFloat value to be searched for
 */
BBInt bbArrayFindFloat(const BBFloatArray *arr, BBFloat num, BBInt start_pos) __attribute__ ((pure));

/** Finds the first occurance of a BBDouble value in an array
 *
 * \param	arr The array to be searched
 * \param	search The BBDouble value to be searched for
 */
BBInt bbArrayFindDouble(const BBDoubleArray *arr, BBDouble num, BBInt start_pos) __attribute__ ((pure));

/** Finds the first occurance of an object in an array
 *
 * \param	arr The array to be searched
 * \param	obj The BBObject to be searched for
 */
BBInt bbArrayFindObject(const BBObjectArray *arr, const BBObject *obj, BBInt start_pos) __attribute__ ((pure));

/** Finds the first occurance of a string in an array
 *
 * \param	arr The array to be searched
 * \param	str The BBString to be searched for
 */
BBInt bbArrayFindString(const BBStringArray *arr, const BBString *str, BBInt start_pos) __attribute__ ((pure));



/** Checks if an array starts with another one
 *
 * \param	arr The array to be checked against
 * \param	rhs The subarray to be checked for
 * \retval	BBTRUE If the array starts with the given subarray
 * \retval	BBFALSE If the subarray does not match the beginning of the array
 */
BBBool bbArrayStartsWith(const BBArray *arr, const BBArray *rhs) __attribute__ ((pure));

/** Checks if an array ends with another one
 *
 * \param	arr The array to be checked against
 * \param	rhs The subarray to be checked for
 * \retval	BBTRUE If the array ends with the given subarray
 * \retval	BBFALSE If the subarray does not match the end of the array
 */
BBBool bbArrayEndsWith(const BBArray *arr, const BBArray *rhs) __attribute__ ((pure));



/** Checks if an array contains another one
 *
 * \param	arr The array to be searched
 * \param	sub The subarray to be searched for
 * \retval	BBTRUE If the array contains the subarray
 * \retval	BBFALSE If the subarray could not be found in the array
 */
BBBool bbArrayContains(const BBArray *arr, const BBArray *sub) __attribute__ ((pure));

/** Checks if an array contains a BBInt number
 *
 * \param	arr The array to be searched
 * \param	num The number which is searched for
 * \retval	BBTRUE If the array contains the number
 * \retval	BBFALSE If the number could not be found in the array
 */
BBBool bbArrayContainsInt(const BBArray *arr, BBInt num) __attribute__ ((pure));

/** Checks if an array contains a BBLong number
 *
 * \param	arr The array to be searched
 * \param	num The number which is searched for
 * \retval	BBTRUE If the array contains the number
 * \retval	BBFALSE If the number could not be found in the array
 */
BBBool bbArrayContainsLong(const BBLongArray *arr, BBLong num) __attribute__ ((pure));

/** Checks if an array contains a BBFloat number
 *
 * \param	arr The array to be searched
 * \param	num The number which is searched for
 * \retval	BBTRUE If the array contains the number
 * \retval	BBFALSE If the number could not be found in the array
 */
BBBool bbArrayContainsFloat(const BBFloatArray *arr, BBFloat num) __attribute__ ((pure));

/** Checks if an array contains a BBDouble number
 *
 * \param	arr The array to be searched
 * \param	num The number which is searched for
 * \retval	BBTRUE If the array contains the number
 * \retval	BBFALSE If the number could not be found in the array
 */
BBBool bbArrayContainsDouble(const BBDoubleArray *arr, BBDouble num) __attribute__ ((pure));

/** Checks if an array contains an object
 *
 * \param	arr The array to be searched
 * \param	obj The object which is searched for
 * \retval	BBTRUE If the array contains the object
 * \retval	BBFALSE If the object could not be found in the array
 */
BBBool bbArrayContainsObject(const BBObjectArray *arr, const BBObject *obj) __attribute__ ((pure));

/** Checks if an array contains a string
 *
 * \param	arr The array to be searched
 * \param	str The string which is searched for
 * \retval	BBTRUE If the array contains the string
 * \retval	BBFALSE If the string could not be found in the array
 */
BBBool bbArrayContainsString(const BBStringArray *arr, const BBString *str) __attribute__ ((pure));



/** Reverses an array
 *
 * \param	arr The input array which should be reversed
 * \returns	A reversed version of the given array, or a pointer to the given \a arr.
 */
BBArray* bbArrayReverse(const BBArray *arr);

/** Rotates all array indexes
 *
 * The sign of \a count controls the direction of the rotation.
 * If \a count is negative, the array's indexes will be rotated to the left.
 * If \a count is positive, the array's indexes will be rotated to the right.
 * In the case that \a count is \c 0, the given array is returned without any rotation.
 *
 * The number of rotations given in \a count can be greater than the length of the array.
 * In that case the array's indexes will be rotateted from the beginning/end again.
 *
 * \param	arr The array to be rotated
 * \param	count The number of rotations
 * \returns	A rotated version of the array, or a pointer to the given \a arr.
 */
BBArray* bbArrayRotate(const BBArray *arr, BBInt count);

#ifndef BB_DISABLE_DEPRECATED

#	define BBARRAYSIZE(q,n)	(20+(n)*sizeof(int)+(q))
#	define BBARRAYDATA(p,n)	((void*)((char*)(p)+20+(n)*sizeof(int)))

#endif	// !BB_DISABLE_DEPRECATED

BB_END_DECLS

#endif	// BLITZ_ARRAY_H
