/**
 * \file	blitz_class.h
 * \brief	The class description used by every object
 *
 * Classes contain general information and functionality to work with instances
 * of the respective object. They do not store individual information for every
 * object instance.
 *
 * Apart from the obvious functionality via virtual functions and information
 * necessary for allocation and such, each BBClass also contains a \a debug_scope
 * property that holds descriptions used by BlitzMax.
 *
 * \author	Mark Sibly
 * \author	Fabian Niemann
 */

#ifndef BLITZ_CLASS_H
#define BLITZ_CLASS_H

#include "_common.h"

#include "blitz_debug.h"
#include "blitz_gc.h"

BB_BEGIN_DECLS

/** The maximum size an instance of a class (object) can take up
 *
 * \note	While theoretically this could be SIZE_MAX,
 *			it is not since BlitzMax operates on an int basis.
 *			This however should not be a problem, as it is very unlikely
 *			to ever have objects which are even nearly that large.
 */
#define BBCLASS_MAX_INSTANCESIZE	BBSIZE_MAX

/** The maximum number of classes that can be registered with the runtime.
 *
 * Used internally by \c bbClassRegister and \c bb_class_register_multiple to check
 * against overflow.
 */
#define BBCLASS_MAX_NUM_REGISTERED	BBSIZE_MAX


#define BBCLASS_NUM_VFNS			32

/** Contains critical information and functionality for object instances
 *
 * A class is a structure which contains all necessary information to allocate
 * and initialize an new instance of it with the GC.
 *
 * It also contains all functions that the user can call to work
 * with the class' instances that hold the data of an object.
 *
 * Every class has a basic set of named function pointers:
 * free, ctor, dtor, ToString, Compare, SendMessage and Clone.
 * These provide some often used funtionality and should be defined
 * in every class available to users.
 *
 * In contrast to the named function pointers, the class specific functionality
 * is contained in the virtual function table (vfns). Extending classes copy
 * these functions from their parent class and overwrite them if necessary.
 */
struct BBClass
{
	/** The class' parent class
	 * 
	 * \note	At the end of the cascade this will always point to
	 *			bbObjectClass for user-defined classes.
	 */
	BBClass			*super;
	
	/** The function used to free an instance from memory
	 *
	 * \param	obj The instance to be freed
	 */
	void			(*free)(BBObject *obj);
	
	/** The debug scope containing additional information about the class
	 *
	 * An often used property of this structure is the name of the class.
	 * Also the reflection functionality depends on the information in this
	 * property to be complete and correct.
	 *
	 * \see		BBDebugScope
	 */
	BBDebugScope	*debug_scope;

	/** The amount of memory an instance of the class takes
	 *
	 * This information is used to allocate a new instance with the GC.
	 * Usually this is automatically set by BlitzMax' compiler.
	 *
	 * \note	bbStringClass and bbArrayClass are exceptions as they have
	 *			an instance_size of 0. That is because their size depends
	 *			on how big the variable data they encapsulate really is.
	 *
	 * \note	This value must not be negative at any time.
	 */
	BBSize			instance_size;
	
	/** The class constructor
	 *
	 * This initializes the instance's data.
	 *
	 * \param	obj The class instance to be initialized
	 */
	void			(*ctor)(BBObject *obj);
	
	/** The class destructor
	 *
	 * This deinitializes any data if necessary.
	 *
	 * \note	Due to BlitzMax' managed memory nature, you can not be sure when
	 *			a destructor is called.
	 *
	 * \param	obj The class instance to be deinitialized
	 */
	void			(*dtor)(BBObject *obj);
	
	/** Gives a string representation of the class instance
	 *
	 * The actual format of the returned string is not specified.
	 * By default the bbObjectToString function – which only returns
	 * a string containing the memory address of the instance – is called.
	 *
	 * But especially for debugging purposes it can be usefull to actually
	 * return information which have a higher descriptive value.
	 *
	 * \param	obj The instance to get a description of
	 * \return	A string representing/describing the instance
	 */
	BBString*		(*ToString)(BBObject *obj);
	
	/** Compares two instances of a class
	 *
	 * \param	obj The object to compare against
	 * \param	rhs The right-hand side object which is compared to the first one
	 * \retval	0 If both objects are equal.
	 * \retval	>0 If the object is greater than the one on the right-hand side
	 * \retval	<0 If the object is lesser than the one on the right-hand side
	 */
	BBInt			(*Compare)(BBObject *obj, BBObject *rhs);
	
	/** Sends a message to an object
	 *
	 * \param	msg The action or notice that should be processed
	 * \param	src The object that send the message and is associated with it
	 * \returns	The response to the send message.
	 *			If there is none, BBNULL should be returned instead.
	 */
	BBObject*		(*SendMessage)(BBObject *msg, BBObject *src);
	
	/** Creates an exact copy of the object
	 *
	 * This should take the given object and create a new instance of
	 * the same class with the values set to be equal to the ones in
	 * the given object.
	 *
	 * If cloning is not desirable (static class) or impossible in a
	 * certain situation, a pointer to bbNullObject should be returned.
	 *
	 * \note	The actual impementation depends on the specific class,
	 *			these are just suggestions on how the function should behave.
	 *
	 * \note	Take caution when cloning instances containing pointers
	 *			to other complex data types. Remember to BB_INCREFS.
	 *
	 * \param	obj The class instance to be cloned
	 * \returns	A new instance of the class with the same properties
	 *			as the given object. If that is not possible, BBNULL should
	 *			be returned.
	 */
	BBObject*		(*Clone)(BBObject *obj);

	void			(*_reserved1_)();
	void			(*_reserved2_)();
	
	/** The virtual function table
	 *
	 * This contains all additional (user-defined) functions.
	 */
	void*			vfns[BBCLASS_NUM_VFNS];
};

//	The offset of the BBClass member 'vfns' 
#define BBCLASS_VFN_OFFSET			offsetof(struct BBClass, vfns)

//	Gives the offset of a virtual function pointer at the given index
#define BBCLASS_VFN_OFFSETOF(key)	( BBCLASS_VFN_OFFSET + (key) * sizeof (void*) )

/** Declares a virtual function to be an abstract method
 *
 * When assigned to a (named) virtual function, each call to that function
 * will result in a new BBNullMethodException to be thrown.
 *
 * \see		BBNullMethodException
 */
#define BBCLASS_ABSTRACT_METHOD		bbNullMethodError

/** Declares a virtual function to be an abstract function
 *
 * When assigned to a (named) virtual function, each call to that function
 * will result in a new BBNullFunctionException to be thrown.
 *
 * \see		BBNullFunctionException
 */
#define BBCLASS_ABSTRACT_FUNCTION	bbNullFunctionError

//	Intern

typedef BBString*	(*BBClassToStringFunc)(BBObject*);
typedef BBInt		(*BBClassCompareFunc)(BBObject*, BBObject*);
typedef BBObject*	(*BBClassSendMessageFunc)(BBObject*, BBObject*);
typedef BBObject*	(*BBClassCloneFunc)(BBObject*);

/** \internal
 *
 * \briefRegister multiple classes at once with the runtime
 *
 * \param	count The number of classes given in the variable argument list
 */
void bb_class_register_multiple(BBSize count, ...);

//	API

//	Needed for binary compatibility since the compiler depends on them.
//	Everywhere else the new function names, which follow the naming scheme
//	better, should be used instead of the old ones.

#define bbClassRegister				bbObjectRegisterType
#define bbClassAllRegistered		bbObjectRegisteredTypes


/** Registers a class with the runtime
 *
 * Internally the runtime keeps track of all registered classes to
 * use them when necessary, for example reflection.
 *
 * \note	You can use classes implemented in C without calling this,
 *			if you include them in the ".i" files of the module.
 *			Then, of course, they are not visible when using reflection.
 *
 * \warning	This does not check if a class is already included, since this
 *			is usually done by the compiler. So be careful not to include
 *			something twice.
 *
 * \param	clas The class to register
 */
void bbClassRegister(const BBClass *clas);

/** Lists all registered classes
 *
 * \param	count An pointer to an integer in which the number of registered classes is stored.
 * \returns	A pointer to the first element in the list of registered classes
 */
BBClass** bbClassAllRegistered(BBInt *count);

BB_END_DECLS

#endif	// BLITZ_OBJECT_H
