/**
 * \file	blitz_debug.h
 * \brief	Additional information about BlitzMax code
 *
 * Using the structs in this header allows for reading information
 * from BlitzMax code from C/C++. Also the other way around, meaning
 * filling out these data structures with the correct values allows
 * you to write pure C code that can be seemlessly accessed from BlitzMax.
 *
 * That includes everything from extending classes to accessing meta data
 * and methods via the reflection functionality of BlitzMax.
 *
 * \author	Mark Sibly
 */

#ifndef BLITZ_DEBUG_H
#define BLITZ_DEBUG_H

#include "_common.h"

BB_BEGIN_DECLS

#ifdef BB_DEBUG
void bb_assert(const char *expr, const char *file, int line, const char *msg, ...);

/** Runtime time assertion
 *
 * This ensures that the \a expr statement is always true at run time.
 * If it fails a BBRuntimeException describing the problem is thrown.
 *
 * \warning	Asserts are only evaluated in debug mode (BB_DEBUG).
 * \see		BBRuntimeException
 *
 * \param	expr The expression that must always be true.
 * \param	msg An error message describing what went wrong
 */
#	define BB_ASSERT(expr, msg, ...) \
	if (!(expr)) { bb_assert(#expr, __FILE__, __LINE__, (msg), ##__VA_ARGS__); }
#else
#	define BB_ASSERT(expr, msg, ...)
#endif

/** All possible debug declarations
 *
 * \see		BBDebugDecl
 */
enum BBDebugDeclKind
{
	/** Marks the end of a scopes debug declarations */
	BBDEBUGDECL_END,
	
	/** Declares a constant */
	BBDEBUGDECL_CONST,
	
	/** Declares a local variable */
	BBDEBUGDECL_LOCAL,
	
	/** Declares a class member variable */
	BBDEBUGDECL_FIELD,
	
	/** Declares a static class member variable */
	BBDEBUGDECL_GLOBAL,
	
	/** Declares a variable pointer */
	BBDEBUGDECL_VARPARAM,

	/** Declares a class virtual function
	 *
	 * Virtual functions are stored in a \c BBClass struct's \c vfns member.
	 */
	BBDEBUGDECL_TYPEMETHOD,
	
	/** Declares a static class function */
	BBDEBUGDECL_TYPEFUNCTION
};

typedef enum BBDebugDeclKind		BBDebugDeclKind;

/** A debug declaration containing information about a piece of user code
 *
 * This describes a single elment of BlitzMax code, for example
 * a Local variable or the Field of a Type.
 */
struct BBDebugDecl
{
	/** The kind of the following declaration */
	BBDebugDeclKind		kind;
	
	/** The name of the declared entity */
	const char			*name;
	
	/** The type tag of the declared entity */
	const char			*type_tag;
	
	union {
		BBString*		const_value;
		BBSize			local_offset;
		BBSize			field_offset;
		void			*global_address;
	};
};

typedef struct BBDebugDecl			BBDebugDecl;

/** All valid debug scopes
 *
 * \see		BBDebugScope
 */
enum BBDebugScopeKind
{
	/** Declares a function */
	BBDEBUGSCOPE_FUNCTION = 1,
	
	/** Declares a user-defined type */
	BBDEBUGSCOPE_USERTYPE,
	
	/** Declares a local block */
	BBDEBUGSCOPE_LOCALBLOCK
};

typedef enum BBDebugScopeKind		BBDebugScopeKind;

/** A scope of information for a block of code
 *
 * Although the name implies something different, debug scopes are not
 * solely used for debugging. For example every user type includes its
 * own debug scope containing information about its members.
 *
 * Reflection depends heavily on these information.
 */
struct BBDebugScope
{
	/** The category of code block this scope represents */
	BBDebugScopeKind	kind;
	
	/** The unique name under which the code can be found */
	const char			*name;
	
	/** The declarations describing each element of the scope */
	BBDebugDecl			decls[1];
};

#ifdef BB_ENABLE_HELPER

//	Used internally to create debug scopes for BlitzMax classes
//	implemented in C e.g. bbRuntimeExceptionClass
typedef struct bb_debugscope_s
{
	BBDebugScopeKind	kind;
	const char			*name;
	BBDebugDecl			decls[];
} bb_debugscope_t;

#endif	// BB_ENABLE_HELPER

/** A debug statement pointing to the location of the error
 *
 */
struct BBDebugStm
{
	/** The source file within the error was found */
	const char			*source_file;
	
	/** The line containing the invalid code */
	BBSize				line_num;
	
	/** The character position at which the compiler found an invalid instruction */
	BBSize				char_num;
};

typedef struct BBDebugStm		BBDebugStm;
typedef struct BBDebugScope		BBDebugScope;

extern void (*bbOnDebugStop)();
extern void (*bbOnDebugLog)(BBString *msg);
extern void (*bbOnDebugEnterStm)(BBDebugStm *stm);
extern void (*bbOnDebugEnterScope)(BBDebugScope *scope, void *inst);
extern void (*bbOnDebugLeaveScope)();
extern void (*bbOnDebugPushExState)();
extern void (*bbOnDebugPopExState)();
extern void (*bbOnDebugUnhandledEx)(BBObject *ex);


#ifndef BB_DISABLE_DEPRECATED

extern void bbCAssertEx();

#ifndef NDEBUG
#	define bbassert(expr) if (!(expr)) { bbCAssertEx(); }
#else
#	define bbassert(expr)
#endif

#endif	// !BB_DISABLE_DEPRECATED

BB_END_DECLS

#endif	// BLITZ_DEBUG_H
