/**
 * \file	blitz_ex.h
 * \brief	Exception handling functionality
 *
 * Throwing an exception instead of returning an error code has the advantage
 * of giving a more detailed description of the error without sacrificing the
 * ability to return any value you want.
 *
 * Also, it gives you the ability to write libraries (modules) that leave the task
 * of handling the error to the application's programmer. So the application can
 * deal with the problem in an appropriate way, given it has information about the
 * current situation that an external library simply does not have.
 *
 * In BlitzMax any object can be used thrown and catched to handle exceptional
 * events. The runtime has it's own exceptions that can be found in the 'exceptions'
 * directory. DO NOT EXTEND THOSE (except maybe bbRuntimeExceptionClass).
 *
 * \author	Mark Sibly
 */

#ifndef BLITZ_EX_H
#define BLITZ_EX_H

#include "_common.h"

BB_BEGIN_DECLS

#if defined BB_OS_MACOS && defined __i386__
#	define BB_ARGP		1
	void* bbArgp(int offset);
#endif

/** \internal
 *
 * \brief	Gives the exception stack's CPU state.
 *
 * Automatically called by every BlitzMax \c Try statement.
 *
 * \returns	The exception stack's CPU state before any operation in the \c Try
 *			block was performed.
 */
void* bbExEnter();

/** \internal
 *
 * \brief	Pops the exception stack by one.
 *
 * Automatically called by a BlitzMax \c EndTry statement when no
 * exception was thrown in the \c Try block.
 */
void bbExLeave();

/** Throws an object as as an exception
 *
 * \param	obj An exception object describing the problem
 */
void bbExThrow(BBObject *obj);

/** Throws an exception message
 *
 * This is propably the fastest way to throw a message out of a C/C++ context.
 * Simply pass the message you want to be thrown as a "const char*" and this
 * will automatically convert it to a BBString. This string can than be catched
 * from within BlitzMax.
 *
 * \param	c_str A message describing the problem
 */
void bbExThrowCString(const char *c_str);


//	Intern

void _bbExEnter(void *_cpu_state);
void* _bbExThrow(void *_cpu_state, void *ptr);

BB_END_DECLS

#endif	// BLITZ_EX_H
