/**
 * \file	blitz_gc.h
 * \brief	Interface for the intern garbage collectors
 *
 * Garbage collection means that the language takes care of cleaning up
 * and freeing of unused memory. Therefore you don't usually call destructors
 * in BlitzMax. The GC (Garbage Collector) takes care of this for you.
 *
 * In case you need to manually handle your objects you can also halt
 * the GC's execution via bbGCSuspend.
 *
 * You will notice that there are two different GC's included in BlitzMax.
 * The one found in "blitz_gc_rc.c" is a reference-counting GC. It is used
 * in single threaded mode of BlitzMax. You will work with this one most of the time.
 *
 * For multi-threaded applications a different approach is needed.
 * The "Mark Sibly GC" found in "blitz_gc_ms.c" is capable of cyclic references
 * and multi-threading and is therefore used in threaded building mode.
 *
 * \author Mark Sibly
 */

#ifndef BLITZ_GC_H
#define BLITZ_GC_H

#include "_common.h"

BB_BEGIN_DECLS

#ifdef BB_THREADED
#	include "blitz_thread.h"
#	ifndef BB_GC_MS
#		define BB_GC_MS
#	endif
#else
#	ifndef BB_GC_RC
#		define BB_GC_RC
#	endif
#endif

#define BBGC_MANYREFS				0x40000000

// Probably shouldn't be here...
#if __ppc__
#	define BBGC_NUM_ROOTREGS		19
#else
#	define BBGC_NUM_ROOTREGS		4
#endif	// __ppc__

void* bbGCRootRegs(void *ptr);
void bbGCStartup();

/** Operating modes for the GC
 * 
 * \see		bbGCGetMode, bbGCSetMode
 */
enum BBGCMode
{
	/** Sets the GC to automatic mode
	 * 
	 * In automatic mode, unused objects are collected as the GC sees fit.
	 *
	 * This is the default mode.
	 */
	BBGC_AUTOMATIC	= 1,
	
	/** Sets the GC to manual mode
	 *
	 * In manual mode, unused objects won't be collected until bbGCCollect
	 * is called manually by the programmer. Hence the name.
	 *
	 * \see		bbGCCollect
	 */
	BBGC_MANUAL		= 2,
	
	BBGC_AGGRESSIVE	= -1
};

#define BBGC_MODE_ISVALID(mode) \
	(																			\
		(mode) == BBGC_AUTOMATIC ||												\
		(mode) == BBGC_MANUAL ||												\
		(mode) == BBGC_AGGRESSIVE												\
	)

/** Gets the GC's operating mode
 *
 * \see		BBGCMode
 * \returns	The GC's current operating mode.
 */
BBGCMode bbGCGetMode();

/** Sets the GC's operating mode
 *
 * In case the \a mode parameter is invalid, operating mode is set to BBGC_AUTOMATIC.
 *
 * \param	mode The mode at which the GC should operate
 */
void bbGCSetMode(BBGCMode mode);

/** 
 *
 * \see		
 */
enum BBGCDebugMode
{
	/** Disables the GC debug mode
	 *
	 * This is the default mode.
	 */
	BBGC_NODEBUG,
	
	/** Enables the GC debug mode via stdout */
	BBGC_STDOUTDEBUG
};

/** Sets the debug mode of the GC
 *
 * In case the \a mode parameter is invalid, debug mode is set to \c BBGC_NODEBUG.
 *
 * \param	mode The debug mode of the GC
 */
void bbGCSetDebug(BBGCDebugMode mode);

/** All possible flags affecting the allocation and deallocation of an object with the GC
 *
 * \note	At the moment, only the multi-threaded garbage collector
 *			is affected by these flags.
 */
enum BBGCAllocFlags
{
	/** 
	 *
	 */
	BBGC_DEFAULT,
	
	/** 
	 *
	 */
	BBGC_ATOMIC,
	
	/** 
	 *
	 */
	BBGC_FINALIZE
};

/** Allocates a block of memory with the GC
 *
 * \param	size The size of the memory block in bytes
 * \returns	A pointer to the newly allocated memory
 *			or \c NULL in case of an error.
 */
void* bbGCMalloc(BBInt size, BBInt flags);

/** Allocates a new object managed by the GC
 *
 * \param	size The size of the instance to be allocated
 * \param	clas The new instance's class
 * \param	flags Modifying allocation/deallocation of the new instance
 * \returns	A new instance of the given class.
 *			In case of an error NULL is returned.
 *
 * \relates	bbObjectClass
 */
BBObject* bbGCAllocObject(BBSize size, const BBClass *clas, BBInt flags);

/** Validates a block of memory
 *
 * \retval	BBTRUE
 *			BBFALSE
 */
BBBool bbGCValidate(const void *ptr);

/** Gets the amount of memory currently allocated by the GC
 *
 * \returns	The currently amount of allocated memory.
 */
BBInt bbGCMemAlloced();

/** Collects unused memory
 *
 * If the GC is suspended, this has no effect.
 *
 * \see		bbGCResume
 *
 * \returns	The amount of memory collected in bytes.
 */
BBInt bbGCCollect();

/** The maximum number of times the GC can be suspended.
 *
 * It's "-1" because the ref-counting GC's bbGCCollect increases the count by one
 * at the beginning and resets it later. This value prevents an overflow.
 */
#define BBGC_MAX_SUSPENDED			(BBINT_MAX - 1)

/** Suspends the GC's execution
 *
 * Temporarily halts the execution of the GC. To keep track of multiple
 * commands to suspend, an internal counter is increased with every call.
 * 
 * Note that this function can only increase the counter up to BBGC_MAX_SUSPENDED
 * without a matching call to bbGCResume. Therefore the limit for the number
 * of supensions is also BBGC_MAX_SUSPENDED. After that all calls to this function
 * have no effect until bbGCResume is called.
 *
 * \note	Suspending and resuming are nested. That means that each call to
 *			bbGCSuspend must be matched by a call to bbGCResume.
 *
 * \see		bbGCResume, bbGCIsSuspended, BBGC_MAX_SUSPENDED
 */
void bbGCSuspend();

/** Checks if the GC is currently suspended
 *
 * \see		bbSuspend, bbGCResume
 *
 * \retval	BBTRUE if the GC is currently suspended
 * \retval	BBFALSE if the GC is currently running (not suspended)
 */
BBBool bbGCIsSuspended();

/** Resumes a suspended GC
 *
 * After suspending the GC with bbGCSuspend, this resumes its execution.
 * If the GC is already running (not suspended) this has no effect.
 *
 * To check if the GC is running, use bbGCIsSuspended
 *
 * \see		bbGCSuspend, bbGCISSuspended
 */
void bbGCResume();

/** 
 *
 */
void bbGCRetain(BBObject *obj);

/** 
 *
 */
void bbGCRelease(BBObject *obj);

// BB_RETAIN/BB_RELEASE should be used to prevent an object from garbage collection.
//
// This is mainly of use if an object is being stored outside of BlitzMax's 'sight' - say, in a C++ table.
// You can also use bbGCRetain/bbGCRelease functions above if necessary - MACROS are just faster.

#ifdef BB_GC_RC
#	define BB_RETAIN(obj)		{ ++(obj)->refs; }
#	define BB_RELEASE(obj)		{ if (!--(obj)->refs) { bbGCFree(obj); } }
#else
#	define BB_RETAIN(obj)		bbGCRetain(((BBObject *)(obj)));
#	define BB_RELEASE(obj)		bbGCRelease(((BBObject *)(obj)));
#endif

//	Internal use only
#ifdef BB_GC_RC

	/*
	 * Called when there are no more references (obj->refs == 0)
	 *
	 * \param	obj The object to free
	 */
	void	bbGCFree(BBObject *obj);

	/*
	 * Called after the destructor to send the memory to nirvana.
	 *
	 * \param	obj The object to deallocate
	 * \param	size The size of the object instance
	 */
	void	bbGCDeallocObject(BBObject *obj, BBSize size);

#	define	BB_INCREFS(obj)		{ ++(obj)->refs; }
#	define	BB_DECREFS(obj)		{ if (!--(obj)->refs) { bbGCFree(obj); } }
#else
#	define	BB_INCREFS(obj)		{}
#	define	BB_DECREFS(obj)		{}
#endif

//	DEPERECATED

#ifndef BB_DISABLE_DEPRECATED

#	ifdef BB_GC_RC
#		define BBINCREFS(obj)	{ ++(obj)->refs; }
#		define BBDECREFS(obj)	{ if (!--(obj)->refs) { bbGCFree(obj); } }
#		define BBRETAIN(obj)	{ ++(obj)->refs; }
#		define BBRELEASE(obj)	{ if (!--(obj)->refs) { bbGCFree(obj); } }
#	else
#		define BBINCREFS(obj)	{}
#		define BBDECREFS(obj)	{}
#		define BB_RETAIN(obj)	bbGCRetain(((BBObject *)(obj)));
#		define BB_RELEASE(obj)	bbGCRelease(((BBObject *)(obj)));
#	endif	// !BB_GC_RC

#endif	// !BB_DISABLE_DEPRECATED
 
BB_END_DECLS

#endif	// BLITZ_GC_H
