#include "blitz_gc.h"
#include "blitz_memory.h"
#include "blitz_string.h"

#ifdef BB_GC_MS

//#define DEBUG_GC

#define ALIGN_SIZE		BBMEM_ALIGNSIZE
#define ALIGN_MASK		BBMEM_ALIGNMASK

#define BBGC_MARKED		4
#define BBGC_LOCKED		8

#define PAGE_BITS		12
#define PAGE_SIZE		(1 << PAGE_BITS)
#define PAGE_MASK		((PAGE_SIZE / 4) - 1)

typedef struct GCBlock	GCBlock;

struct GCBlock
{
	GCBlock			*succ;
	int				flags;		// low 4 bits - rest is size
	char			data[0];
};

static int* Pages[1 << (25 - PAGE_BITS)];
static unsigned int PagesAlloced;

static GCBlock *UsedBlocks;
static GCBlock *FinalizedBlocks;

static char *FreeBuf;
static size_t FreeBufSize;
static GCBlock *FreeBlocks[256];

static BBGCMode Mode = BBGC_AUTOMATIC;
static BBGCDebugMode DebugMode = BBGC_NODEBUG;
static unsigned int gc_alloced;
static uint32_t gc_suspended = 0;	// todo: why not atomic?

static unsigned int NumGlobalVars = 0;
static void ***GlobalVars;

#ifdef BB_OS_WIN32
extern void *_bss_start__;
extern void *_bss_end__;
extern void *_data_start__;
extern void *_data_end__;
extern void *_start_data__;
extern void *_stop_data__;
extern void *_end__;
#endif	// BB_OS_WIN32

#ifdef BB_OS_LINUX
extern void *__data_start;
extern void *__bss_start;
extern void *_end;
#endif	// BB_OS_LINUX

static void **DATA_START;
static void **DATA_END;

#ifdef DEBUG_GC

static const char* object_get_class_name(const BBObject *ptr)
{
	BBDebugScope *scope = obj->clas->debug_scope;

	return scope ? scope->name : "?";
}

#endif	// DEBUG_GC

static void gc_error(void *ptr, const char *msg)
{
	printf("GC ERROR: %s, object=$%p\n", msg, ptr);
	fflush(stdout);	
	bbExThrowCString(msg);
}

static int gc_set_membit(void *ptr)
{
	unsigned page = (unsigned)ptr >> (PAGE_BITS + 7);
	
	if (!Pages[page]) {
		++PagesAlloced;
		Pages[page] = malloc(PAGE_SIZE);
		memset(Pages[page], 0, PAGE_SIZE);
	}
	
	unsigned offset = ((unsigned)ptr >> 9) & PAGE_MASK;
	unsigned mask = 1 << (((unsigned)ptr >> 4) & 31 );
	
	if (Pages[page][offset] & mask) {
		gc_error(ptr, "setMemBit error: membit already set");
	}
	
	Pages[page][offset] |= mask;
}

static void gc_clear_membit(void *ptr)
{
	unsigned page = (unsigned)ptr >> (PAGE_BITS + 7);
	
	if (!Pages[page]) {
		gc_error(ptr, "clrMemBit error: mempage does not exist");
	}

	unsigned offset = ((unsigned)ptr >> 9) & PAGE_MASK;
	unsigned mask = 1 << (((unsigned)ptr >> 4) & 31);

	if (!(Pages[page][offset] & mask)) {
		gc_error(ptr, "clrMemBit error: membit not set");
	}
		
	Pages[page][offset] &= ~mask;
}

static int gc_test_membit(const void *ptr)
{
	if ((unsigned) ptr & 15) {
		return 0;
	}
		
	unsigned int page = (unsigned)ptr >> (PAGE_BITS + 7);
	if (!Pages[page]) {
		return 0;
	}
		
	unsigned offset = ((unsigned)ptr >> 9) & PAGE_MASK;
	unsigned mask = 1 << (((unsigned)ptr >> 4) & 31);

	return Pages[page][offset] & mask;
}

static void gc_collect_mem();

static size_t GcHeapAlloced;
static size_t GcHeapSize;

static void* gc_heap_alloc(size_t size)
{
	size += ALIGN_MASK + 4;
	
	void *p = malloc(size);
	
	if (!p) {
		bbGCCollect();
		p = malloc(size);
		if (!p) {
			return NULL;
		}
	}

	GcHeapAlloced += size;

	if (GcHeapAlloced > GcHeapSize) {
		GcHeapSize = GcHeapAlloced;
#ifdef DEBUG_GC
		printf("heap_size=%i\n", GcHeapSize);
		fflush(stdout);
#endif
	}
	
	void *q = (void *)(((unsigned)p + ALIGN_MASK + 4) & ~ALIGN_MASK);
	*((void **)q - 1) = p;

	return q;
}

static void gc_heap_free(void *ptr, int size)
{
	GcHeapAlloced -= size;
	free(*((void **)ptr - 1));
}

static void* gc_alloc(size_t size, int flags)
{
	size = (size + sizeof (GCBlock) + 15) & ~15;
	int i= size / 16;
	
	GCBlock *block;
	
	if (i < 256 && (block = FreeBlocks[i])) {
		FreeBlocks[i] = block->succ;
	} else
	{
		static int alloced;
		if (Mode == BBGC_AUTOMATIC && (gc_alloced - alloced) > GcHeapSize / 3) {
			gc_collect_mem();
			alloced = gc_alloced;
		}
	
		if (i > 255) {
			block = gc_heap_alloc(size);
			gc_set_membit(block);
		} else if (block = FreeBlocks[i]) {
			FreeBlocks[i] = block->succ;
		} else
		{
			if (size > FreeBufSize)
			{
				if (FreeBufSize)
				{
					int i = FreeBufSize / 16;
					
					GCBlock *to_free = FreeBuf;
					to_free->flags = BBGC_MARKED;
					to_free->succ  = FreeBlocks[i];
					
					FreeBlocks[i] = to_free;
					gc_set_membit(to_free);
				}
				
				FreeBufSize = 65536;
				FreeBuf = gc_heap_alloc(FreeBufSize);
			}
			
			block = FreeBuf;
			FreeBuf += size;
			FreeBufSize -= size;
			gc_set_membit(block);
		}
	}
	
	block->succ  = UsedBlocks;
	block->flags = size | flags;
	UsedBlocks = block;

	gc_alloced += size;
	return block->data;
}

static GCBlock* gc_get_block(void *ptr)
{
	GCBlock *block = (GCBlock*)ptr - 1;
	
	if (gc_test_membit(block)) {
		return block;
	} else {
		return NULL;
	}
}

static void gc_free_block(GCBlock *block)
{
	int flags	= block->flags;
	int size	= flags & ~15;
	int i		= size / 16;

	if (i > 255) {
		gc_clear_membit(block);
		gc_heap_free(block, size);
	} else {
		block->flags= BBGC_MARKED;
		block->succ = FreeBlocks[i];
		FreeBlocks[i] = block;
	}
	
	gc_alloced -= size;
}

static void gc_mark(void *ptr)
{
	GCBlock *block = gc_get_block(ptr);
	if (!block) {
		return;
	}
		
	if (block->flags & BBGC_MARKED) {
		return;
	}
		
	block->flags |= BBGC_MARKED;
	
	if (block->flags & BBGC_ATOMIC) {
		return;
	}
		
	int size = block->flags & ~15;

	for (int i = sizeof (GCBlock); i < size; i += 4) {
		void **r = (char *)block + i;
		gc_mark(*r);
	}
}

void gc_collect_mem()
{
	static bool is_recursive;
	if (is_recursive) {
		return;
	}
	is_recursive = true;

#ifdef DEBUG_GC
	int ms = bbMilliSecs();
#endif
	
	{
		const BBThread * const current_thread = bbThreadGetCurrent();
		BBThread *locked_threads = _bbThreadLockThreads();
		
		unsigned int reg_index;
		void **reg;	
		
		for (BBThread *thread = locked_threads; thread; thread = thread->succ)
		{
			for (unsigned int data_index = 0; data_index < BBTHREAD_DATA_SIZE; ++data_index) {
				gc_mark(thread->data[data_index]);
			}

			if (thread == current_thread)
			{
				void *regs[BBGC_NUM_ROOTREGS];
				void **sp = bbGCRootRegs(regs);
				
				for (reg_index = 0; reg_index < BBGC_NUM_ROOTREGS; ++reg_index) {
					gc_mark(regs[reg_index]);
				}
				for (reg = sp; reg != thread->stackTop; ++reg) {
					gc_mark(*reg);
				}
			} else
			{
				for (reg_index = 0; reg_index < BBTHREAD_NUM_REGS; ++reg_index) {
					gc_mark((void *)thread->locked_regs[reg_index]);
				}
				for (reg = (void **)thread->locked_sp; reg != thread->stackTop; ++reg) {
					gc_mark(*reg);
				}
			}
		}
	}
	
	for (unsigned int i = 0; i < NumGlobalVars; ++i) {
		gc_mark(*GlobalVars[i]);
	}

#ifdef DEBUG_GC
	int mark_ms = bbMilliSecs();
#endif
	
	GCBlock *block;

	// mark locked blocks
	for (block = UsedBlocks; block; block = block->succ) {
		if (block->flags & BBGC_LOCKED) {
			gc_mark(block->data);
		}
	}

	// resurrect or free finalized blocks
	while (block = FinalizedBlocks)
	{
		FinalizedBlocks = block->succ;
		if (block->flags & BBGC_MARKED)
		{
			// resurrect me!
			block->succ = UsedBlocks;
			UsedBlocks = block;
			
			if (block->flags & BBGC_FINALIZE)
			{
				block->flags &= ~BBGC_FINALIZE;
#ifdef DEBUG_GC
				BBObject *obj = (BBObject *)block->data;
				printf("GC resurrected:%s @%p\n", object_get_class_name(obj), obj);
				fflush(stdout);
#endif
			}
		} else {
			gc_free_block(block);
		}
	}

	GCBlock **used_blocks = &UsedBlocks;

	size_t num_finalized = 0;
	
	while (block = *used_blocks)
	{
		if (block->flags & BBGC_MARKED) {
			used_blocks = &block->succ;
			block->flags &= ~BBGC_MARKED;
		} else
		{
			*used_blocks = block->succ;
			
			if (block->flags & BBGC_FINALIZE)
			{
				++num_finalized;
				
				BBObject *obj = (BBObject *)block->data;
				BBClass *clas = obj->clas;
				
				clas->free(obj);
				obj->clas = clas;
			}
			
			block->succ = FinalizedBlocks;
			FinalizedBlocks = block;
		}
	}
	
	if (!num_finalized)
	{
		// No finalizers were run, so it's OK to free blocks NOW.
		while (block = FinalizedBlocks) {
			FinalizedBlocks = block->succ;
			gc_free_block(block);
		}
	}
	
	_bbThreadUnlockThreads();
	
#ifdef DEBUG_GC
	int end_ms = bbMilliSecs();
	printf("gc ms=%i, marked=%i, marked_ms=%i, freed=%i, freed_ms=%i\n",
		   end_ms - ms, n_marked, mark_ms - ms, n_freed, end_ms - mark_ms);
	fflush(stdout);
#endif

	is_recursive = false;	
}

static bool gc_var_is_global(const int *ptr)
{
	if (ptr == &bbNullObject || ptr == &bbEmptyString || ptr == &bbEmptyArray) {
		return true;
	}

	return (ptr >= DATA_START && ptr < DATA_END - 1)
		&& (*(void **)ptr == &bbStringClass)
		&& (*(int **)(ptr + 1) == 0x7fffffff);
}

//***** GC Interface *****//

void bbGCStartup()
{
#ifdef BB_OS_LINUX
	DATA_START	= &__data_start;
	DATA_END	= &_end;
#endif
	
#ifdef BB_OS_MACOS
	int *seg = getsegbyname("__DATA");
	DATA_START	= (void **)seg[6];
	DATA_END	= (void **)(seg[6] + seg[7]);
#endif

#ifdef BB_OS_WIN32
	DATA_START	= &_data_start__;
	DATA_END	= &_bss_end__;
#endif
	
#ifdef DEBUG_GC
	printf("DATA_START=%p, DATA_END=%p\n", DATA_START, DATA_END);
	fflush(stdout);
#endif
	
	void **reg;
	NumGlobalVars = 0;
	for (reg = DATA_START; reg != DATA_END; ++reg)
	{
		void *p = *reg;
		if (gc_var_is_global(p)) {
			++NumGlobalVars;
		}
	}
	
#ifdef DEBUG_GC
	printf("Found %i global vars\n", NumGlobal);
	fflush(stdout);
#endif
	
	GlobalVars = (void ***)malloc(NumGlobalVars * 4);
	unsigned int i = 0;
	for (reg = DATA_START; reg != DATA_END; ++reg)
	{
		void *p = *reg;
		if (gc_var_is_global(p)) {
			GlobalVars[i++] = reg;
		}
	}

}

BBGCMode bbGCGetMode()
{
	return Mode;
}

void bbGCSetMode(BBGCMode mode)
{
	if (mode == BBGC_AUTOMATIC || mode == BBGC_MANUAL || mode == BBGC_AGGRESSIVE) {
		Mode = mode;
	} else {
		Mode = BBGC_AUTOMATIC;
	}
}

void bbGCSetDebug(BBGCDebugMode mode)
{
	if (mode == BBGC_NODEBUG || mode == BBGC_STDOUTDEBUG) {
		DebugMode = mode;
	} else {
		DebugMode = BBGC_NODEBUG;
	}
}


void* bbGCMalloc(BBInt size, BBInt flags)
{
	if (size <= 0) {
		return NULL;
	}
	
	void *mem_ptr;

	BB_LOCK
	
	mem_ptr = gc_alloc(size, flags & BBGC_ATOMIC);
	
	BB_UNLOCK
	
	return mem_ptr;
}

BBObject* bbGCAllocObject(BBSize size, const BBClass *clas, int flags)
{
	BB_ASSERT(size >= sizeof (BBObject), "The object's size is too small.");
	BB_ASSERT(size <= BBSIZE_MAX, "The object's size is too large.");	
	
	BBObject *obj;
	
	BB_LOCK
	
	obj = (BBObject *)gc_alloc(size, flags);
	obj->clas = clas;
	obj->refs = 0;
	
	BB_UNLOCK
	
	return obj;
}

BBBool bbGCValidate(const void *ptr)
{
	BBBool is_valid;

	BB_LOCK
	is_valid = gc_get_block(ptr) ? BBTRUE : BBFALSE;
	BB_UNLOCK
	
	return is_valid;
}

BBInt bbGCMemAlloced()
{
	return gc_alloced;
}

BBInt bbGCCollect()
{
	if (gc_suspended) {
		return 0;
	}
	
	BBInt mem_collected;
	
	BB_LOCK

	mem_collected = gc_alloced;
	gc_collect_mem();
	mem_collected -= gc_alloced;

	BB_UNLOCK
	
	return mem_collected;
}

void bbGCSuspend()
{
#ifdef BB_ENABLE_EXPERIMENTAL
#	ifdef BB_OS_MACOS
	if (gc_suspended < BBGC_MAX_SUSPENDED) {
		OSAtomicIncrement32(&gc_suspended);
	}
#	else
	bbAtomicAdd(&gc_suspended, 1);
#	endif	// !BB_OS_MACOS
#else
	if (gc_suspended < BBGC_MAX_SUSPENDED) {
		BB_LOCK
		++gc_suspended;
		BB_UNLOCK
	}
#endif	// !BB_ENABLE_EXPERIMENTAL
}

BBBool bbGCIsSuspended()
{
	return !gc_suspended;
}

void bbGCResume()
{
	if (gc_suspended)
	{
		BB_LOCK
		--gc_suspended;
		BB_UNLOCK
	}
}

void bbGCRetain(BBObject *obj)
{
	BB_LOCK
	
	GCBlock *block = gc_get_block(obj);
	
	if (block) {
		if (!obj->refs++) {
			block->flags |= BBGC_LOCKED;
		}
	}

	BB_UNLOCK
}

void bbGCRelease(BBObject *obj)
{
	BB_LOCK
	
	GCBlock *block = gc_get_block(obj);
	
	if (block) {
		if (!--obj->refs) {
			block->flags &= ~BBGC_LOCKED;
		}
	}

	BB_UNLOCK
}

#endif	// BB_GC_MS
