/**
 * \file	blitz_handle.h
 * \brief	Accessing objects through internally stored handles
 *
 * Handles can be used to access objects from out of BlitzMax. Since the handles
 * are simple integers, they can be used by about any programming language.
 *
 * Since BlitzMax is garbage collected, simply casting the BBObject* to an integer
 * could cause errors, since the GC would have no knowledge of the external usage
 * and would collect objects that are still used.
 *
 * \author	Mark Sibly
 */

#ifndef BLITZ_HANDLE_H
#define BLITZ_HANDLE_H

#include "_common.h"

BB_BEGIN_DECLS

/** Gives an integer handle for an object
 *
 * If the given object already has an integer handle, the existing one is returned.
 *
 * \param	obj The class instance to use
 * \returns	A integer handle representing the given object,
 *			or '0' in case the parameter points to bbNullObject.
 */
BBInt bbHandleFromObject(BBObject *obj);

/** Gives the object associated with an integer handle
 *
 * This function only retrieves a pointer to an object previously
 * created by a call to bbHandleFromObject. Any non-existant handles
 * will always result in a BBNULL as a return value.
 *
 * \see		bbHandleFromObject
 *
 * \param	handle An integer handle of an object
 * \returns	A pointer to the associated object,
 *			or BBNULL in case the handle was not created yet.
 */
BBObject* bbHandleToObject(BBInt handle);

/** Releases the internal hash for the handle
 *
 * This releases the internal integer handle as well as the object
 * associated with it.
 *
 * \param	handle An integer handle of an object
 */
void bbHandleRelease(BBInt handle);

BB_END_DECLS

#endif	// BLITZ_HANDLE_H
