/**
 * \file	blitz_incbin.h
 * \brief	Functionality to include any binary data in the binary
 *
 * 'Incbin' is a BlitzMax keyword. When executed, the compiler looks for the
 * file which is given as the parameter. If the file is found, it is included
 * in the application's binary file by the compiler.
 *
 * When the application is run, the function bbIncbinAdd is called and
 * automatically given the right parameters. This fills the internal structure
 * with the information which can then be accessed via bbIncbinPtr and bbIncbinLen.
 *
 * At least that's what I figured out so far - Fabian
 *
 * \author	Mark Sibly
 */

#ifndef BLITZ_INCBIN_H
#define BLITZ_INCBIN_H

#include "_common.h"
#include "blitz_memory.h"

BB_BEGIN_DECLS

/** The maximum size of a file that can be included via bbIncbinAdd
 *
 * \see		bbIncbindAdd, bbIncbindLen
 */
#define BBINCBIN_MAX_SIZE			BBMEM_MAX_ALLOCSIZE

/** Adds a file to the binary included data
 *
 * \param	file The file where the data came from
 * \param	buf A pointer to the binary included data
 * \param	size The size of the data in the range [0 .. BBINCBIN_MAX_SIZE]
 * \retval	1 if the file was successfully incbin'd
 * \retval	0 if the file was already included
 */
BBInt bbIncbinAdd(BBString *file, const void *buf, BBSize size);

/** Gives a pointer to the data stored under the given file name
 *
 * \param	file The file where the data came from
 * \returns	A pointer to the included data or NULL if there is none for this file
 */
void* bbIncbinPtr(const BBString *file);

/** Gives the length of the data stored under the given file name
 *
 * \param	file The file where the data came from
 * \returns	The length of the included data or 0 if the file was not incbin'd yet
 */
BBSize bbIncbinLen(const BBString *file);

BB_END_DECLS

#endif	// BLITZ_INCBIN_H
