/**
 * \file	blitz_macros.h
 * \brief	Low-level macros and defines to ease cross-platform programming
 *
 * \author	Fabian Niemann
 */

#ifndef BLITZ_MACROS_H
#define BLITZ_MACROS_H

#ifdef __cplusplus
#	define BB_BEGIN_DECLS	extern "C" {
#	define BB_END_DECLS		}
#else
#	define BB_BEGIN_DECLS
#	define BB_END_DECLS
#endif

BB_BEGIN_DECLS

//	Check OS to build for

/*! \def	BB_OS_LINUX 
 *
 * \brief	Defined if the operating system is Linux
 *
 * Equals the BlitzMax precompiler define ?Linux:
 * <tt>#ifdef BB_OS_LINUX</tt> does the same as <tt>?Linux</tt>
 */
#ifdef __linux
#	define BB_OS_LINUX		TRUE
#endif

/** Defined if the operating system is MacOS (X)
 *
 * Equals the BlitzMax precompiler define ?MacOS:
 * <tt>#ifdef BB_OS_MACOS</tt> does the same as <tt>?MacOS</tt>
 */
#ifdef __APPLE__
#	define BB_OS_MACOS		TRUE
#endif

/** Defined if the operating system is Windows
 *
 * Equals the BlitzMax precompiler define ?Win32:
 * <tt>#ifdef BB_OS_WIN32</tt> does the same as <tt>?Win32</tt>
 */
#ifdef _WIN32
#	define BB_OS_WIN32		TRUE
#endif

//	Check the system's CPU

#if defined (__ppc__)
/** Defined if the cpu's architecture is PPC.
 *
 * Equals the BlitzMax precompiler define ?Win32:
 * <tt>#ifdef BB_CPU_PPC</tt> does the same as <tt>?PPC</tt>
 */
#	define BB_CPU_PPC		TRUE
#endif

#if defined (__i386__) || defined (_X86_)
/** Defined if the cpu's architecture is X86.
 *
 * Equals the BlitzMax precompiler define ?X86:
 * <tt>#ifdef BB_CPU_X86</tt> does the same as <tt>?X86</tt>
 */
#	define BB_CPU_X86		TRUE
#endif

//	Check for C version

#if defined (__STDC_VERSION__)

#	if __STDC_VERSION__ >= 199901L
#		define BB_ENV_C99	TRUE
#	endif	// C99

#endif	// __STDC_VERSION__

//	Check for 32/64-bit environment

#if defined (__GNUC__)

#	if defined (__x86_64__) || __ppc64__
#		define BB_ENV_64BIT	TRUE
#	else
#		define BB_ENV_32BIT	TRUE
#	endif

#elif defined (_WIN32) || defined (_WIN64)

#	if defined (_WIN64)
#		define BB_ENV_64BIT	TRUE
#	else
#		define BB_ENV_32BIT	TRUE
#	endif

#endif	// _WIN32 || _WIN64

//	Double-check compile-time configuration in case something went wrong
//	on the way from blitz.bmx to gcc

#if defined (THREADED) && !defined (BB_THREADED)
#	define BB_THREADED		TRUE
#endif

#if !defined (NDEBUG) && !defined (BB_DEBUG)
#	define BB_DEBUG			TRUE
#endif

//	Set compile options

#if defined (__GNUC__) && defined (BB_ENV_C99)
//	Helpers are various (inline) functions and macros to ease development
#	define BB_ENABLE_HELPER
#endif	// GCC & C99

//	Error checking

#if defined (BB_ENABLE_EXPERIMENTAL) && !defined (BB_ENV_C99)
#	error "Experimental compilation mode requires C99."
#endif

//	Macros that actually do something

#define BB_MIN(x, y)		((x) < (y) ? (x) : (y))
#define BB_MAX(x, y)		((x) > (y) ? (x) : (y))
#define BB_ABS(x)			((x) > 0 ? (x) : -(x))

BB_END_DECLS

#endif	// BLITZ_MACROS_H
