/**
 * \file	blitz_memory.h
 * \brief	Basic memory handling functions
 *
 * The garbage collector needs memory to aligned on a 16-bit boundary.
 * So if you have to interact with it, use the following wrapper functions.
 *
 * \author Mark Sibly
 * \author Fabian Niemann
 */

#ifndef BLITZ_MEMORY_H
#define BLITZ_MEMORY_H

#include "_common.h"

BB_BEGIN_DECLS

#define BBMEM_ALIGNSIZE			16
#define BBMEM_ALIGNMASK			(BBMEM_ALIGNSIZE - 1)

#define BBMEM_MAX_ALLOCSIZE		(BBSIZE_MAX - BBMEM_ALIGNSIZE - 4)

/** Allocates a continous block of aligned memory.
 *
 * Both versions of BlitzMax's GC need the memory to be aligned on a 16-byte boundary
 * in order to work correctly. This function provides such an alignment.
 *
 * If the allocation failed, it tries to free unused memory by manually calling
 * the GC's collect function. Only if this also fails, NULL is returned.
 *
 * \param	size The size of memory in bytes
 * \return	A pointer to the newly allocated memory, or NULL.
 */
void* bbMemAlloc(BBInt size) __attribute__ ((malloc));

/** Frees a previously allocated block of aligned memory.
 *
 * Before attempting to free the given block of memory, a check for
 * a non-null pointer is performed.
 *
 * \param	mem A pointer to the block of memory to be freed.
 */
void bbMemFree(void *mem);

/** Changes the size of a block of memory
 *
 * The name of this function can be misleading, as it actually allocates
 * a new block of memory with the given \a new_size and copies into it
 * the contents of the old one.
 *
 * If \a new_size is larger than \a size, than the data from \a mem is copied
 * completely and everything after its original \a size contains undefined
 * values (all the stuff that has been there before).
 *
 * If on the other hand \a new_size is smaller than \a size, only the beginning
 * of the data from \a mem is copied there. You will lose the rest.
 *
 * The original memory block pointed to by \a mem is freed after copying.
 *
 * \see		bbMemAlloc
 *
 * \param	mem A pointer to the block of memory to be copied
 * \param	size The size of the memory block in bytes
 * \param	new_size The size of the newly allocated memory block
 * \returns	A pointer to the newly allocated block of memory, or NULL on failure.
 */
void* bbMemExtend(void *mem, BBInt size, BBInt new_size);

/** Clears a block of memory to 0
 *
 * When allocated, a block of memory is not filled with anything
 * one could expect. Therefore it is always a good idea to clear the
 * block before using it or ensuring that it is filled completely
 * with the right data.
 *
 * \param	dst The memory block which is to be cleared
 * \param	size The size of the memory block in bytes
 */
void bbMemClear(void *dst, BBInt size);

/** Copies the content of a block of memory into another.
 *
 * \warning	The two blocks (destination and source) may NOT overlap in memory,
 *			otherwise the behaviour is undefined.
 *
 * \param	dst The memory block to which the data is copied
 * \param	src The memory block containing the data
 * \param	size The exact number of bytes that are copied
 */
void bbMemCopy(void *dst, const void *src, BBInt size);

/** Moves the data from one memory location to another
 *
 * \note	The two blocks (destination and source) can overlap.
 *
 * \param	dst The memory location to which the data is moved
 * \param	src The memory block containing the data
 * \param	size The exact number of bytes that are moved
 */
void bbMemMove(void *dst, const void *src, BBInt size);

BB_END_DECLS

#endif	// BLITZ_MEMORY_H
