
#include "blitz_object.h"
#include "exceptions/blitz_nullmethodexception.h"

static BBDebugScope ClassScope =
{
	.kind			= BBDEBUGSCOPE_USERTYPE,
	.name			= "Object",
	.decls			= {
		BBDEBUGDECL_END
	}
};

BBClass bbObjectClass =
{
	.super			= NULL,
	.free			= bbObjectFree,
	.debug_scope	= &ClassScope,
	.instance_size	= sizeof (BBObject),
	
	.ctor			= bbObjectCtor,
	.dtor			= bbObjectDtor,
	
	.ToString		= bbObjectToString,
	.Compare		= bbObjectCompare,
	.SendMessage	= bbObjectSendMessage,
	.Clone			= bbObjectClone,
	
	._reserved1_	= bbObjectReserved,
	._reserved2_	= bbObjectReserved
};

BBObject bbNullObject =
{
	.clas	= NULL,
	.refs	= BBGC_MANYREFS
};

BBObject* bbObjectNew(const BBClass * const clas)
{
	//	At the moment only the MSGC uses the flag parameter,
	//	so we don not need to calculate it for the other one.
	
#ifdef BB_GC_MS
	BBInt flags = (clas->dtor != bbObjectDtor) ? BBGC_FINALIZE : BBGC_DEFAULT;
	BBObject *obj = bbGCAllocObject(clas->instance_size, clas, flags);
#else
	BBObject *obj = bbGCAllocObject(clas->instance_size, clas, BBGC_DEFAULT);
#endif
	
	clas->ctor(obj);
	return obj;
}

void bbObjectFree(BBObject *obj)
{
	const BBClass *clas = obj->clas;

#ifdef BB_GC_RC

	if (obj == &bbNullObject) {
		obj->refs = BBGC_MANYREFS;
		return;
	}

	clas->dtor(obj);
	bbGCDeallocObject(obj, clas->instance_size);

#else
	clas->dtor(obj);
#endif
}

void bbObjectCtor(BBObject *obj)
{
	obj->clas = &bbObjectClass;
}

void bbObjectDtor(BBObject *obj)
{
	obj->clas = NULL;
}

BBString* bbObjectToString(const BBObject *obj)
{
	char buf[32];
	snprintf(buf, 32, "%p", obj);
	return bbStringFromCString(buf);
}

BBInt bbObjectCompare(const BBObject *obj, const BBObject *rhs)
{
	return (char *)obj - (char *)rhs;
}

BBObject* bbObjectSendMessage(const BBObject *obj, const BBObject *src)
{
	return &bbNullObject;
}

BBObject* bbObjectClone(BBObject *obj)
{
	if (!obj || !obj->clas || obj == &bbNullObject) {
		return &bbNullObject;
	}
	
	BBObject *out = bbObjectNew(obj->clas);
	
	if (!out) {
		return &bbNullObject;	// todo: return NULL like bbObjectNew?
	}

	void *in_field, *out_field;
	
	for (const BBDebugDecl *decl = obj->clas->debug_scope->decls; decl->kind != BBDEBUGDECL_END; ++decl)
	{
		//	Only copy "Field" members
		if (decl->kind == BBDEBUGDECL_FIELD)
		{
			BBTypeKey tkey = BBTYPETAG_KEY(decl->type_tag);
			
			in_field	= (char *)obj + decl->field_offset;	
			out_field	= (char *)out + decl->field_offset;
			
			if (BBTYPEKEY_ISCOMPLEX(tkey))
			{
				printf("refs = %i\n", ((BBObject *)in_field)->refs);
				
				BBObject *out_objfield = out_field;
				out_field = in_field;
				bbGCRetain(out_objfield);
				
				printf("refs = %i\n", ((BBObject *)out_objfield)->refs);
			} else
			{
				//	Copies the raw memory data from the given object's "Field"
				//	to the new one by pointing to the offsets stored in debug_decl
				memcpy(out_field, in_field, BBTYPEKEY_SIZE(tkey));
			}
		}
	}
	
	return out;
}

#ifdef BB_ENABLE_EXPERIMENTAL

void bbObjectReserved()
{
	const BBChar message_buf[] = {
		'I','l','l','e','g','a','l',' ',
		'c','a','l','l',' ','t','o',' ',
		'r','e','s','e','r','v','e','d',' ',
		'm','e','t','h','o','d'
	};
	
	bbExThrow(
		bbRuntimeExceptionCreate(
			BBSTRING_FROMBUFFER(message_buf, 31),
			BBNULL
		)
	);
}

#else	// BB_ENABLE_EXPERIMENTAL

void bbObjectReserved()
{
	bbExThrowCString("Illegal call to reserved method");
}

#endif	// !BB_ENABLE_EXPERIMENTAL

BBObject* bbObjectDowncast(const BBObject *obj, const BBClass *clas)
{
	const BBClass *parent = obj->clas;
	
	while (parent && parent != clas) {
		parent = parent->super;
	}
	
	return parent ? obj : &bbNullObject;
}
