/**
 * \file    blitz_object.h
 * \brief   The root type which all other objects extend
 *
 * While classes contain information and functionality for all instances of a kind.
 * An object is one of said instances and holds variables ("Field") that describe
 * its current state. These member variables can be modified via the virtual functions
 * from the respective BBClass.
 *
 * \author	Mark Sibly
 * \author	Fabian Niemann
 */

#ifndef BLITZ_OBJECT_H
#define BLITZ_OBJECT_H

#include "_common.h"

#include "blitz_class.h"
#include "blitz_gc.h"

BB_BEGIN_DECLS

/** Basic BlitzMax object data type
 *
 * Every class extends BBObject in some way, therefore allowing
 * you to store any instance of any class in a BBObject*.
 *
 * \see		BBClass
 */
struct BBObject
{
	/** The class of the object
	 *
	 * Since all functionality and identification is based on classes,
	 * changes to this value affect how an object is handled by the runtime.
	 */
	BBClass		*clas;
	
	/** The number of times an instance is currently used
	 *
	 * \note	Only used in single-threaded mode (by the reference-counting GC).
	 * \warning	Do not change or read this manually, as it is controlled be the runtime.
	 *
	 * \protected
	 */
	int			refs;
};

/** The parent class to all other user-defined classes
 *
 * It contains default implementations for all named virtual functions
 * that are used if a class (and its parents) do not defined their own.
 */
extern BBClass bbObjectClass;

/** A typesafe \c NULL pointer for all user-defined objects
 *
 * This is used not only for variables of type \c BBObject, but for every
 * other user-defined object (not \c BBString and \c BBArray).
 *
 * All uninitialized user-defined objects should be initialized with this.
 *
 * \attention Keep in mind that arrays and strings got their own \c NULL objects.
 * \warning	Do not free this compile-time object.
 */
extern BBObject bbNullObject;

/** A typesafe \c NULL pointer for all user-defined objects
 *
 * Can be used as a NULL pointer for all \c BBObject* and user-defined types.
 * 
 * \note	Strings and arrays have their own special NULL pointers
 * \see		bbNullObject, BBNULLOBJECT
 */
#define BBNULL						(&bbNullObject)

//	Alias for bbNullObject that fits to the naming scheme
//	of BBNULLSTRING and BBNULLARRAY.
#define BBNULLOBJECT				(&bbNullObject)

/** Creates a new instance of a class.
 *
 * This allocates a new object with the GC and initializes it with
 * the given class' constructor function.
 *
 * All new instances of user defined classes (BlitzMax "Type") are created
 * with this function. The compiler calls it whenever he comes across the 
 * "New" keyword in BlitzMax code.
 *
 * \param	clas The class of the object to create
 * \returns	A pointer to the newly allocated and initialized object
 */
BBObject* bbObjectNew(const BBClass *clas);

/** Frees an object from memory
 *
 * Calls the class' destructor on the given \c obj.
 * In single-threaded mode this also deallocates the instance using the GC.
 *
 * \note	Passing BBNULL to this function does no harm.
 *
 * \param	obj A pointer to the object to be freed
 */
void bbObjectFree(BBObject *obj);

/** Initializes a generic object
 *
 * This sets the object's "clas" member to bbObjectClass.
 *
 * \param	obj A pointer to any object
 */
void bbObjectCtor(BBObject *obj);

/** Deinitializes a generic object
 *
 * This sets the object's "clas" member to NULL.
 *
 * \param	obj A pointer to any object
 */
void bbObjectDtor(BBObject *obj);

/** Gives the memory location of an object in a string
 *
 * \param	obj The object of which the memory location is returned
 * \returns	A string containing the location of the object instance in memory.
 */
BBString* bbObjectToString(const BBObject *obj);

/** Compares the memory location of two objects
 *
 * \param	obj The object to compare with
 * \param	rhs The righthand side object to compare against the first one
 * \returns	The difference of both objects memory locations.
 *			0 means they are at the same location and therefore equal.
 */
BBInt bbObjectCompare(const BBObject *obj, const BBObject *rhs);

/** Does not perform any action
 *
 * This is only a place holder for the named virtual function in each class.
 *
 * \param	obj The object
 * \param	src The source
 * \returns	BBNULL
 */
BBObject* bbObjectSendMessage(const BBObject *obj, const BBObject *src);

/** Creates an exact clone of the given object
 *
 * Cloning means that the returned object is a completely new instance
 * with the exact same property ("Field") values as the given one.
 *
 * This default implementatin reads the necessary information for creating
 * a new class instance from the debug_scope property of the given object's class.
 * Every "Field" declaration that is found, is then copied depending on its type:
 *
 * All primitive data types are copied by value using memcpy.
 * The complex ones (BBObject, BBString & BBArray) are copied by reference,
 * which means the cloned object's members point to the same instance.
 *
 * If the given object is either NULL, does not have a valid class
 * or is a pointer to bbNullObject, BBNULL is returned.
 *
 * \warning	
 *
 * \param	obj Any class instance
 * \return	A new instance of the given object's class with the same values.
 */
BBObject* bbObjectClone(BBObject *obj);

/** Throws a warning that the called virtual function is reserved for future use
 *
 * \note	If you try to access either "_reserved1_" or "_reserved2_" of
 *			a BBClass this function is called.
 */
void bbObjectReserved();

BBObject* bbObjectDowncast(const BBObject *obj, const BBClass *clas);

BB_END_DECLS

#endif	// BLITZ_OBJECT_H
