/**
 * \file	blitz_string.h
 * \brief	Functionality for UCS-2 character strings
 *
 * \author	Mark Sibly
 * \author	Fabian Niemann
 */

#ifndef BLITZ_STRING_H
#define BLITZ_STRING_H

#include "_common.h"

#include "blitz_array.h"
#include "blitz_object.h"

BB_BEGIN_DECLS

//	Used internally to calculate the amount of memory necessary
//	to allocate a string with a certain length.
#define BBSTRING_SIZE_CALC(len)			(sizeof (BBString) + (len) * sizeof (BBChar))

/** Gets the size a string takes up in memory
 *
 * The amount of memory the string uses is calculated from
 * the size of "struct BBString" and the size of its buffer.
 *
 * \param	str The string to calculate the size of
 * \returns	The complete size of the given string in bytes.
 */
#define BBSTRING_SIZE(str)				BBSTRING_SIZE_CALC((str)->length)

#ifndef BB_ENABLE_EXPERIMENTAL
/** The maximum length of a string object
 *
 * \todo	New formular. Experimental one is absolute maximum and not guaranteed to work.
 *			The other one is stable but could be too small.
 */
#	define BBSTRING_MAX_LENGTH				((BBSize)(65536 / sizeof (BBChar) - sizeof (BBString)))
#else
#	define BBSTRING_MAX_LENGTH				((BBSize)(BBSIZE_MAX / sizeof (BBChar)) - sizeof (BBString))
#endif	// BB_ENABLE_EXPERIMENTAL

/** Allocates a new string using the GC
 *
 * This is a utility macro which does not perform any
 * additional checking. It only fills in the right parameters
 * for allocating a string object with the GC.
 *
 * Only use this if you can ensure a valid length.
 * Otherwise it's better to use the bbStringNew function which
 * does the checking for you and is the standard way of creating
 * new string objects.
 *
 * \note	This does not set the length of the string object
 * \see		bbGCAllocObject, bbStringNew
 *
 * \param	len The length of the new string
 * \returns	A pointer to the newly allocated string
 */
#define BBSTRING_ALLOC(len) \
	({																			\
		(BBString *)															\
		bbGCAllocObject(														\
			BBSTRING_SIZE_CALC(len),											\
			&bbStringClass,														\
			BBGC_ATOMIC															\
		);																		\
	})

/* Creates a string from a BBChar buffer
 *
 * \param	buffer A buffer containing UCS-2 encoded string data
 * \param	len The length of the buffer and the new string
 * \returns	A pointer to a newly allocated string with the given contents
 */
#define BBSTRING_FROMBUFFER(buffer, len) \
	({																			\
		BB_ASSERT((len) > 0 && (len) < BBSTRING_MAX_LENGTH, "Invalid buffer length"); \
		BBString *_str	= BBSTRING_ALLOC((len));								\
		_str->length	= (len);												\
		memcpy(_str->buf, (buffer), (len) * sizeof (BBChar));					\
		_str;																	\
	})

/** Fixed-length string object
 *
 * In BlitzMax a string is a complex data type extending BBObject.
 * That means it is usually handled by the GC and can be stored in
 * \c BBObject pointers for convenience.
 *
 * All of the string's data is stored in a UCS-2 encoded \c BBChar buffer.
 * Therefore the length of the buffer equals the string's length.
 *
 * It can be assumed that every string's class is the bbStringClass,
 * since it's declared as a "Final Type" from the side of BlitzMax.
 * Therefore every pointer of the type <tt>BBString*</tt> can be used directly
 * with the string functions, instead of looking up the right one in the
 * virtual functions table (bbStringClass.vfns).
 *
 * \extends	BBObject
 */
struct BBString
{
	/** The class of the string
	 *
	 * \note	Should always point to bbStringClass
	 * \see		bbStringClass
	 */
	BBClass*		clas;
	int				refs;

	/** The length of the string
	 * 
	 * Since UCS-2 is used internally to store string data,
	 * this equals the number of BBChars stored in the buf property.
	 *
	 * \note	This value must be in the range [0 .. BBSTRING_MAX_LENGTH]
	 */
	BBSize			length;
	
	/** Contains the actual string data in UCS-2 encoding */
	BBChar			buf[];
};

/** Contains all operations that can be performed on strings
 *
 * Since strings are a final built-in class, in this case the class is not as
 * important as it is for user-defined classes.
 */
extern BBClass bbStringClass;

/** A typesafe \c NULL pointer for strings
 *
 * This \c NULL equivalent for a \c BBString pointer can be used in a wide range
 * of situations to show that an operation was out of range, could not be
 * performed or a string could not be allocated.
 *
 * \see		BBNULLSTRING, bbNullObject, bbEmptyArray
 * \warning	Do not free this compile-time object.
 */
extern BBString bbEmptyString;

/** Convience macro for using a pointer to bbEmptyString
 *
 * Returned in case a search, calculation or parsing action was unsuccessful.
 *
 * \see		BBNULL, BBNULLARRAY
 */
#define BBNULLSTRING					(&bbEmptyString)

#ifdef BB_ENABLE_HELPER

bool bb_string_buffer_equals(const BBChar * restrict str_buf,
							 const BBChar * restrict cmp_buf,
							 BBSize len) __attribute__ ((pure));

#endif	// BB_ENABLE_HELPER



//	Built-in

/** Passes a string through without changing anything
 *
 * This function only exists to ensure uniformity of classes.
 * When you operate from C/C++ and know that your object is a string,
 * it makes no sense to use it.
 *
 * \param	str The string to be passed through
 * \returns	The given string
 */
BBString* bbStringToString(const BBString *str);

/** Compares two strings
 *
 * The strings are compared by testing all of their characters for equality.
 * When a non-equal character is found, the difference between the code point
 * of \a str and \a rhs at that position is returned.
 *
 * If all characters are equal the difference between both strings' length is returned.
 *
 * \note	Passing the same string to \a str and \a rhs will return \c 0.
 *
 * \see		bbStringEquals
 *
 * \param	str The string to be compared
 * \param	rhs The right hand side string that is compared against the first one
 * \returns	The difference between the two strings, \c 0 if they are equal.
 */
BBInt bbStringCompare(const BBString *str, const BBString *rhs) __attribute__ ((pure));

/** Creates an exact new copy of a string
 *
 * This creates a new string with the same attribtues and content
 * as the one given as a parameter.
 *
 * Note that this is only useful in special cases, as the string API
 * always returns a new string object when anything is to be changed.
 * For example bbStringToLower already creates a new string with a buffer
 * filled with the lowercase characters.
 *
 * \param	str The string which should be cloned
 * \returns	A new string object, or BBNULLSTRING on failure.
 */
BBString* bbStringClone(const BBString *str);



//	API

/** Creates a new string
 *
 * Allocates a new string object with the GC and initializes
 * it with the given length.
 *
 * If the len parameter is negative or zero, a pointer to bbEmptyString
 * is returned instead of a new string. The same happens if memory allocation
 * with the GC failed, or the \a len is greater than \c BBSTRING_MAX_LENGTH.
 *
 * \param	len The length of the new string
 * \returns	A pointer to the newly created string object if successful, otherwise &bbEmptyString.
 */
BBString* bbStringNew(BBInt len);

/** Creates a new string from a character code
 *
 * If the given code is out of UCS-2 range, a \c BBRuntimeException will be thrown.
 *
 * \param	codepoint A UCS-2 character code
 * \returns	A new one character long string.
 */
BBString* bbStringFromChar(BBInt codepoint);

/** Creates a new string from an integer number
 *
 * Use this to convert \c BBByte, \c BBShort, \c BBInt, \c BBSize and \c BBBool variables to a string.
 *
 * \param	number The integer number that should be converted to a string
 * \returns	A new string containing the integer's value, or \c BBNULLSTRING.
 */
BBString* bbStringFromInt(BBInt number);

/** Creates a new string from a long integer number
 *
 * \param	number The long integer number that should be converted to a string
 * \returns	A new string containing the long integer's value, or \c BBNULLSTRING.
 */
BBString* bbStringFromLong(BBLong number);

/** Creates a new string from a single-precision floating point number
 *
 * \param	number The single-precision floating point number that should be converted to a string
 * \returns	A new string containing the \c float's value, or \c BBNULLSTRING.
 */
BBString* bbStringFromFloat(BBFloat number);

/** Creates a new string from a double-precision floating point number
 *
 * \param	number The double-precision floating point number that should be converted to a string
 * \returns	A new string containing the \c double's value, or \c BBNULLSTRING.
 */
BBString* bbStringFromDouble(BBDouble number);

/** Creates a new string from a buffer of byte data
 *
 * All values in the buffer are interpreted as <tt>unsigned char</tt>,
 * therefore the first 256 UCS-2 code points can be read from the buffer.
 *
 * If \a size is negative or \c 0, \c BBNULLSTRING is returned.
 * If \a size is larger than \c BBSTRING_MAX_LENGTH it is clamped.
 * In debug mode a \c BBRuntimeException is also thrown in that case.
 *
 * \param	buf A buffer of byte values
 * \param	size The length of the buffer
 * \returns	A new string containing the buffer's data, or \c BBNULLSTRING.
 */
BBString* bbStringFromBytes(const char *buf, BBInt size);

/** Creates a new string from a buffer of short integer data
 *
 * All values in the buffer are interpreted as <tt>unsigned short int</tt>,
 * therefore all code points from the Basic Multilingual Plane (BMP) can be read from the buffer.
 *
 * If \a size is negative or \c 0, \c BBNULLSTRING is returned.
 * If \a size is larger than \c BBSTRING_MAX_LENGTH it is clamped.
 * In debug mode a \c BBRuntimeException is also thrown in that case.
 *
 * \param	buf A buffer of short integer values
 * \param	size The length of the buffer
 * \returns	A new string containing the buffer's data, or \c BBNULLSTRING.
 */
BBString* bbStringFromShorts(const BBShort *buf, BBInt size);

/** Creates a new string from a buffer of integer data
 *
 * All values in the buffer are interpreted as <tt>signed int</tt>,
 * therefore all code points from the Basic Multilingual Plane can be read from the buffer.
 *
 * Code points out of the BMP will be clamped.
 * 
 * \todo	Better way to deal with code points above BMP.
 *
 * If \a size is negative or \c 0, \c BBNULLSTRING is returned.
 * If \a size is larger than \c BBSTRING_MAX_LENGTH it is clamped.
 * In debug mode a \c BBRuntimeException is also thrown in that case.
 *
 * \param	buf A buffer of integer values
 * \param	size The length of the buffer
 * \returns	A new string containing the buffer's data, or \c BBNULLSTRING.
 */
BBString* bbStringFromInts(const BBInt *buf, BBInt size);

/** Creates a new string from an array's data
 *
 * Reads all elements from a \c BByte, \c BBShort or \c BBInt array
 * and uses them as UCS-2 code points to create a string.
 * 
 * If the array's length is negative or \c 0, \c BBNULLSTRING is returned.
 * If the array's length is larger than \c BBSTRING_MAX_LENGTH it is clamped.
 * In debug mode a \c BBRuntimeException is also thrown in that case.
 *
 * \param	arr An array containing UCS-2 code points
 * \returns	A string containing all of the array's elements as code points, or \c BBNULLSTRING.
 */
BBString* bbStringFromArray(const BBArray *arr);

/** Creates a new string from a null-terminated string.
 *
 * If \a c_str is \c NULL, then \c BBNULLSTRING is returned.
 *
 * \param	c_str A null-terminated ASCII string
 * \returns	A new string containing the C-style string's data, or \c BBNULLSTRING.
 */
BBString* bbStringFromCString(const char *c_str);

/** Creates a new string from a null-terminated wide character string.
 *
 * If \c w_str is NULL then \c BBNULLSTRING is returned.
 *
 * \param	w_str A null-terminated wide string.
 * \param	A new string containing the wide string's data, or \c BBNULLSTRING.
 */
BBString* bbStringFromWString(const BBChar *w_str);

/** Creates a new string from an UTF-8 encoded null-terminated string
 *
 * If \a utf8_str is \c NULL, \c BBNULLSTRING is returned.
 *
 * \throws	BBRuntimeException In case a Code Point is out of UCS-2 range.
 *
 * \param	utf8_str A UTF-8 encoded null-terminated string
 * \returns	A string containing the UTF-8 string's data in UCS-2 encoding, or \c BBNULLSTRING.
 */
BBString* bbStringFromUTF8String(const char *utf8_str);


/** Checks if a string begins with a given substring
 *
 * \param	str The string to be searched in
 * \param	sub The substring which is searched for
 * \retval	BBTRUE if the given string starts with the substring
 * \retval	BBFALSE in any other case.
 */
BBBool bbStringStartsWith(const BBString *str, const BBString *sub) __attribute__ ((pure));

/** Checks if a string ends with a given substring
 *
 * \param	str The string to be searched in
 * \param	sub The substring which is searched for
 * \returns	BBTRUE if the given string ends with the substring, otherwise BBFALSE.
 */
BBBool bbStringEndsWith(const BBString *str, const BBString *sub) __attribute__ ((pure));

/** Checks if a string contains a substring
 *
 * If the given substring is longer than the string in which it is seached in,
 * or is empty, than BBFALSE is returned.
 *
 * \see		bbStringCount, bbStringFind
 *
 * \param	str The string to be searched in
 * \param	sub The substring which is searched for
 * \retval	BBTRUE if the given string contains the substring
 * \retval	BBFALSE if the substring is not found in the given string
 */
BBBool bbStringContains(const BBString *str, const BBString *sub) __attribute__ ((pure));



/** Concatenates two strings together to a new one
 *
 * \note	The BlitzMax compiler automatically concatenates constant strings
 *			at compile time. All run time adding operations use this function.
 *
 * \param	str The string to be at the beginning of the newly composed string
 * \param	rhs The string to be appended to the first one
 * \returns	A newly created string containing both given strings' data.
			In case of an error BBNULLSTRING is returned.
 */
BBString* bbStringConcat(const BBString *str, const BBString *rhs);


/** Specifies the side of a string
 *
 * Choosing a specific side of a string to be worked with/upon can be
 * useful to make string handling function more flexible.
 *
 * \see		bbStringTrim
 */
enum BBStringSide
{
	/** The operation will affect both sides of the string */
	BBSTRING_SIDE_BOTH	= 0,
	
	/** The operation will only affect the beginning of the string */
	BBSTRING_SIDE_LEFT	= -1,
	
	/** The operation will only affect the end of the string */
	BBSTRING_SIDE_RIGHT	= 1
};

typedef enum BBStringSide	BBStringSide;

#define BBSTRING_SIDE_ISVALID(side) \
	((side) >= BBSTRING_SIDE_LEFT && (side) <= BBSTRING_SIDE_RIGHT)

/** Removes whitespace from the specified sides of a string
 *
 * \arg	\c BBSTRING_SIDE_LEFT	Removes whitespace from the beginning of the string
 * \arg \c BBSTRING_SIDE_RIGHT	Removes whitespace from the end of the string
 * \arg \c BBSTRING_SIDE_BOTH	Removes whitespace from both sides of the string
 *
 * Any other value is handled like BBSTRING_SIDE_BOTH
 *
 * \param	str The string to remove whitespace from
 * \param	side The side or sides of the string to be edited
 * \returns	A string with all whitespace characters removed.
 */
BBString* bbStringTrim(const BBString *str, BBStringSide side);

/** Pads a string to a given length
 *
 * If the resulting length \a len is smaller than or equal to the length of the
 * given string \a str, the same string is returned without any padding.
 *
 * \arg	\c BBSTRING_SIDE_LEFT	Pads the beginning of the string with whitespace
 * \arg \c BBSTRING_SIDE_RIGHT	Pads the end of the string with whitespace
 * \arg \c BBSTRING_SIDE_BOTH	Pads both sides of the string with whitespace
 *
 * \param	str The string to be padded
 * \param	len The final length of the padded string
 * \param	side The side or sides of the string to be edited
 * \param	A new string padded to the given length, or the given one.
 */
BBString* bbStringPad(const BBString *str, BBInt len, BBStringSide side);

/** Gives a section of a string
 *
 * If the \a end position is lesser than the \a begin position, \c BBNULLSTRING is returned.
 *
 * \see		bbStringSub
 *
 * \param	str The string to be sliced
 * \param	begin The first position in \a str to start sclicing
 * \param	end The last position in \str to end slicing
 * \returns	A new string containing the given string section's data, or \c BBNULLSTRING. 
 */
BBString* bbStringSlice(const BBString *str, BBInt begin, BBInt end);

/** Replaces all occurances of a given substring in a string
 *
 * If the resulting string's length is \c 0, \c BBNULLSTRING is returned.
 * If the resulting string's length is greater than \c BBSTRING_MAX_LENGTH, is is clamped.
 * In debug mode a \c BBRuntimeException is also thrown in that case.
 *
 * \param	str The input string
 * \param	sub The substring which should be replaced
 * \param	replacement The string that will be used to replace \a sub in \a str
 * \returns	A new string with all occurances of \a sub replaced with \a replacement,
 *			or \c BBNULLSTRING.
 */
BBString* bbStringReplace(const BBString *str, const BBString *sub, const BBString *replacement);

/** Gives the UCS-2 character code of the first character of a string
 *
 * \param	str The string of which the code point will be extracted
 * \returns	The code point as a BBInt number, or BBSTRING_NOTFOUND in case of an empty string.
 */
BBInt bbStringAsc(const BBString *str) __attribute__ ((pure));



/** Shows that a string was not found within the given data
 *
 * \see		bbStringFind, bbStringFindLast
 */
#define BBSTRING_NOTFOUND	(-1)

/** Finds the first occurance of a substring in a string
 *
 * The given string is searched from the beginning to the end for
 * a matching substring. If none is found, BBSTRING_NOTFOUND is returned.
 *
 * If the given substring is longer than the haystack string in which is
 * searched, BBSTRING_NOTFOUND is returned.
 *
 * \note	Any negative index parameter is set to 0 before searching.
 * \see		bbStringFindLast, bbStringFindAll
 *
 * \param	str The string which is searched in
 * \param	sub The substring which is looked for
 * \param	index The start position for the search
 * \returns	The position of the match in the str if found, otherwise BBSTRING_NOTFOUND.
 */
BBInt bbStringFind(const BBString *str, const BBString *sub, BBInt index) __attribute__ ((pure));

/** Finds the last occurance of a substring in a string
 *
 * In contrast to bbStringFind, the search is performed in reverse.
 * That means the given string's buffer is searched from the end to
 * the beginning. Therefore making this version more performant when
 * it can be expected that the substring is located at the end.
 *
 * \see		bbStringFind
 *
 * \param	str The string which is searched in
 * \param	sub The substring which is looked for
 * \param	index The start position for the search
 * \returns	The position of the match in the str if found, otherwise BBSTRING_NOTFOUND.
 */
BBInt bbStringFindLast(const BBString *str, const BBString *sub, BBInt index) __attribute__ ((pure));

/** Finds all occurances of a subtring in a string
 *
 * This is an easier and faster way to find all substrings in a string than
 * calling bbStringFind multiple times.
 *
 * If the given substring is longer than the haystack string in which is
 * searched, BBNULLARRAY is returned.
 *
 * \author	Fabian
 * \see		bbStringFind
 *
 * \param	str The string which is searched in
 * \param	sub The substring which is looked for
 * \param	index The start position for the search
 * \returns	An array containing all indexes where the substring was found.
 *			If no match was found BBNULLARRAY is returned.
 */
BBArray* bbStringFindAll(const BBString* str, const BBString *sub, BBInt index);

/** Finds the first occurance of any of the given substrings
 *
 * If either the length of \a str or \a all_subs is \c 0, \c BBSTRING_NOTFOUND is returned.
 * The same happens if \a all_subs is not a string array.
 *
 * If the given start \a index is negative, it is set to \c 0.
 * If it is larger than the input string's length, \c BBNULLSTRING is returned.
 *
 * \param	str The string which is searched in
 * \param	all_subs A string array containing all substrings to be searched for
 * \param	index The start position of the search
 * \returns	The first position any of the given substrings was found at, or \c BBSTRING_NOTFOUND.
 */
BBInt bbStringFindAny(const BBString *str, const BBStringArray *all_subs, BBInt index) __attribute__ ((pure));

/** Gives a lowercase version of a string
 *
 * \see		bbStringToUpper
 *
 * \param	str The string to convert to lower case
 * \returns	A new lowercase-only string or BBNULLSTRING
 */
BBString* bbStringToLower(const BBString *str);

/** Gives an uppercase version of a string
 *
 * \see		bbStringToLower
 *
 * \param	str The string to convert to upper case
 * \returns	A new uppercase-only string or BBNULLSTRING
 */
BBString* bbStringToUpper(const BBString *str);

/** Parses an integer number from a string
 *
 * \param	str The string containing the number as character string
 * \returns	The parsed number as a BBInt value
 */
BBInt bbStringToInt(const BBString *str);

/** Parses a single-precision floating point number from a string
 *
 * \param	str The string containing the number as character string
 * \returns	The parsed number as a BBFloat value
 */
BBFloat bbStringToFloat(const BBString *str);

/** Parses a double-precision floating point number from a string
 *
 * \param	str The string containing the number as character string
 * \returns	The parsed number as a BBDouble value
 */
BBDouble bbStringToDouble(const BBString *str);

/** Parses a long integer number from a string
 *
 * \param	str The string containing the number as character string
 * \param	result A pointer to a BBLong variable in which the result is stored
 * \returns	Nothing, the parsed value is stored in the variable pointed to by the result parameter.
 */
void bbStringToLong(const BBString *str, BBLong *result);

/** Converts a string to a null-terminated character string
 *
 * 
 * 
 * \note	The memory references by the returned pointer must be manually freed.
 * \see		bbStringToWString, bbStringToUTF8String
 *
 * \param	str The string to be converted
 * \returns	A pointer to a newly allocated null-terminated string.
 *			NULL in case of an error.
 */
char* bbStringToCString(const BBString *str);

/** Converts a string to a null-terminated wide character string
 *
 * \note	The memory references by the returned pointer must be manually freed.
 *
 * \param	str The string that is to be converted to a null-terminated wide character string
 * \returns	A pointer to the newly allocated wide character string data.
 *			NULL in case of an error.
 */
BBChar* bbStringToWString(const BBString *str);

/** Converts a string to a null-terminated UTF-8 encoded character string
 *
 * This encodes the internal UCS-2 data of a given string to a valid UTF-8 string.
 * Obviously the encoded string will only contain code points from the BMP. 
 *
 * \note	The memory references by the returned pointer must be manually freed.
 *
 * \param	str The string to be converted to a null-terminated UTF-8 string
 * \returns	A pointer to the newly allocated UTF-8 string data.
 *			NULL in case of an error.
 */
char* bbStringToUTF8String(const BBString *str);

/** Splits a string into bits delimited by a separator
 *
 * Searched the input string for occurances of the \a separator and splits
 * it at every found position. If no match is found, the input string is
 * returned as the first element in the result string array.
 *
 * \param	str The string to be split
 * \param	separator The separator that is used to split the string
 * \returns	An array with all the parts of the split string.
 */
BBStringArray* bbStringSplit(const BBString *str, const BBString *separator);

/** Joins an array of bits together using the given string as glue
 *
 */
BBString* bbStringJoin(const BBString *str, const BBStringArray *bits);



//	Additions by BME

/** Checks two strings for equality
 *
 * Two strings equal when they have the same length and all of their
 * characters are the same. The memory location does not matter here.
 *
 * If both strings are empty, the result is BBTRUE.
 *
 * \note	When given two strings pointing to the same location, BBTRUE is returned.
 *
 * \param	str The first string against which the other one is compared
 * \param	rhs The string to compare against the first one
 * \returns	BBTRUE in case the strings equal, otherwise BBFALSE.  
 */
BBBool bbStringEquals(const BBString *str, const BBString *rhs) __attribute__ ((pure));

/** Counts all occurances of substring in a string
 *
 * If either the string or the substring is empty, the result is 0.
 *
 * \param	str The string which is searched
 * \param	sub The substring to search for
 * \returns	The number of times a match for the substring is found in the string.
 */
BBInt bbStringCount(const BBString *str, const BBString *sub) __attribute__ ((pure));

/** Reverses a string
 *
 * If the given string is empty, BBNULLSTRING is returned.
 *
 * \param	str The string which should be reversed
 * \returns	A new string object with reversed characters or the given \a str.
 */
BBString* bbStringReverse(const BBString *str);

/** Rotates a string
 *
 * The rotation is circular, therefore a number of rotations larger than the
 * string's length – positive or negative – will start the rotation again at the
 * beginning or end of the string respectively.
 *
 * If either the given string is too short (<tt>&lt; 2</tt>), \a count is \c 0
 * or equals the string's length, the given string \a str is returned without rotation.
 *
 * \param	str The string to be rotated
 * \param	count The amount of which the characters should be rotated
 * \returns	A new string with every character rotated or the given \a str.
 */
BBString* bbStringRotate(const BBString *str, BBInt count);

/** Repeats and concatenates a string multiple times
 *
 * This adds the given string multiple times to itself.
 * 
 * If the count parameter is zero or less, \c BBNULLSTRING is returned.
 *
 * If the count parameter has the exact value of \c 1,
 * the given string is simply returned.
 *
 * If the length of the resulting string exceeds \c BBSTRING_MAX_LENGTH,
 * the number of concatenations will be reduced until it no longer exceeds that maximum length.
 * In addition to that a \c BBRuntimeException will be thrown in debug mode.
 *
 * \param	str The string which should be repeated
 * \param	count The number of times the string should be repeated
 * \returns	A new string containing the given string multiple times.
 */
BBString* bbStringTimes(const BBString *str, BBInt count);

/** Gives a substring of a string
 *
 * This is an alternate way to get a substring out of a given string.
 * The difference between this one and bbStringSlice is, that here the length
 * of the substring is specified, not the last position in the original string.
 *
 * A negative index results in counting the given value down from the end of the
 * original string. If this goes below zero (when the negative index is too long)
 * \c BBNULLSTRING is returned. If the index is larger than the length of the
 * original string, \c BBNULLSTRING is returned immediately.
 *
 * The length parameter can be longer than the given string. In this case
 * the returned substring is as long as possible. If the given length is
 * zero or lower, \c BBNULLSTRING is returned.
 *
 * \note	The len parameter specifies the \b maximal length of the substring.
 * \see		bbStringSlice
 *
 * \param	str The string from which the substring is taken
 * \param	index The start position from where the substring is taken
 * \param	len The maximum length of the substring.
 * \returns	A new string which is a part of the given one, or BBNULLSTRING.
 */
BBString* bbStringSub(const BBString *str, BBInt index, BBInt len);

/** Checks if the string only consists of lowercase characters
 *
 * When given an empty string as a paremeter \c BBTRUE is returned.
 *
 * \param	str The string to check
 * \retval	BBTRUE if the string only contains lowercase characters
 * \retval	BBFALSE if there is at least one uppercase character
 */
BBBool bbStringIsLower(const BBString *str) __attribute__ ((pure));

/** Checks if the string only consists of lowercase characters
 *
 * When given an empty string as a paremeter \c BBTRUE is returned.
 *
 * \param	str The string to check
 * \retval	BBTRUE if the string only contains uppercase characters
 * \retval	BBFALSE if there is at least one lowercase character
 */
BBBool bbStringIsUpper(const BBString *str) __attribute__ ((pure));

/** Checks if a string only contains ascii characters
 *
 * When given an empty string as a paremeter \c BBTRUE is returned.
 *
 * \param	str The string to be checked
 * \retval	BBTRUE if the string only contains ASCII characters
 * \retval	BBFALSE if there are any character codes above 127
 */
BBBool bbStringIsASCII(const BBString *str) __attribute__ ((pure));

//	intern

char* bbTmpCString(const BBString *str);
BBChar* bbTmpWString(const BBString *str);
char* bbTmpUTF8String(const BBString *str);

BB_END_DECLS

#endif	// BLITZ_STRING_H
