#include "blitz_thread.h"
#include "blitz_object.h"
#include "blitz_app.h"

#ifdef BB_ENABLE_EXPERIMENTAL

extern bool bb_mutex_init(bb_mutex_t *mutex);
extern bool bb_mutex_destroy(bb_mutex_t *mutex);
extern bool bb_mutex_lock(bb_mutex_t *mutex);
extern bool bb_mutex_unlock(bb_mutex_t *mutex);
extern bool bb_mutex_trylock(bb_mutex_t *mutex);

extern bool bb_sem_init(bb_sem_t *sem, size_t count);
extern bool bb_sem_destroy(bb_sem_t *sem);
extern bool bb_sem_wait(bb_sem_t *sem);
extern bool bb_sem_post(bb_sem_t *sem);

#endif	// BB_ENABLE_EXPERIMENTAL

//#define DEBUG_THREADS

//***** Common *****

int _bbNeedsLock;
bb_mutex_t _bbLock;

static int ThreadDataId;

static BBThread *Threads;
static BBThread *DeadThreads;

static BBThread *MainThread;

static void flush_dead_threads()
{
	BBThread **p = &DeadThreads;
	BBThread *thread;
	
	while (thread = *p)
	{
		if (thread->detached)
		{
			*p = thread->succ;
#ifdef BB_OS_WIN32
			CloseHandle(thread->handle);
#endif
			free(thread);
		} else {
			p = &thread->succ;
		}
	}
}

static void add_thread(BBThread *thread)
{
	flush_dead_threads();
	thread->succ = Threads;
	Threads = thread;
}

static void remove_thread(BBThread *thread)
{
	BBThread **p = &Threads;
	BBThread *t;
	
	while (t = *p)
	{
		if (t == thread)
		{
			*p = t->succ;
			if (t->detached) {
#ifdef BB_OS_WIN32
				CloseHandle(t->handle);
#endif
				free(t);
			} else {
				t->succ = DeadThreads;
				DeadThreads = t;
			}
			break;
		} else {
			p = &t->succ;
		}
	}
}

//	API

BBInt bbThreadAllocData()
{
	if (ThreadDataId < BBTHREAD_DATA_SIZE - 1) {
		return ++ThreadDataId;
	}
	
	return 0;
}

void bbThreadSetData(BBInt index, BBObject *data)
{
	BB_ASSERT(index >= 0 && index < BBTHREAD_DATA_SIZE, "Invalid data index.");
	
	bbThreadGetCurrent()->data[index] = data;
}

BBObject* bbThreadGetData(BBInt index)
{
	BB_ASSERT(index >= 0 && index < BBTHREAD_DATA_SIZE, "Invalid data index.");
	
	BBObject *data = bbThreadGetCurrent()->data[index];
	return data ? data : &bbNullObject;
}

//***** Windows threads *****
#ifdef BB_OS_WIN32

static DWORD CurrentThreadTls;

static DWORD WINAPI threadProc(void *p)
{
	BBThread *thread = p;
	
	TlsSetValue(CurrentThreadTls, thread);
	
	DWORD result = (DWORD)thread->proc(thread->data[0]);
	
	BB_LOCK
	remove_thread(thread);
	BB_UNLOCK
	
	return result;
}

void bbThreadStartup()
{
	if (bb_mutex_init(&_bbLock) < 0) {
		exit(-1);
	}
		
	CurrentThreadTls = TlsAlloc();

	BBThread *thread = malloc(sizeof (BBThread));
	
	thread->proc = 0;
	memset(thread->data, 0, sizeof (thread->data));
	
	thread->detached	= 0;
	thread->stackTop	= bbGCStackTop;
	thread->id			= GetCurrentThreadId();
	
	if (!DuplicateHandle(GetCurrentProcess(), GetCurrentThread(),
						 GetCurrentProcess(), &thread->handle, 0,
						 FALSE, DUPLICATE_SAME_ACCESS))
	{
		exit(-1);
	}

	TlsSetValue(CurrentThreadTls, thread);
	
	thread->succ = Threads;
	Threads = thread;
	MainThread = thread;
}

BBThread* bbThreadCreate(BBThreadProc proc, BBObject *data)
{
	BBThread *thread = malloc(sizeof (BBThread));
	
	thread->proc = proc;
	memset(thread->data, 0, sizeof (thread->data));
	
	thread->data[0]	= data;
	thread->detached= 0;
	thread->handle	= CreateThread(0, 0, threadProc, thread, CREATE_SUSPENDED, &thread->id);

	CONTEXT ctx = { CONTEXT_CONTROL };
	GetThreadContext( thread->handle,&ctx );
	thread->stackTop = ctx.Esp;

	BB_LOCK
	add_thread(thread);
	BB_UNLOCK

	_bbNeedsLock = 1;
	
	return thread;
}

void bbThreadDetach(BBThread *thread)
{
	thread->detached = 1;
}

BBObject* bbThreadWait(BBThread *thread)
{
	if (WaitForSingleObject(thread->handle, INFINITE) == WAIT_OBJECT_0)
	{
		BBObject *obj;
		if (GetExitCodeThread(thread->handle, (DWORD*) &obj)) {
			thread->detached = 1;
			return obj;
		} else {
			printf("ERROR! bbThreadWait: GetExitCodeThread failed!\n");
		}
	} else {
		printf("ERROR! bbThreadWait: WaitForSingleObject failed!\n");
	}
	printf("LastError=%i\n", GetLastError());
	
	return &bbNullObject;
}

BBThread* bbThreadGetMain()
{
	return MainThread;
}

BBThread* bbThreadGetCurrent()
{
	return TlsGetValue(CurrentThreadTls);
}

BBInt bbThreadSuspend(BBThread *thread)
{
	return SuspendThread(thread->handle);
}

BBInt bbThreadResume(BBThread *thread)
{
	return ResumeThread(thread->handle);
}

BBThread* _bbThreadLockThreads()
{
	const BBThread * const current = bbThreadGetCurrent();

	for (BBThread *t = Threads; t; t = t->succ)
	{
		if (t != current)
		{
			SuspendThread(t->handle);
			
			CONTEXT ctx = { CONTEXT_INTEGER | CONTEXT_CONTROL };
			GetThreadContext(t->handle, &ctx);
			
			t->locked_regs[0] = ctx.Edi;
			t->locked_regs[1] = ctx.Esi;
			t->locked_regs[2] = ctx.Ebx;
			t->locked_regs[3] = ctx.Edx;
			t->locked_regs[4] = ctx.Ecx;
			t->locked_regs[5] = ctx.Eax;
			t->locked_regs[6] = ctx.Ebp;
			t->locked_sp = ctx.Esp;
		}
	}
	
	return Threads;
}

void _bbThreadUnlockThreads()
{
	const BBThread *current = bbThreadGetCurrent();
	
	BBThread *t;
	for (t = Threads; t; t = t->succ)
	{
		if (t != current) {
			ResumeThread(t->handle);
		}
	}
}

#else	// BB_OS_WIN32 -> now POSIX threads

#if defined BB_OS_LINUX
#	define MUTEX_RECURSIVE 1
#elif defined BB_OS_MACOS
#	define MUTEX_RECURSIVE 2
#endif

pthread_mutexattr_t _bb_mutexattr;

static BBThread *Threads;
static pthread_key_t CurrentThreadTls;

static void suspendSigHandler(int sig)
{
	BBThread *thread = pthread_getspecific(CurrentThreadTls);
	
	thread->locked_sp = bbGCRootRegs(thread->locked_regs);
	
#ifdef DEBUG_THREADS
	printf("In suspendSigHandler! thread=%p locked_sp=%p\n", thread, thread->locked_sp);
	fflush(stdout);
#endif
	
	bb_sem_post(&thread->acksema);
	
	// wait for resume - apparently very naughty!
	bb_sem_wait(&thread->runsema);

#ifdef DEBUG_THREADS	
	printf("Got resume!\n");
	fflush(stdout);
#endif
}

void bbThreadStartup()
{
	if (pthread_mutexattr_init(&_bb_mutexattr) < 0) {
		exit(-1);
	}
	
	if (pthread_mutexattr_settype(&_bb_mutexattr, MUTEX_RECURSIVE) < 0) {
		exit(-1);
	}
	
	if (pthread_key_create(&CurrentThreadTls, 0) < 0) {
		exit(-1);
	}
		
	if (bb_mutex_init(&_bbLock) < 0) {
		exit(-1);
	}
		
	struct sigaction act;
	memset(&act, 0, sizeof (act));
	
	act.sa_handler	= suspendSigHandler;
	act.sa_flags	= SA_RESTART;
		
	if (sigaction(SIGUSR2, &act, 0) < 0) {
		exit(-1);
	}
		
	BBThread *thread = malloc(sizeof (BBThread));
	memset(thread->data, 0, sizeof (thread->data));
	
	thread->proc		= 0;
	thread->detached	= 0;
	thread->suspended	= 0;
	thread->handle		= pthread_self();
	
	if (!bb_sem_init(&thread->runsema, 0)
	 || !bb_sem_init(&thread->acksema, 0))
	{
		exit(-1);
	}

	thread->stackTop = bbGCStackTop;
	pthread_setspecific(CurrentThreadTls, thread);
	
	thread->succ = Threads;
	Threads = thread;
	MainThread = thread;
}

static void* threadProc(void *ptr)
{
	BBThread *thread = ptr;
	
	thread->stackTop = bbGCRootRegs(thread->locked_regs);
	pthread_setspecific(CurrentThreadTls, thread);
	
	BB_LOCK
	add_thread(thread);
	BB_UNLOCK
	
	bb_sem_post(&thread->acksema);
	bb_sem_wait(&thread->runsema);
	
#ifdef DEBUG_THREADS
	printf("Thread %p added, stackTop=%p\n", thread, thread->stackTop);
	fflush(stdout);
#endif
	
	void *result = thread->proc(thread->data[0]);
	
	BB_LOCK
	remove_thread(thread);
	BB_UNLOCK
	
	bb_sem_destroy(&thread->runsema);
	bb_sem_destroy(&thread->acksema);
	
#ifdef DEBUG_THREADS
	printf("Thread %p removed\n", thread);
	fflush(stdout);
#endif
	
	return result;
}

BBThread* bbThreadCreate(BBThreadProc proc, BBObject *data)
{
	BBThread *thread = malloc(sizeof (BBThread));
	memset(thread->data, 0, sizeof (thread->data));
	
	thread->proc		= proc;
	thread->data[0]		= data;
	thread->detached	= 0;
	thread->suspended	= 1;
	
	if (bb_sem_init(&thread->runsema, 0))
	{
		if (bb_sem_init(&thread->acksema, 0))
		{
			if (pthread_create(&thread->handle, 0, threadProc, thread) >= 0)
			{
				bb_sem_wait (&thread->acksema);
				_bbNeedsLock = 1;
				return thread;
			}
			bb_sem_destroy(&thread->acksema);
		}
		bb_sem_destroy(&thread->runsema);
	}
	
	free(thread);
	return NULL;
}

void bbThreadDetach(BBThread *thread)
{
	thread->detached = 1;
	pthread_detach(thread->handle);
}

BBObject* bbThreadWait(BBThread *thread)
{
	BBObject *obj = 0;
	thread->detached = 1;
	pthread_join(thread->handle, &obj);
	return obj;
}

BBThread* bbThreadGetMain()
{
	return MainThread;
}

BBThread* bbThreadGetCurrent()
{
	return pthread_getspecific(CurrentThreadTls);
}

BBInt bbThreadSuspend(BBThread *thread)
{
	BB_LOCK
	
	int num_suspended = thread->suspended++;
	
	if (num_suspended == 0) {
		pthread_kill(thread->handle, SIGUSR2);
		bb_sem_wait(&thread->acksema);
	}

	BB_UNLOCK
	
	return num_suspended;
}

BBInt bbThreadResume(BBThread *thread)
{
	BB_LOCK
	
	int num_suspended = thread->suspended--;
	
	if (num_suspended == 1) {
		bb_sem_post(&thread->runsema);
	}
	
	BB_UNLOCK
	
	return num_suspended;
}

BBThread* _bbThreadLockThreads()
{
	const BBThread * const current = bbThreadGetCurrent();
	
	for (BBThread *t = Threads; t; t = t->succ)
	{
		if (t != current) {
			if (!t->suspended++) {
				pthread_kill(t->handle, SIGUSR2);
				bb_sem_wait(&t->acksema);
			}
		}
	}
	
	return Threads;
}

void _bbThreadUnlockThreads()
{
	const BBThread * const current = bbThreadGetCurrent();
	
	for (BBThread *t = Threads; t; t = t->succ)
	{
		if (t != current) {
			if (!--t->suspended) {
				bb_sem_post(&t->runsema);
			}
		}
	}
}

#endif	// !BB_OS_WIN32

//***** Atomic ops *****
#ifdef BB_CPU_PPC

BBInt bbAtomicCAS(volatile BBInt *target, BBInt old, BBInt new_val)
{
	BBInt oldval;
	BBInt result = 0;

	__asm__ __volatile__(
		"1:lwarx %0,0,%2\n"   /* load and reserve              */
		"cmpw %0, %4\n"      /* if load is not equal to 	*/
		"bne 2f\n"            /*   old, fail			*/
		"stwcx. %3,0,%2\n"    /* else store conditional         */
		"bne- 1b\n"           /* retry if lost reservation      */
		"li %1,1\n"	     /* result = 1;			*/
		"2:\n"
		: "=&r"(oldval), "=&r"(result)
		: "r"(target), "r"(new_val), "r"(old), "1"(result)
		: "memory", "cc");

	return result;
}

BBInt bbAtomicAdd(volatile BBInt *target, BBInt incr)
{
	BBInt old;
	for (;;)
	{
		old = *target;
		if (bbAtomicCAS(target, old, old + incr)) {
			return old;
		}
	}
}

#else	// BB_CPU_PPC

BBInt bbAtomicCAS(volatile BBInt *target, BBInt old, BBInt new_val)
{
	char result;

	__asm__ __volatile__(
		"lock; cmpxchgl %3, %0; setz %1"
		: "=m"(*target), "=q"(result)
		: "m"(*target), "r" (new_val), "a"(old) : "memory");

	return (BBInt) result;
}

BBInt bbAtomicAdd(volatile BBInt *target, BBInt incr)
{
	BBInt result;

	__asm__ __volatile__ ("lock; xaddl %0, %1" :
			"=r" (result), "=m" (*target) : "0" (incr), "m" (*target)
			: "memory");

	return result;
}

#endif	// !BB_CPU_PPC
