/** \file	blitz_thread.h
 * \brief	Functionality to utilize multi-threading capabilities
 *
 * \author	Mark Sibly
 */

#ifndef BLITZ_THREAD_H
#define BLITZ_THREAD_H

#include "_common.h"

#ifdef BB_OS_LINUX
#	include <semaphore.h>
#endif

#ifdef BB_OS_MACOS
#	include <mach/semaphore.h>
#	include <mach/task.h>
#	include <mach/mach_init.h>

#	include <libkern/OSAtomic.h>	// test
#endif

#ifndef BB_OS_WIN32
#	include <pthread.h>
#	include <unistd.h>
#	include <signal.h>
#endif	// !BB_OS_WIN32

BB_BEGIN_DECLS

#ifdef BB_OS_LINUX
typedef pthread_mutex_t		bb_mutex_t;
typedef sem_t				bb_sem_t;
#endif

#ifdef BB_OS_MACOS
typedef pthread_mutex_t		bb_mutex_t;
typedef semaphore_t			bb_sem_t;
#endif

#ifdef BB_OS_WIN32
typedef CRITICAL_SECTION	bb_mutex_t;
typedef HANDLE				bb_sem_t;
#endif

#ifndef BB_OS_WIN32
extern pthread_mutexattr_t _bb_mutexattr;
#endif	// !BB_OS_WIN32

// Replacing macros with actual inline functions
#ifdef BB_ENABLE_EXPERIMENTAL

inline bool bb_mutex_init(bb_mutex_t *mutex)
{
#ifdef BB_OS_WIN32
	InitializeCriticalSection(mutex);
	return true;
#else
	return pthread_mutex_init(mutex, &_bb_mutexattr) >= 0;
#endif
}

inline bool bb_mutex_destroy(bb_mutex_t *mutex)
{
#ifdef BB_OS_WIN32
	DeleteCriticalSection(mutex);
	return true;
#else
	return pthread_mutex_destroy(mutex) >= 0;
#endif	
}

inline bool bb_mutex_lock(bb_mutex_t *mutex)
{
#ifdef BB_OS_WIN32
	DeleteCriticalSection(mutex);
	return true;
#else
	return pthread_mutex_lock(mutex) >= 0;
#endif
}

inline bool bb_mutex_unlock(bb_mutex_t *mutex)
{
#ifdef BB_OS_WIN32
	LeaveCriticalSection(mutex);
	return true;
#else
	return pthread_mutex_lock(mutex) >= 0;
#endif	
}

inline bool bb_mutex_trylock(bb_mutex_t *mutex)
{
#ifdef BB_OS_WIN32
	return TryEnterCriticalSection(mutex) != 0;
#else
	return pthread_mutex_trylock(mutex) == 0;
#endif
}



inline bool bb_sem_init(bb_sem_t *sem, size_t count)
{
#if defined (BB_OS_LINUX)
	return sem_init(sem, 0, count) >= 0;
#elif defined (BB_OS_MACOS)
	return semaphore_create(mach_task_self(), sem, SYNC_POLICY_FIFO, count) >= 0;
#elif defined (BB_OS_WIN32)
	*sem = CreateSemaphore(0, count, 0x7fffffff, 0);
	return *sem != 0;
#endif
}

inline bool bb_sem_destroy(bb_sem_t *sem)
{
#if defined (BB_OS_LINUX)
	return sem_destroy(sem);
#elif defined (BB_OS_MACOS)
	return semaphore_destroy(mach_task_self(), *sem);
#elif defined (BB_OS_WIN32)
	CloseHandle(*sem);
#endif
}

inline bool bb_sem_wait(bb_sem_t *sem)
{
#if defined (BB_OS_LINUX)
	return sem_wait(sem);
#elif defined (BB_OS_MACOS)
	return semaphore_wait(*sem);
#elif defined (BB_OS_WIN32)
	return WaitForSingleObject(*sem, INFINITE);
#endif
}

inline bool bb_sem_post(bb_sem_t *sem)
{
#if defined (BB_OS_LINUX)
	return sem_post(sem);
#elif defined (BB_OS_MACOS)
	return semaphore_signal(*sem);
#elif defined (BB_OS_WIN32)
	return ReleaseSemaphore(*sem, 1, 0);
#endif
}

#else	// BB_ENABLE_EXPERIMENTAL

#ifdef BB_OS_WIN32

#	define bb_mutex_init(MUTPTR)		(InitializeCriticalSection(MUTPTR), 1)
#	define bb_mutex_destroy(MUTPTR)		DeleteCriticalSection(MUTPTR)
#	define bb_mutex_lock(MUTPTR)		EnterCriticalSection(MUTPTR)
#	define bb_mutex_unlock(MUTPTR)		LeaveCriticalSection(MUTPTR)
#	define bb_mutex_trylock(MUTPTR)		(TryEnterCriticalSection(MUTPTR) != 0)

#	define bb_sem_init(SEMPTR,COUNT)	((*(SEMPTR) = CreateSemaphore(0, (COUNT), 0x7fffffff, 0)) != 0)
#	define bb_sem_destroy(SEMPTR)		CloseHandle(*(SEMPTR))
#	define bb_sem_wait(SEMPTR)			WaitForSingleObject(*(SEMPTR), INFINITE)
#	define bb_sem_post(SEMPTR)			ReleaseSemaphore(*(SEMPTR), 1, 0)

#else	// BB_OS_WIN32

extern pthread_mutexattr_t _bb_mutexattr;

#	define bb_mutex_init(MUTPTR)		(pthread_mutex_init((MUTPTR),&_bb_mutexattr) >= 0)
#	define bb_mutex_destroy(MUTPTR)		pthread_mutex_destroy(MUTPTR)
#	define bb_mutex_lock(MUTPTR)		pthread_mutex_lock(MUTPTR)
#	define bb_mutex_unlock(MUTPTR)		pthread_mutex_unlock(MUTPTR)
#	define bb_mutex_trylock(MUTPTR)		(pthread_mutex_trylock(MUTPTR) == 0)

#endif	// !BB_OS_WIN32

#ifdef BB_OS_MACOS
#	define bb_sem_init(SEMPTR,COUNT)	(semaphore_create(mach_task_self(), (SEMPTR), SYNC_POLICY_FIFO, (COUNT)) >= 0)
#	define bb_sem_destroy(SEMPTR)		semaphore_destroy(mach_task_self(), *(SEMPTR))
#	define bb_sem_wait(SEMPTR)			semaphore_wait(*(SEMPTR))
#	define bb_sem_post(SEMPTR)			semaphore_signal(*(SEMPTR))
#endif	// BB_OS_MACOS

#ifdef BB_OS_LINUX
#	define bb_sem_init(SEMPTR, COUNT)	(sem_init((SEMPTR), 0, (COUNT)) >= 0)
#	define bb_sem_destroy				sem_destroy
#	define bb_sem_wait					sem_wait
#	define bb_sem_post					sem_post
#endif	// BB_OS_LINUX

#endif	// !BB_ENABLE_EXPERIMENTAL

#ifdef BB_OS_WIN32
#	define BBTHREAD_NUM_REGS	7
#elif BB_CPU_PPC
#	define BBTHREAD_NUM_REGS	19
#else
#	define BBTHREAD_NUM_REGS	4
#endif

/** Signature of a function that contains the thread's logic
 *
 * \see		bbThreadCreate
 */
typedef BBObject *(*BBThreadProc)(BBObject*);

typedef struct BBThread			BBThread;

#define BBTHREAD_DATA_SIZE		32

/** Represents a single thread of execution
 *
 */
struct BBThread
{
	BBThread		*succ;
	BBThreadProc	proc;
	void			*data[BBTHREAD_DATA_SIZE];
	int				detached;
	void			*stackTop;
	void			*locked_sp;
	int				locked_regs[BBTHREAD_NUM_REGS];
#ifdef BB_OS_WIN32
	HANDLE			handle;
	DWORD			id;
#else
	pthread_t		handle;
	int				suspended;
	bb_sem_t		runsema;
	bb_sem_t		acksema;
#endif	// !BB_OS_WIN32
};

//	Sets up the threading environment
//	Automatically called by bbStartup (blitz_app.c)
void bbThreadStartup();

/** Creates a new thread
 *
 * \param	entry The callback function executing the thread's business logic
 * \paramm	data Data passed to the callback functio
 * \returns	A new thread that will execute the callback function
 */
BBThread* bbThreadCreate(BBThreadProc entry, BBObject *data);

/**
 *
 */
void bbThreadDetach(BBThread *thread);

/**
 *
 */
BBObject* bbThreadWait(BBThread *thread);

/** Gets the application's main thread of execution
 *
 * \returns	A pointer to the main thread
 */
BBThread* bbThreadGetMain();

/** Gets the current thread in which a function is running
 *
 * Can be used with \c bbThreadGetMain to determine if a code section
 * is currently being executed in the main thread.
 *
 * \returns	A pointer to the currently executing thread
 */
BBThread* bbThreadGetCurrent();

/**
 *
 */
BBInt bbThreadSuspend(BBThread *thread);

/**
 *
 */
BBInt bbThreadResume(BBThread *thread);



/**
 *
 */
BBInt bbThreadAllocData();

/**
 *
 */
void bbThreadSetData(BBInt index, BBObject *data);

/**
 *
 */
BBObject* bbThreadGetData(BBInt index);

// These MUST be inside a BB_LOCK/BB_UNLOCK
BBThread* _bbThreadLockThreads();
void _bbThreadUnlockThreads();

/**
 *
 */
BBInt bbAtomicCAS(volatile BBInt *target, BBInt old_value, BBInt new_value);

/** Atomically adds a number to an integer
 *
 * \param	target A pointer to the value that should be increased
 * \param	incr The number which will be added to the \a target number
 * \returns	The result of the addition as an integer
 */
BBInt bbAtomicAdd(volatile BBInt *target, BBInt incr);

//	Internal locks //

extern int _bbNeedsLock;
extern bb_mutex_t _bbLock;

#define BB_LOCK		if (_bbNeedsLock) { bb_mutex_lock(&_bbLock); }
#define BB_UNLOCK	if (_bbNeedsLock) { bb_mutex_unlock(&_bbLock); }

//	Deprecated //

#ifndef BB_DISABLE_DEPRECATED
#	define BB_THREADREGS		BBTHREAD_NUM_REGS
#endif	// !BB_DISABLE_DEPRECATED

BB_END_DECLS

#endif	// BLITZ_THREAD_H
