/**
 * \file	blitz_types.h
 * \brief	Introduces most used data types
 *
 * BlitzMax has two kinds of data types: Primitive and complex types.
 *
 * Primitive types store one value or a pointer to one. They include:
 * BBByte, BBShort, BBInt, BBLong, BBFloat, BBDouble and BBChar
 *
 * Complex data types are BlitzMax objects, consisting of a C struct
 * to store the data and a  \c BBClass that defines the functionality.
 * They include: \c BBObject, \c BBString, \c BBArray and every user-defined class
 *
 * \author	Mark Sibly
 * \author	Fabian Niemann
 */

#ifndef BLITZ_TYPES_H
#define BLITZ_TYPES_H

#include <float.h>
#include <limits.h>
#include <stdint.h>

#include "_common.h"

#ifdef BB_ENV_C99
#	include <stdbool.h>
#endif

BB_BEGIN_DECLS

#if defined BB_ENV_C99

typedef int8_t						BBInt8;
typedef uint8_t						BBUInt8;
typedef int16_t						BBInt16;
typedef uint16_t					BBUInt16;
typedef int32_t						BBInt32;
typedef uint32_t					BBUInt32;
typedef int64_t						BBInt64;
typedef uint64_t					BBUInt64;

#elif defined _MSC_VER

typedef __int8						BBInt8;
typedef unsigned __int8				BBUInt8;
typedef __int16						BBInt16;
typedef unsigned __int16			BBUInt16;
typedef __int32						BBInt32;
typedef unsigned __int32			BBUInt32;
typedef __int64						BBInt64;
typedef unsigned __int64			BBUInt64;

#else

typedef signed char					BBInt8;
typedef unsigned char				BBUInt8;
typedef short						BBInt16;
typedef unsigned short				BBUInt16;
typedef int							BBInt32;
typedef unsigned int				BBUInt32;
typedef long long					BBInt64;
typedef unsigned long long			BBUInt64;

#endif

//	Primitive data types

/** The BlitzMax "Byte" data type
 *
 * \see		BBBYTE_MIN, BBBYTE_MAX
 */
typedef unsigned char				BBByte;

/** The BlitzMax "Short" data type
 *
 * \see		BBSHORT_MIN, BBSHORT_MAX
 */
typedef unsigned short				BBShort;

/** The BlitzMax "Int" data type
 *
 * \see		BBINT_MIN, BBINT_MAX
 */
typedef signed int					BBInt;

/** The BlitzMax "Long" data type
 *
 * \see		BBLONG_MIN, BBLONG_MAX
 */
typedef BBInt64						BBLong;

/** The BlitzMax "Float" data type
 */
typedef float						BBFloat;

/** The BlitzMax "Double" data type
 */
typedef double						BBDouble;

/** The minimum value a variable of type BBByte can hold */
#define BBBYTE_MIN					((BBByte) 0x00)
/** The maximum value a variable of type BBByte can hold */
#define BBBYTE_MAX					((BBByte) 0xff)

/** The minimum value a variable of type BBShort can hold */
#define BBSHORT_MIN					((BBShort) 0x0000)
/** The maximum value a variable of type BBShort can hold */
#define BBSHORT_MAX					((BBShort) 0xffff)

/** The minimum value a variable of type BBInt can hold */
#define BBINT_MIN					((BBInt) 0x80000000)
/** The maximum value a variable of type BBInt can hold */
#define BBINT_MAX					((BBInt) 0x7fffffff)

/** The minimum value a variable of type BBLong can hold */
#define BBLONG_MIN					((BBLong) 0x800000000000000L)
/** The maximum value a variable of type BBLong can hold */
#define BBLONG_MAX					((BBLong) 0x7ffffffffffffffL)

/** The minimum value a variable of type BBFloat can hold */
#define BBFLOAT_MIN					((BBFloat) -FLT_MAX)
/** The maximum value a variable of type BBFloat can hold */
#define BBFLOAT_MAX					((BBFloat) FLT_MAX)

/** The minimum value a variable of type BBDouble can hold */
#define BBDOUBLE_MIN				((BBDouble) -DBL_MAX)
/** The maximum value a variable of type BBDouble can hold */
#define BBDOUBLE_MAX				((BBDouble) DBL_MAX)

/** The BlitzMax internal character type
 *
 * Represents a single UCS-2 character code.
 * 
 * \see		BBCHAR_MIN, BBCHAR_MAX
 */
typedef BBUInt16					BBChar;

/** The minimum value a variable of type BBChar can hold */
#define BBCHAR_MIN					((BBChar) 0x0000)

/** The maximum value a variable of type BBChar can hold */
#define BBCHAR_MAX					((BBChar) 0xffff)

/** Boolean data type
 *
 * Although BlitzMax does not have a designated boolean data type
 * as of now, this is useful for improving readability.
 *
 * \see		BBTRUE, BBFALSE
 *
 * \note	In BlitzMax everything with a value 0 is considered 'False'
 *			and everything non-zero is considered 'True'. No matter the type.
 */
typedef enum { BBFALSE, BBTRUE }	BBBool;

#ifdef BB_ENABLE_EXPERIMENTAL

/** Stores a size or length
 *
 * This should be used as an equivalent to C's "size_t" type.
 * Since the BlitzMax language and runtime are based on signed integers,
 * the value of a \c BBSize variable should also never exceed \c BBINT_MAX.
 *
 * \see		BBSIZE_MIN, BBSIZE_MAX
 */
typedef unsigned int				BBSize;

#else
typedef int							BBSize;
#endif	// !BB_ENABLE_EXPERIMENTAL

/** The minimal value a variable of type BBSize can hold */
#define BBSIZE_MIN					((BBSize) 0)

/** The maximum value a variable of type BBSize *SHOULD* store
 *
 * Since the BlitzMax language aswell as it's runtime operate mostly on a
 * signed integer basis, values stored in variables of this data type should
 * not exceed \c BBINT_MAX. Otherwise there could be problems with legacy code.
 *
 */
#define BBSIZE_MAX					((BBSize) INT_MAX)

//	Complex data types

typedef struct BBClass				BBClass;
typedef struct BBObject				BBObject;
typedef struct BBString				BBString;
typedef struct BBArray				BBArray;

extern const char *bbVoidTypeTag;	// "?"
extern const char *bbByteTypeTag;	// "b"
extern const char *bbShortTypeTag;	// "s"
extern const char *bbIntTypeTag;	// "i"
extern const char *bbLongTypeTag;	// "l"
extern const char *bbFloatTypeTag;	// "f"
extern const char *bbDoubleTypeTag;	// "d"
extern const char *bbStringTypeTag;	// "$"
extern const char *bbObjectTypeTag;	// ":Object"
extern const char *bbBytePtrTypeTag;// "*b"

//	Additions by BME

/** \internal
 *
 * \brief	An integer identifier mapping a type tag
 *
 * BlitzMax uses a system of null-terminated strings internally to inform
 * about the type of an argument, for example in the array functions.
 * 
 * Sometimes it can be benefiticial to execution speed to just look up information
 * about a data type in an array instead of checking by hand. The problem is,
 * that by using the first letters of the type tag string as the array's keys
 * would waste memory without need. Type keys solve this problem.
 *
 * By mapping type tags to type keys you can create arrays containing extra
 * information for each type without wasting precious memory.
 */
enum BBTypeKey
{
	/** Identifies the type BBByte */
	BBTYPEKEY_BYTE,
	
	/** Identifies the type BBShort */
	BBTYPEKEY_SHORT,
	
	/** Identifies the type BBInt */
	BBTYPEKEY_INT,
	
	/** Identifies the type BBLong */
	BBTYPEKEY_LONG,
	
	/** Identifies the type BBFloat */
	BBTYPEKEY_FLOAT,
	
	/** Identifies the type BBDouble */
	BBTYPEKEY_DOUBLE,
	
	/** Identifies a pointer to a variable */
	BBTYPEKEY_POINTER,

	/** Identifies the type void */
	BBTYPEKEY_VOID,	
	
	/** Identifies a function */
	BBTYPEKEY_FUNCTION,
	
	/** Identifies the type BBObject* */
	BBTYPEKEY_OBJECT,
	
	/** Identifies the type BBString* */
	BBTYPEKEY_STRING,
	
	/** Identifies the type BBArray* */
	BBTYPEKEY_ARRAY
};

typedef enum BBTypeKey				BBTypeKey;

/** Checks if a type key is valid
 *
 * All known keys are considered to be valid.
 *
 * \param	key The type key to check as an integer
 * \retval	true If the key is valid
 * \retval	false If the key is not valid
 */
#define BBTYPEKEY_ISVALID(key)		((key) >= BBTYPEKEY_BYTE && (key) <= BBTYPEKEY_ARRAY)

/** Checks if a type key identifies a numeric variable
 *
 * The following data types are considered to be "numeric":
 * \c BBByte, \c BBShort and \c BBInt, \c BBLong, \c BBFloat and \c BBDouble
 *
 * Therefore the following type keys will check positive for complex data type:
 * \c BBTYPEKEY_BYTE, \c BBTYPEKEY_SHORT and \c BBTYPEKEY_INT, \c BBTYPEKEY_LONG,
 * \c BBTYPEKEY_FLOAT and \c BBTYPEKEY_DOUBLE
 *
 * \see		BBTYPEKEY_ISPRIMITIVE
 *
 * \param	key The type key to check
 * \retval	true If the key identifies a numeric variable
 * \retval	false If the does not identify a valid numeric variable
 */
#define BBTYPEKEY_ISNUMERIC(key)	((key) >= BBTYPEKEY_BYTE && (key) <= BBTYPEKEY_DOUBLE)

/** Checks if a type key identifies a primitive data type
 *
 * The following data types are considered to be "primitive":
 * \c BBByte, \c BBShort, \c BBInt, \c BBLong, \c BBFloat, \c BBDouble,
 * \c void*, \c void and \c (*)
 *
 * Therefore the following type keys will check positive for primitive data type:
 * \c BBTYPEKEY_BYTE, \c BBTYPEKEY_SHORT and \c BBTYPEKEY_INT, \c BBTYPEKEY_LONG,
 * \c BBTYPEKEY_FLOAT, \c BBTYPEKEY_DOUBLE, \c BBTYPEKEY_POINTER, \c BBTYPEKEY_VOID,
 * \c BBTYPEKEY_FUNCTION
 *
 * \see		BBTYPEKEY_ISCOMPLEX, BBTYPEKEY_ISNUMERIC
 *
 * \param	key The type key to check
 * \retval	true If the key identifies a primitive data type
 * \retval	false If the does not identify a valid primitive variable
 */
#define BBTYPEKEY_ISPRIMITIVE(key)	((key) >= BBTYPEKEY_BYTE && (key) <= BBTYPEKEY_OBJECT)

/** Checks if a type key identifies a complex data type
 *
 * The following data types are considered to be "complex":
 * \c BBObject, \c BBString and \c BBArray
 *
 * Therefore the following type keys will check positive for complex data type:
 * \c BBTYPEKEY_OBJECT, \c BBTYPEKEY_STRING and \c BBTYPEKEY_ARRAY
 *
 * \see		BBTYPEKEY_ISPRIMITIVE
 *
 * \param	key The type key to check
 * \retval	true If the key identifies a complex data type
 * \retval	false If the does not identify a valid complex variable
 */
#define BBTYPEKEY_ISCOMPLEX(key)	((key) >= BBTYPEKEY_OBJECT && (key) <= BBTYPEKEY_ARRAY)

/** Gets the corresponding BBTypeKey for a type tag
 *
 * \see		BBTypeKey
 * \param	tag The type tag as a null-terminated string
 * \returns	The corresponding type key
 */
#define BBTYPETAG_KEY(tag)			bbToTypeKeyData[(tag)[0]]

/** \internal
 *
 * \brief	Contains the conversion information from TypeTag to \c BBTypeKey.
 */
extern const BBTypeKey bbToTypeKeyData[];

/** Contains useful information about each type key.
 *
 * \see		BBTypeKey
 */
struct BBTypeKeyInfo
{
	/** The size of one instance in bytes
	 *
	 * Basically this is just the value of 'sizeof ($DATATYPE)' stored.
	 * It can be used for allocating structures able to store different
	 * type values quickly.
	 */
	BBSize			size;
	
	/** The value to which each uninitialized instance should be set
	 *
	 * This can be a simple value '0' or one of the built-in object type \c NULL pointers.
	 */
	void			*init;
	
	/** The identifier of the type
	 *
	 * This contains a human readable string with the BlitzMax name of the type. 
	 */
	const char		*ident;
};

typedef struct BBTypeKeyInfo		BBTypeKeyInfo;

/** Stores the TypeKey information for all known BlitzMax data types
 *
 * \see		BBTypeKeyInfo
 */
extern const BBTypeKeyInfo bbTypeKeyInfo[];

#define BBTYPEKEY_INFO(key)			bbTypeKeyInfo[(key)]
#define BBTYPEKEY_SIZE(key)			bbTypeKeyInfo[(key)].size
#define BBTYPEKEY_INIT(key)			bbTypeKeyInfo[(key)].init
#define BBTYPEKEY_IDENT(key)		bbTypeKeyInfo[(key)].ident

#define BBTYPETAG_INFO(tag)			bbTypeKeyInfo[BBTYPETAG_KEY(tag)]
#define BBTYPETAG_SIZE(tag)			BBTYPETAG_INFO(tag).size
#define BBTYPETAG_INIT(tag)			BBTYPETAG_INFO(tag).init
#define BBTYPETAG_IDENT(tag)		BBTYPETAG_INFO(tag).ident



#ifndef BB_DISABLE_DEPRECATED

//	Don't use these, as they are confusing (BBSTRING is a pointer, BBINT isn't)

typedef unsigned char				BBBYTE;
typedef unsigned short				BBSHORT;
typedef signed int					BBINT;
typedef BBInt64						BBLONG;
typedef float						BBFLOAT;
typedef double						BBDOUBLE;
typedef BBClass*					BBCLASS;
typedef BBObject*					BBOBJECT;
typedef BBString*					BBSTRING;
typedef BBArray*					BBARRAY;

#endif	// !BB_DISABLE_DEPRECATED

BB_END_DECLS

#endif	// BLITZ_TYPES_H
