/**
 * \file	blitz_exception.h
 * \brief	Basic exception class
 *
 * A minimalist common interface for all exceptions. 
 *
 * \author	Fabian Niemann
 */

#ifndef BLITZ_EX_EXCEPTION_H
#define BLITZ_EX_EXCEPTION_H

#include "_common.h"

BB_BEGIN_DECLS

/** Base for all exceptions
 *
 * Although any object can be thrown as an exception, extending this
 * base exception gives you the advantage of a common interface as well as
 * the possiblity to trace back a chain of exceptions that caused each other.
 *
 * \extends	BBObject
 */
struct BBException
{
	/** The exception's clas
	 *
	 * This should always point to a class that extends bbExceptionClass.
	 *
	 * \see		bbExceptionClass
	 */
	BBClass			*clas;
	
	int				refs;
	
	/** The cause of the exception
	 * 
	 * The cause is a previous exception that caused this one to be thrown.
	 * If there is no such exception, this must be BBNULL.
	 *
	 * \see		bbExceptionGetCause, bbExceptionInitCause
	 */
	BBException		*cause;
};

//	For internal usage only
enum bb_exception_vfnkey_e
{
	__BBEXCEPTION_VFN_INITCAUSE__,	
	__BBEXCEPTION_VFN_GETCAUSE__
};

struct bb_exception_vfns_s
{
	// Initializes an exception's cause using its own class' virtual function
	void			(*InitCause)(BBException *ex, BBException *cause);	
	
	// Gets an exception's cause using its own class' virtual function
	BBException*	(*GetCause)(BBException *ex);
};

typedef struct bb_exception_vfns_s	__BBExceptionVfns;

extern BBClass bbExceptionClass;

void bbExceptionCtor(BBException *ex);
void bbExceptionDtor(BBException *ex);

BBInt bbExceptionCompare(const BBException *ex, const BBException *rhs);
BBException* bbExceptionClone(BBException *ex);

/** Gets the cause of an exception
 *
 * \see		bbExceptionInitCause
 *
 * \param	ex The exception whose cause is returned
 * \returns	The cause of the exception or BBNULL if no known one was given.
 */
BBException* bbExceptionGetCause(const BBException *ex);

/** Initializes the cause of an exception
 *
 * Initializing an exception's cause can only be done once, usually in the
 * classes own "Create" function. However it can also be done afterwards, in case
 * the process of creation does not automatically take care of it.
 *
 * \note	An exception can not be caused by itself.
 *			Therefore initializing an exception object with itself will not work.
 * 
 * \param	ex The exception whose cause should be set
 * \param	cause The previous exception that caused the current one
 */
void bbExceptionInitCause(BBException *ex, BBException *cause);

BB_END_DECLS

#endif	// BLITZ_EX_EXCEPTION_H
